/*************************************************************************
**  source       * Acc_Control.c
**               *
**  description  * Solid accelerator Control APIs usage sample:
**               *
**               * (C) Copyright Solid Information Technology Ltd 2005
\*************************************************************************/
/* Common system headers */

#ifdef SS_NT
#  include <windows.h>
#endif

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#include <solidodbc3.h>

/* Solid SQL Accelerator API */
#include <sscapi.h>

/* Solid portability header */
#include "c.h"

/* global declarations */

SQLCHAR *username   = "dba"; 
SQLCHAR *password   = "dba"; 
SQLCHAR *catalog    = "dba";

#ifndef bool
#  define bool int
#endif

/* prototypes */

SscRetT StartAccelerator(SscServerT* sf_handle, SQLCHAR* uid, SQLCHAR* pwd,
                         SQLCHAR* catalog, SscStateT runflags,
                         SQLCHAR* TestTask);
SscRetT StopAccelerator(SscServerT sf_handle, bool force);
SscRetT ConnectToServer(SQLHENV* henv, SQLHDBC* hdbc, SQLHSTMT* hstmt);

void Disconnect(SQLHENV* henv, SQLHDBC* hdbc, SQLHSTMT* hstmt);
void PrintError(SQLSMALLINT handleType, SQLHANDLE handle, SQLINTEGER line,
                SQLCHAR* fname);
void CheckReturnT(SQLRETURN rc, SQLSMALLINT handletype, SQLHANDLE handle,
                  SQLINTEGER line, SQLCHAR* file);
void CheckReturnTN(SQLRETURN rc, SQLSMALLINT handletype, SQLHANDLE handle,
                   SQLINTEGER line, SQLCHAR* file);
void SscHeader(char* str);
void SscRunTest(SscServerT SscHandle, SQLHENV henv, SQLHDBC hdbc);

void tssc_setstate(SQLHDBC hdbc, SscServerT SscHandle);
void tssc_running(SscServerT SscHandle);
void tssc_lserver(SscServerT SscHandle);
void tssc_notifier_cp(SscServerT SscHandle, SQLHDBC hdbc);
void tssc_getactive(SscServerT SscHandle);
void tssc_taskclassprio(SscServerT SscHandle); 
void tssc_suspend(SscServerT SscHandle);

void PrintAssertionInfo(SQLCHAR* fname, SQLINTEGER lineno);

/* utility macros */

#define SscPrintf printf

#define CheckRetCode(rc, handleType, handle) \
            CheckReturnT(rc, handleType, handle, __LINE__, (SQLCHAR*)__FILE__);

#define CheckRetCodeError(rc, handleType, handle)        \
            CheckReturnTN(rc, handleType, handle,        \
                          __LINE__, (SQLCHAR*)__FILE__);

#define CheckSscRetT(rc) if (!rc) \
                            PrintAssertionInfo((SQLCHAR*)__FILE__, __LINE__);
#define SscAssertT assert
#define MAXCOLS 50


/*##**********************************************************************\
 *
 *      CheckReturnT
 *
 * Check return code for error.  If so, display it and assert.
 *
 * Parameters:
 *      rc - in
 *          return code to check
 *
 *      handleType - in
 *          type of handle
 *
 *      handle - in, use
 *          handle to which the return code belongs
 *
 *      line - in
 *          line number of call
 *
 *      file - in
 *          filename from which we're called.
 *
 */
void CheckReturnT(SQLRETURN rc, SQLSMALLINT handleType, SQLHANDLE handle,
                  SQLINTEGER line, SQLCHAR* file)
{
        if (rc != SQL_SUCCESS) {
            PrintError(handleType, handle, line, file);
        }

        if (!(rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO)) {
            PrintAssertionInfo(file, line);
        }
}

/*##**********************************************************************\
 *
 *      CheckReturnTN
 *
 * Check return code expecting an error. If it's an error, display it.
 * If not, assert.
 *
 * Parameters:
 *      rc - in
 *          return code to check
 *
 *      handleType - in
 *          type of handle
 *
 *      handle - in, use
 *          handle to which the return code belongs
 *
 *      line - in
 *          linenumber of call
 *
 *      file - in
 *          filename from which we're called
 *
 */
void CheckReturnTN(SQLRETURN rc, SQLSMALLINT handleType, SQLHANDLE handle,
                   SQLINTEGER line, SQLCHAR* file)
{
        if (rc == 1 || rc == SQL_ERROR) {
            PrintError(handleType, handle, line, file);
        }

        if (rc != SQL_ERROR) {
            PrintAssertionInfo(file, line);
        }
}

/*##**********************************************************************\
 *
 *      PrintAssertionInfo
 *
 * "Assert" ie. exit.
 *
 * Parameters:
 *      fname - in
 *          file of assertion
 *
 *      lineno - in
 *          line of assertion
 *
 */
void PrintAssertionInfo(SQLCHAR* fname, SQLINTEGER lineno)
{
        printf("\nASSERTION FAILED in %d@%s\n", lineno, fname);
        exit(0);
}

/*##**********************************************************************\
 *
 *      SscHeader
 *
 * Print a nice frame for text. Oh, and also the text in it.
 *
 * Parameters:
 *      str - in
 *          text in the frame
 *
 */
void SscHeader(char* str)
{
        SscPrintf("\n**************************************************************************\n");
        SscPrintf("  %s", str);                    
        SscPrintf("\n**************************************************************************\n\n");
}
        
/*##**********************************************************************\
 *
 *      Disconnect
 *
 * Disconnect from server and free resources.
 *
 * Parameters:
 *      henv - in, use
 *          env handle of the connection
 *
 *      hdbc - in, use
 *          dbc handle of the connection
 *
 *      hstmt - in, use
 *          statement handle used with the connection
 *
 */
void Disconnect(SQLHENV* henv, SQLHDBC* hdbc, SQLHSTMT* hstmt)
{
        SQLFreeHandle(SQL_HANDLE_STMT, *hstmt);
        SQLDisconnect(*hdbc);
        SQLFreeHandle(SQL_HANDLE_DBC, *hdbc);
        SQLFreeHandle(SQL_HANDLE_ENV, *henv);
}

/*##**********************************************************************\
 *
 *      StartAccelerator
 *
 * Start Accelerator server
 *
 * Parameters:
 *      h - in, use
 *          handle for the server
 *
 *      user - in
 *          username
 *
 *      pwd - in
 *          password
 *
 *      catalog - in
 *          catalog
 *
 *      runflags - in
 *          settings for the server
 *
 *      fname - in
 *          name of the binary in which accelerator is used
 *
 * Return value:
 *      SSC_SUCCESS if successful, something else if not.
 *
 */
SscRetT StartAccelerator(SscServerT* h, SQLCHAR* user,
                         SQLCHAR* pwd, SQLCHAR* catalog, SscStateT runflags,
                         SQLCHAR* fname)
{
        SQLINTEGER  g_argc;
        SQLCHAR*    g_argv[4];
        SscRetT     rc = SSC_SUCCESS;
  
        g_argc = 4;
        g_argv[0] = fname; 
        g_argv[1] = user;
        g_argv[2] = pwd;  
        g_argv[3] = catalog;

        SscPrintf("\nStarting the Accelerator ...\n");
        rc = SSCStartServer(g_argc, g_argv, h, runflags);  
        if (rc != SSC_SUCCESS) {
            SscPrintf("\nFailed to start the Accelerator, returned %d\n", rc);
        } else {
            SscPrintf("\nAccelerator started successfully\n");
        }

        return (rc);
}

/*##**********************************************************************\
 *
 *      StopAccelerator
 *
 * Stop Accelerator server
 *
 * Parameters:
 *      h - in, use
 *          server handle
 *
 *      force - in
 *          force stop or not
 *
 * Return value:
 *      SSC_SUCCESS if successful, something else if not.
 *
 */
SscRetT StopAccelerator(SscServerT h, bool force)
{
        SscRetT rc = SSC_SUCCESS;

        rc = SSCStopServer(h, force);
        if (rc != SSC_SUCCESS) {
            SscPrintf("\nFailed to stop the Accelerator returned %d\n", rc);
        } else {
            SscPrintf("\nAccelerator stopped successfully\n");
        }
        
        return(rc);
}

/*##**********************************************************************
 *
 *      ConnectToServer
 *
 * Connect to default server, allocate handles, set autocommit ON.
 *
 * Parameters:
 *      henv - use, out
 *          allocated environment
 *
 *      hdbc - use, out
 *          allocated connection
 *
 *      hstmt - use, out
 *          allocated statement
 *
 * Return value:
 *      SQL_SUCCESS if ok, something else if not.
 *
 */
SscRetT ConnectToServer(SQLHENV* henv, SQLHDBC* hdbc, SQLHSTMT* hstmt)
{
        SQLRETURN rc = SQL_SUCCESS;

        /* Connect to the Accelerator server */

        rc = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, henv);
        CheckRetCode(rc, SQL_HANDLE_ENV, *henv);

        rc = SQLSetEnvAttr(*henv, SQL_ATTR_ODBC_VERSION, (void *)SQL_OV_ODBC3,
                           SQL_NTS);
        CheckRetCode(rc, SQL_HANDLE_ENV, *henv);
        
        rc = SQLAllocHandle(SQL_HANDLE_DBC, *henv, hdbc);
        CheckRetCode(rc, SQL_HANDLE_ENV, *henv);
        
        rc = SQLConnect(*hdbc, "", SQL_NTS, username, SQL_NTS,
                        password, SQL_NTS);
        CheckRetCode(rc, SQL_HANDLE_DBC, *hdbc);
        
        rc = SQLSetConnectAttr(*hdbc, SQL_ATTR_AUTOCOMMIT,
                               (SQLPOINTER)SQL_AUTOCOMMIT_ON, SQL_NTS);
        CheckRetCode(rc, SQL_HANDLE_DBC, *hdbc);

        rc = SQLAllocHandle(SQL_HANDLE_STMT, *hdbc, hstmt);
        CheckRetCode(rc, SQL_HANDLE_DBC, *hdbc);   

        return(rc);
}

/*==========================================================================
FunctionName    : PrintError
Purpose         : if error occured, get the error and print it
Parameters      : 
    handleType  : type of handle
        handle  : handle
        line    : line number at which test failed
        file    : name of the file
Returns         : none
==========================================================================*/
/*##**********************************************************************\
 *
 *      PrintError
 *
 * Get error record from a handle and print it.
 *
 * Parameters:
 *      handleType - in
 *          type of handle
 *
 *      handle - in, use
 *          handle in error state
 *
 *      line - in
 *          linenumber of call
 *
 *      fname - in
 *          filename in which error occured
 *
 */
void PrintError(SQLSMALLINT handleType, SQLHANDLE handle, SQLINTEGER line,
                SQLCHAR* fname)
{
        SQLCHAR      SqlState[6];
        SQLCHAR      MsgText[SQL_MAX_MESSAGE_LENGTH];
        SQLINTEGER   naterr;
        SQLRETURN    rc;
        
        rc = SQLGetDiagRec(handleType, handle, 1, (SQLCHAR *)&SqlState,
                           (SQLINTEGER *)&naterr, (SQLCHAR *)&MsgText,
                           255, NULL);
        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {
            SscPrintf("\nError %d@%s:[%s][%s]\n", line, fname, SqlState,
                      MsgText);
        }                       
}

/*##**********************************************************************\
 *
 *      TSscNotifyFunc
 *
 * Notify callback, returning success.  Called by the server.
 *
 * Parameters:
 *      SscHandle - in
 *          server handle
 *
 *      what - in
 *
 *      userdata - in
 *          the data we gave server when we set the notifier.
 *
 * Return value:
 *      SSC_SUCCESS
 *
 * Limitations:
 *
 * Globals used:
 */
int SSC_CALL TSscNotifyFunc(SscServerT SscHandle, SscNotFunT what,
                            void* userdata)
{
        SscPrintf("\nNotifier: %s\n", userdata);
        return SSC_SUCCESS;
}

/*##**********************************************************************\
 *
 *      TSscNotifyFuncA
 *
 * Notifier callback, returning abort.  Called by the server.
 *
 * Parameters:
 *      SscHandle - in
 *          server handle
 *
 *      what - in
 *
 *      userdata - in
 *          the data we gave to server when we set the notifier.
 *
 * Return value:
 *      SSC_ABORT
 *
 */
int SSC_CALL TSscNotifyFuncA(SscServerT SscHandle, SscNotFunT what,
                             void* userdata)
{
        SscPrintf("\nNotifier: %s\n", userdata);
        return SSC_ABORT;
}

/*=========================================================================
FunctionName    : TSscResultSet
Purpose         : Used to get fetch and display the result set
Parameters      : 
        handles : Statement Handle :hstmt
Returns         : RowCount,the no of records fetched.
=========================================================================*/
/*##**********************************************************************\
 *
 *      TSscResultSet
 *
 * Fetch and display a result set.
 *
 * Parameters:
 *      hstmt - in, use
 *          statement to read
 *
 * Return value:
 *      Number of rows fetched.
 *
 */
SQLINTEGER TSscResultSet(SQLHSTMT hstmt)
{
        RETCODE     rc=0;
        SQLCHAR     ColName[254]={0};
        SQLSMALLINT ColType;
        SQLSMALLINT ColNameLen;
        SQLSMALLINT Nullable;
        SQLLEN      Collen[MAXCOLS]={0};
        SQLSMALLINT Scale,ColNumber;
        SQLLEN      Len[MAXCOLS]={0};
        SQLSMALLINT ColCount=0;
        SQLCHAR     Data[MAXCOLS][110]={0};
        SQLINTEGER  i=0, RowCount=0;


        rc = SQLNumResultCols(hstmt, &ColCount);
        CheckRetCode(rc,SQL_HANDLE_STMT,hstmt);

        printf("\n");
        for (i = 0; i < ColCount; i++) {
            ColNumber = i + 1;
            rc = SQLDescribeCol(hstmt, ColNumber, ColName, sizeof(ColName),
                                &ColNameLen, &ColType, &Collen[i], &Scale,
                                &Nullable);
            CheckRetCode(rc, SQL_HANDLE_STMT, hstmt);
            printf("%s\t", ColName);

            rc = SQLBindCol(hstmt, ColNumber, SQL_C_CHAR, Data[i], 110,
                            &Len[i]);
            CheckRetCode(rc, SQL_HANDLE_STMT, hstmt);
        }

        printf("\n");
        for (i = 0; i < ColCount; i++) {
            printf("--------");
        }
        printf("\n");

        while ((rc = SQLFetch(hstmt)) != SQL_NO_DATA_FOUND) {
            RowCount++;
            for (i = 0; i < ColCount; i++) {
                if (Len[i] == SQL_NULL_DATA)
                    printf("NULL\t");
                else
                    printf("%s\t", Data[i]);
            }
            printf("\n");
        }
        printf("\nAll selected rows (%d) printed\n", RowCount);

        SQLFreeStmt(hstmt, SQL_CLOSE);
        SQLFreeStmt(hstmt, SQL_UNBIND);

        return(RowCount);
}

#ifdef SS_VXW
#define main ac_main
#endif

int SS_CDECL main(int argc, char** argv)
{
        SQLHENV    henv  = NULL;
        SQLHDBC    hdbc  = NULL;
        SQLHSTMT   hstmt = NULL;
        SQLRETURN  rc;
        SscRetT    src;
        SscServerT SscHandle;
  
        /* Start the server */
        src = StartAccelerator(&SscHandle, "-Udba", "-Pdba", "-Cdba",
                               SSC_STATE_OPEN, argv[0]);
        if (src != SSC_SUCCESS) {
            exit(1);
        }
            
        /* Allocate handles */
        rc = SQLAllocEnv(&henv);
        CheckRetCode(rc, SQL_HANDLE_ENV, henv);

        rc = SQLAllocConnect(henv, &hdbc);
        CheckRetCode(rc, SQL_HANDLE_ENV, henv);

        /* Set autocommit off */
        rc = SQLSetConnectAttr(hdbc, SQL_ATTR_AUTOCOMMIT, SQL_AUTOCOMMIT_OFF,
                               0);
        CheckRetCode(rc, SQL_HANDLE_DBC, hdbc);
    
        /* Perform a set of actions on the server */
        SscRunTest(SscHandle, henv, hdbc);

        rc = SQLFreeConnect(hdbc);
        CheckRetCode(rc,SQL_HANDLE_DBC, hdbc);

        rc = SQLFreeEnv(henv);
        CheckRetCode(rc,SQL_HANDLE_ENV, henv);

        /* Stop the server */
        src = StopAccelerator(SscHandle, 1);
        SscAssertT(src == SSC_SUCCESS);

        printf("\n=======\nSuccess\n");

        return(0);
}                   

/*##**********************************************************************\
 *
 *      SscRunTest
 *
 * Run all demonstrations. Server must be started.
 *
 * Parameters:
 *      SscHandle - in, use
 *          server handle
 *
 *      henv - in, use
 *          env handle
 *
 *      hdbc - in, use
 *          connection
 *
 */
void SscRunTest(SscServerT SscHandle, SQLHENV henv, SQLHDBC hdbc)
{
        tssc_running(SscHandle);            /* SSCIsRunning */ 
        tssc_setstate(hdbc, SscHandle);     /* SSCSetState */
        tssc_lserver(SscHandle);            /* SSCIsThisLocalServer*/ 
        tssc_notifier_cp(SscHandle, hdbc);  /* SSCSetNotifier */
        tssc_getactive(SscHandle);          /* SSCGetActiveTasks*/
        tssc_taskclassprio(SscHandle);      /* SSCSetTaskClassPrio and
                                             * SSCGetTaskClassPrio */
        tssc_suspend(SscHandle);            /* SSCSuspendTaskClass,
                                             * SSCGetSuspendedTaskClass and
                                             * SSCResumeTaskClass */
}

/*##**********************************************************************\
 *
 *      tssc_setstate
 *
 * Open and close the server by setting runflags.
 *
 * Parameters:
 *      hdbc - in, use
 *          handle for connection
 *
 *      SscHandle - in, use
 *          server handle
 *
 */
void tssc_setstate(SQLHDBC hdbc, SscServerT SscHandle)
{
        SscRetT    src;
        SQLRETURN  rc;
       
        SscHeader("SSCSetState");

        /* Connect to the server */
        rc = SQLConnect(hdbc, "", SQL_NTS, "dba", SQL_NTS, "dba", SQL_NTS);
        CheckRetCode(rc, SQL_HANDLE_DBC, hdbc);
        
        /* Disconnect the user */
        rc = SQLDisconnect(hdbc);
        CheckRetCode(rc, SQL_HANDLE_DBC, hdbc);

        /* Set the runflag to ~SSC_STATE_OPEN, ie. CLOSED */
        SscPrintf("\nSetting the SSCSetState to Closed\n");
        src = SSCSetState(SscHandle, ~SSC_STATE_OPEN);
        SscAssertT(src == SSC_SUCCESS);

        /* Now check the remote connection, which should be denied */
        SscPrintf("\nNow Trying to establish connection to the Accelarator (should fail) ...\n");
        rc = SQLConnect(hdbc, "", SQL_NTS, "dba", SQL_NTS, "dba", SQL_NTS);
        CheckRetCodeError(rc, SQL_HANDLE_DBC, hdbc);     
    
        /* Set it back to SSC_STATE_OPEN */
        src = SSCSetState(SscHandle, SSC_STATE_OPEN);
        SscAssertT(src == SSC_SUCCESS);
}

/*##**********************************************************************\
 *
 *      tssc_running
 *
 * Check if server is running.
 *
 * Parameters:
 *      SscHandle - in, use
 *          server handle
 *
 */
void tssc_running(SscServerT SscHandle)
{
        SscRetT src = 0;

        SscHeader("SSCIsRunning");

        /* Checking whether the Accelerator is running or not */
        src = SSCIsRunning(SscHandle);
        if (src != 0) {
            SscPrintf("\nAccelerator is running\n");
        } else {
            SscPrintf("\nAccelerator is not running\n");
        }
}

/*##**********************************************************************\
 *
 *      tssc_lserver
 *
 * Check that server is a local server
 *
 * Parameters:
 *      SscHandle - in, use
 *          server handle
 *
 */
void tssc_lserver(SscServerT SscHandle)
{
        SscRetT    src = 0;
        SQLRETURN  rc  = 0;

        SscHeader("SSCIsThisLocalServer");
        
        /* Check for local server */
        src = SSCIsThisLocalServer();
        if (src == 1) {
            SscPrintf("\nThis is a Local Server\n");
        } else {
            SscPrintf("\nThis is NOT a Local Server!\n");
        }
        
        SscAssertT(src == 1);
}

/*##**********************************************************************\
 *
 *      tssc_notifier_cp
 *
 * Set a notifier to accept a checkpoint, then make a checkpoint.
 *
 * Parameters:
 *      SscHandle - in, use
 *          server handle
 *
 *      hdbc - in, use
 *          connection
 *
 */
void tssc_notifier_cp(SscServerT SscHandle, SQLHDBC hdbc)
{
        SscRetT   src = 0;
        SQLHSTMT  hstmt1 = NULL;
        SQLRETURN rc = 0;

        SscHeader("SSCSetNotifier");
    
        /* Set a notifier to wait for checkpoint creation request */

        src = SSCSetNotifier(SscHandle, SSC_NOTIFY_CHECKPOINT_REQUEST,
                             TSscNotifyFunc, "SSC_NOTIFY_CHECKPOINT-SUCCESS");
        SscAssertT(src == SSC_SUCCESS);

        /* Create a checkpoint. Our notifier is triggered and our message, 
         * "SSC_NOTIFY_CHECKPOINT-SUCCESS", is passed back to us via our
         * notifier callback function. Callback returns SSC_SUCCESS to server
         * and the checkpoint may then proceed.
         */
        
        /* Connect to the server */
        rc = SQLConnect(hdbc, "", SQL_NTS, "dba", SQL_NTS, "dba", SQL_NTS);
        CheckRetCode(rc, SQL_HANDLE_DBC, hdbc);

        rc = SQLAllocStmt(hdbc, &hstmt1);
        CheckRetCode(rc, SQL_HANDLE_DBC, hdbc);

        /* Make checkpoint */
        rc = SQLExecDirect(hstmt1, "ADMIN COMMAND 'MCP'", SQL_NTS);
        CheckRetCode(rc, SQL_HANDLE_STMT, hstmt1);
        TSscResultSet(hstmt1);

        /* Disconnect the user */
        rc = SQLDisconnect(hdbc);
        CheckRetCode(rc, SQL_HANDLE_DBC, hdbc);

}

/*##**********************************************************************\
 *
 *      tssc_getactive
 *
 * Get active tasks from the server.
 *
 * Parameters:
 *      SscHandle - in, use
 *          server handle
 *
 */
void tssc_getactive(SscServerT SscHandle)
{
        SscRetT      src = 0;
        SscTaskSetT  tsetT;

        SscHeader("SSCGetActiveTaskClass");

        /* Get the active tasks from the server */
        SscPrintf("Getting the active tasks from the server ...\n\n"); 
        tsetT = SSCGetActiveTaskClass(SscHandle);

        if (tsetT == 0xff) {
            SscPrintf("Task SSC_TASK_ALL is active\n");
        } else {
            if (!tsetT) {
                SscPrintf("None of the tasks are active\n");
            } else {
                if (tsetT & SSC_TASK_CHECKPOINT) {
                    SscPrintf("Task SSC_TASK_CHECKPOINT is active\n");
                }
                if (tsetT & SSC_TASK_BACKUP) {
                    SscPrintf("Task SSC_TASK_BACKUP is active\n");
                }
                if (tsetT & SSC_TASK_MERGE) {
                    SscPrintf("Task SSC_TASK_MERGE is active\n");
                }
                if (tsetT & SSC_TASK_LOCALUSERS) {
                    SscPrintf("Task SSC_TASK_LOCALUSERS is active\n");
                }
                if (tsetT & SSC_TASK_REMOTEUSERS) {
                    SscPrintf("Task SSC_TASK_REMOTEUSERS is active\n");
                }
                if (tsetT & SSC_TASK_SYNC_HISTCLEAN) {
                    SscPrintf("Task SSC_TASK_SYNC_HISTCLEAN is active\n");
                }
                if (tsetT & SSC_TASK_SYNC_MESSAGE) {
                    SscPrintf("Task SSC_TASK_SYNC_MESSAGE is active\n");
                }
                if (tsetT & SSC_TASK_HOTSTANDBY) {
                    SscPrintf("Task SSC_TASK_HOTSTANDBY is active\n");
                }
            }   
        }
}

/*##**********************************************************************
 *
 *      tssc_taskclassprio
 *
 * Demonstrate task class priority functions.
 *
 * Parameters:
 *      SscHandle - in, use
 *          server handle
 *
 */
void tssc_taskclassprio(SscServerT SscHandle)
{
        SscRetT    src = 0;
        SscTaskPrioT tprioT;

        SscHeader("SSCSetTaskClassPrio and SSCGetTaskClassPrio");
    
        /* Setting the task class priorities */
        SscPrintf("Set checkpoint task priority to SSC_PRIO_HIGH\n");
        src = SSCSetTaskClassPrio(SscHandle,
                                  SSC_TASK_CHECKPOINT, SSC_PRIO_HIGH);
        SscAssertT(src == SSC_SUCCESS);

        SscPrintf("Get checkpoint task priority:\n");
        src = SSCGetTaskClassPrio(SscHandle, SSC_TASK_CHECKPOINT, &tprioT);
        SscAssertT(src == SSC_SUCCESS);
        SscAssertT(tprioT == SSC_PRIO_HIGH);

        SscPrintf("  Checkpoint priority is SSC_PRIO_HIGH\n");
        
}

/*##**********************************************************************
 *
 *      tssc_suspend
 *
 * Demonstrate suspension and resuming of a task.
 *
 * Parameters:
 *      SscHandle - in, use
 *          server handle
 *
 */
void tssc_suspend(SscServerT SscHandle)
{
        SscRetT     src = 0;
        SscTaskSetT tsetT;

        SscHeader("SSCSuspendTaskClass, SSCGetSuspendedTaskClass and SSCResumeTaskClass");
        
        /* Suspend a task */
        SscPrintf("Suspending task: SSC_TASK_BACKUP\n");
        src = SSCSuspendTaskClass(SscHandle, SSC_TASK_BACKUP);
        SscAssertT(src == SSC_SUCCESS);

        /* Get suspended tasks */
        SscPrintf("Get suspended tasks:\n");
        tsetT = SSCGetSuspendedTaskClass(SscHandle);
        SscAssertT((tsetT & SSC_TASK_BACKUP) != 0);
        SscPrintf("  SSCGetSuspendedTaskClass returned SSC_TASK_BACKUP\n");
        
        /* Resume the task */
        SscPrintf("\nResume task: SSC_TASK_BACKUP\n");
        src = SSCResumeTaskClass(SscHandle, SSC_TASK_BACKUP);
        SscAssertT(src == SSC_SUCCESS);
        SscPrintf("  Task resumed\n\n");
}

/* EOF */
