/*-----------------------------------------------------------------------------
 * COPYRIGHT (c) 2001-2003, Solid Information Technology.  All Rights Reserved.
 * NAME: aclib_diskless.c
 * PURPOSE:
 *     This is a C program that uses the SSCStartDisklessServer() 
 *     function in the Solid Database Engine AcceleratorLib to run 
 *     a database server without a disk drive.  
 *
 * POSSIBLE IMPROVEMENTS:
 *     1) This demo uses simplified error-handling.
 *        In a production program, the error-handling should be more
 *        sophisticated.
 * ----------------------------------------------------------------------------
 */

/* Common system headers */
#ifdef SS_NT
#  include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <solidodbc3.h>

/* Solid AcceleratorLib header */
#include "sscapi.h"

/* Solid portabilty header */
#include "c.h"

/* This string contains a copy of the solid.ini configuration file, since 
 * a diskless server cannot read the solid.ini file from the disk. 
 */
char* ini_string =
        "; Diskless + Accelerator\n"
        "; Note that even a diskless server needs FileSpec\n"
        "; info -- primarily to indicate size (not location)\n"
        "; of storage/memory.\n"
        "[IndexFile]\n"
        "FileSpec_1= solidM.db 10485760\n"
        "CacheSize=8388608\n"
        "[Com]\n"
        "Listen=TCP 2315\n";

extern char* lic_string;

/* Global variables.
 * We made these global to simplify the demo slightly, but in most
 * larger applications you will probably want to use local variables
 * for the connection handle and the statement handle, and perhaps
 * for other of these variables also.
 */
SQLHENV    henvm;   /* ODBC "Environment" handle */
SQLHDBC    hdbcm;   /* ODBC "Connection" handle */
SQLHSTMT   hstmtm;  /* ODBC "Statement" handle */
SQLCHAR*   listenat = "localserver"; /* dsn */ 
char*      username = "dba";
char*      password = "dba";
SQLPOINTER com = SQL_AUTOCOMMIT_OFF;

/* The application gets this "handle" to the server when it starts the
 * server and uses it when stopping the server. 
 */
SscServerT serverHandle;

/* Constant definitions */
#define DATALEN  255

/* Maximum length of an SQLSTATE (used in error messages). */
#define MAX_SQLSTATE_NAME_LEN 80

/* Maximum length of a piece of data returned in a query. */
#define MAX_COL_NAME_LEN  128

/* Maximum number of columns in a result set. */
#define MAXCOLS  100



/* Utility macros for error checking */
#define CHECKHStmt(r, hstmt)  \
        if ( ((r) != SQL_SUCCESS) && ((r) != SQL_SUCCESS_WITH_INFO) )  {   \
            print_error(r, SQL_HANDLE_STMT, hstmt, __FILE__, __LINE__);    \
            }

#define CHECKDbc(r, hdbc)  \
        if ( ((r) != SQL_SUCCESS) && ((r) != SQL_SUCCESS_WITH_INFO) )  {   \
            print_error(r, SQL_HANDLE_DBC, hdbc, __FILE__, __LINE__);      \
            }

#define CHECKEnv(r, henv)  \
        if ( ((r) != SQL_SUCCESS) && ((r) != SQL_SUCCESS_WITH_INFO) )  {   \
            print_error(r, SQL_HANDLE_ENV, henv, __FILE__, __LINE__);      \
            }





/*##************************************************************************
 *
 *                  print_error()
 * 
 *       Print an error message.
 * 
 * Parameters
 *     rc.........: Return code
 *     handle_type: Handle type identifier (handle of stmt, or DBC or ...)
 *     handle.....: handle
 *     filename...: The name of the file in which this error was detected
 *                  (this is usually set by passing __FILE__ in the call).
 *     line.......: The line number at which the error occurred (usually
 *                  just pass __LINE__ in the call).
 * Returns: none
 */
void print_error(SQLRETURN rc, SQLSMALLINT handle_type, SQLHANDLE handle,
                 char* filename, int line)
{
        SQLCHAR      SQLstate[MAX_SQLSTATE_NAME_LEN];
        SQLINTEGER   errorcode;
        SQLCHAR      errormsg[SQL_MAX_MESSAGE_LENGTH];
        SQLSMALLINT  msglen;
        SQLRETURN    r;
        
        printf("Error (%d) encountered!\n", rc);
        
        /* Print more error info, if available */
        r = SQLGetDiagRec(handle_type, handle, 1, (SQLCHAR *)SQLstate,
                          (SQLINTEGER *) &errorcode, (SQLCHAR *) errormsg,
                          SQL_MAX_MESSAGE_LENGTH - 1, &msglen);
        if (r != SQL_NO_DATA_FOUND)  {
            printf("%s\n", errormsg);
            printf("SQLSTATE = %s\n", SQLstate);
            printf("Error Code = %d\n", errorcode);
        }
        printf("File name = %s, line number = %d\n\n", filename, line);
}




/*##************************************************************************
 *
 *         PrepareForConnect
 *
 *         This allocates some "handles" that are needed when connecting
 *         to the server and when doing other things.
 *
 * PARAMETERS:
 *     henv_ptr - a pointer to the henv (ODBC Environment); this is 
 *                updated to point to a valid Environment handle
 *     hdbc_ptr - a pointer to the hdbc (ODBC Connection); this is updated
 *                to point to a valid Connection variable.
 *     hstmt_ptr - a pointer to the hstmt (ODBC Statement); this is updated
 *                to point to a valid ODBC Statement variable
 *
 * RETURN VALUE:
 *     If there are no errors, this returns what SQLAllocHandle returns
 *     when there's no error (presumably SQL_SUCCESS).
 *     If there is an error, it returns that
 *     error (presumably SQL_ERROR or SQL_INVALID_HANDLE).  
 */
SQLRETURN PrepareForConnect(SQLHENV *henv_ptr, SQLHDBC *hdbc_ptr,
                            SQLHSTMT *hstmt_ptr)
{
        
        SQLRETURN r = SQL_ERROR;
        
        puts("DDDIAGNOSTIC: SQLAllocHandle for HENV");
        r = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, henv_ptr);
        CHECKEnv(r, *henv_ptr);
        if (r == SQL_ERROR)  {
            return r;
        }
        
        puts("DDDIAGNOSTIC: SQLSetEnvAttr");
        r = SQLSetEnvAttr(*henv_ptr, SQL_ATTR_ODBC_VERSION,
                          (void *)SQL_OV_ODBC3, SQL_NTS);
        CHECKEnv(r, *henv_ptr);
        if (r == SQL_ERROR)  {
            return r;
        }
        
        puts("DDDIAGNOSTIC: SQLAllocHandle for HDBC");
        r = SQLAllocHandle(SQL_HANDLE_DBC, *henv_ptr, hdbc_ptr);
        CHECKDbc(r, *hdbc_ptr);
        if (r == SQL_ERROR)  {
            return r;
        }
        
        return r;
}



/*##************************************************************************
 *
 *         Connect
 *
 *         This tries to connect to the server (using the SQLConnect
 *         function).
 *
 * PARAMETERS:
 *     henv_ptr - a pointer to the ODBC Environment variable.
 *     hdbc_ptr - a pointer to the ODBC Connection variable.
 *     hstmt_ptr - a pointer to the ODBC Statement variable. 
 *     sname - the connect string for the server, for example "tcp 2315"
 *             or "tcp 128.52.196.055" or "tcp foo.bar.com 2315". 
 *     uname - user name of database administrator.
 *     pwd   - password of the database administrator.
 *
 *     Note that the input strings (sname, uname, and pwd) must be
 *     null-terminated strings.
 *
 *     Note that you must have already allocated an henv, and hdbc, and an 
 *     an hstmt, e.g. by calling PrepareForConnect().
 *
 * RETURN VALUE:
 *     If there are no errors, this returns what SQLAllocHandle returns
 *     when there's no error (presumably SQL_SUCCESS).
 *     If there is an error, it returns that
 *     error (presumably SQL_ERROR or SQL_INVALID_HANDLE).  
 */
SQLRETURN Connect(SQLHENV *henv_ptr, SQLHDBC *hdbc_ptr, SQLHSTMT *hstmt_ptr,
                  char *sname, char *uname, char *pwd)
{
        
        SQLRETURN r = SQL_ERROR;
        
        puts("INFO: Calling SQLConnect");
        r = SQLConnect(*hdbc_ptr, sname, SQL_NTS, uname, SQL_NTS,
                       pwd, SQL_NTS);
        CHECKDbc(r, *hdbc_ptr);
        if (r == SQL_INVALID_HANDLE)  {
            puts("ERROR: invalid handle while trying to connect.");
            return r;
        }
        if (r == SQL_ERROR)  {
            puts("ERROR: SQL_ERROR when tried to connect.");
            return r;
        }
        
        puts("INFO: SQLSetConnectAttr");
        SQLSetConnectAttr(*hdbc_ptr, SQL_ATTR_AUTOCOMMIT, com,
                          SQL_IS_INTEGER);
        
        puts("INFO: SQLAllocHandle (hstmt)");
        r = SQLAllocHandle(SQL_HANDLE_STMT, *hdbc_ptr, hstmt_ptr);
        CHECKDbc(r, *hdbc_ptr);
        
        return r;
}




/*##************************************************************************
 *
 *                  execute_sql()
 *
 *                  Executes an SQL query and displays (on the screen)
 *                  any values that were returned.
 * 
 * Parameters
 *      hstmt - Statement Handle
 *      sql   - query to be executed
 *      cbsql - Length of query OR the special value SQL_NTS (which 
 *              indicates that the query is a Null-Terminated String).
 * Returns
 *      This returns the error code that we get when we execute the
 *      SQL statement that was passed to this routine.  Note that
 *      if that statement executes correctly, but there is an error during
 *      some subsequent step (such as printing column names or fetching
 *      the actual data rows), you will NOT get any indication of that
 *      subsequent error.  
 */
int execute_sql(SQLHSTMT hstmt, SQLCHAR *sql, SQLINTEGER cbsql)
{
        /* The Return Code when we execute the statement. */
        SQLRETURN       rc;
        /* The Return Code from various functions involved in printing
         * column names, etc.
         */
        SQLRETURN       rc1;
        /* The Return Code when we fetch the next row of data. */
        SQLRETURN       rc2;
        /* The number of columns returned by the query. */
        SQLSMALLINT     ColCount = 0;
        SQLCHAR         ColName[MAX_COL_NAME_LEN + 1] = { 0 };
        SQLLEN          Len[MAXCOLS] = { 0 };
        SQLCHAR         Data[MAXCOLS][DATALEN] = { 0 };
        SQLSMALLINT     i = 0;
        SQLINTEGER      RowCount = 0;
        SQLSMALLINT	    iCol = 0;

        /* Check that the SQL statement string is not NULL. */
        assert(sql);

        printf("%s\n", sql);	

        /* Execute the SQL command. */
        rc = SQLExecDirect(hstmt, sql, cbsql);

        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {

            rc1 = SQLNumResultCols(hstmt, &ColCount);
            if (ColCount > 0)   {
                /* Print the column names. */
                for (i = 0; i < ColCount; i++)   {
                    iCol = i + 1;
                    rc1 = SQLDescribeCol(hstmt, iCol, ColName, 
                                         sizeof(ColName),
                                         NULL, NULL, NULL, NULL, NULL);
                    printf("%s\t", ColName);
                    rc1 = SQLBindCol(hstmt, iCol, SQL_C_CHAR, 
                                     Data[iCol-1],
                                     sizeof(Data[iCol-1]),
                                     &Len[iCol-1]);
                }
                printf("\n");
                rc2 = SQLFetch(hstmt);   
                while (rc2 != SQL_NO_DATA_FOUND && rc2 != SQL_ERROR) {
                    RowCount++;
                    for (i = 0; i < ColCount; i++)   {
                        if (Len[i] == SQL_NULL_DATA)   {
                            strcpy(Data[i], "NULL");
                        }
                        printf("%s\t", Data[i]);
                    }
                    
                    printf("\n");
                    rc2 = SQLFetch(hstmt);
                }
			
                SQLFreeStmt(hstmt, SQL_RESET_PARAMS);
                SQLFreeStmt(hstmt, SQL_UNBIND);
                SQLFreeStmt(hstmt, SQL_CLOSE);
            }
        } else {
            print_error(rc, SQL_HANDLE_STMT, hstmt, __FILE__, __LINE__);
        }
        
        return(rc);
}



/*##************************************************************************
 *
 *                  execute_sql_and_return_one_string()
 *
 *          This is a specialized version of the 
 *          execute_sql() routine.  This assumes that 
 *          the query returns a single row, and that 
 *          row has 2 columns, the second of which is 
 *          a string.  We then return that string.
 * 
 * Parameters
 *     hstmt - Statement Handle
 *     sql   - query to be executed
 *     cbsql - Length of query OR the special value SQL_NTS
 *             (Null-Terminated String).
 * 
 * Output
 *     This copies a new string to the parameter return_string_ptr.
 * 
 * Returns
 *      This returns the error code that we get when we execute the
 *      SQL statement that was passed to this routine.  Note that
 *      if that statement executes correctly, but there is an error during
 *      some subsequent step (such as printing column names or fetching
 *      the actual data rows), you will NOT get any indication of that
 *      subsequent error.  
 *
 */
int execute_sql_and_return_one_string(SQLHSTMT hstmt, SQLCHAR *sql, 
                                      SQLINTEGER cbsql,
                                      SQLCHAR *return_string_ptr,
                                      unsigned int max_len)
{
        /* The Return Code when we execute the statement. */
        SQLRETURN       rc;
        /* The Return Code from various functions involved in printing
         * column names, etc.
         */
        SQLRETURN       rc1;
        /* The Return Code when we fetch the next row of data. */
        SQLRETURN       rc2;
        /* The number of columns returned by the query. */
        SQLSMALLINT     ColCount = 0;
        SQLCHAR         ColName[MAX_COL_NAME_LEN + 1] = { 0 };
        SQLLEN          Len[MAXCOLS] = { 0 };
        SQLCHAR         Data[MAXCOLS][DATALEN] = { 0 };
        SQLSMALLINT     i = 0;
        SQLINTEGER      RowCount = 0;
        SQLSMALLINT     iCol = 0;

        /* Check that the input string is not NULL. */	
        assert(sql != NULL);

        /* Display the command we were asked to execute. */
        printf("----> %s\n", sql); 

        /* Execute the command. */
        rc = SQLExecDirect(hstmt, sql, cbsql);
        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {
            rc1 = SQLNumResultCols(hstmt, &ColCount);
            assert(ColCount == 2);
            if (ColCount > 0)  {
                /* Print the column names. */
                for (i = 0; i < ColCount; i++)  {
                    iCol = i + 1;
                    rc1 = SQLDescribeCol(hstmt, iCol, ColName, 
                                         sizeof(ColName),
                                         NULL, NULL, NULL, NULL, NULL);
                    printf("%s\t", ColName);
                    rc1 = SQLBindCol(hstmt, iCol, SQL_C_CHAR,
                                     Data[iCol-1],
                                     sizeof(Data[iCol-1]),
                                     &Len[iCol-1]);
                }
                printf("\n");
            }
            rc2 = SQLFetch(hstmt);
            /* Display data for each row.  The caller is expecting only 
             * one row, but we will display (although not return) data for
             * all the rows that match.
             */
            while (rc2 != SQL_NO_DATA_FOUND && rc2 != SQL_ERROR)  {
                RowCount++;
                for (i = 0; i < ColCount; i++) {
                    if (Len[i] == SQL_NULL_DATA) {
                        strcpy(Data[i], "NULL");
                    }
                    printf("%s\t", Data[i]);
                }
                printf("\n");
                rc2 = SQLFetch(hstmt);
            }
            
            /* Save a copy of the data that we called this routine to get! */
            if (Data[1] != NULL
                && (strlen(Data[1]) <= max_len)) {
                
                strcpy(return_string_ptr, Data[1]);
            }
            printf("\n");
            
            SQLFreeStmt(hstmt, SQL_RESET_PARAMS);
            SQLFreeStmt(hstmt, SQL_UNBIND);
            SQLFreeStmt(hstmt, SQL_CLOSE);
        }
        else  {
            print_error(rc, SQL_HANDLE_STMT, hstmt, __FILE__, __LINE__);
        }
        return(rc);
}

/* -------------------------------------------------------------------------
 * Function Name	: Disconnect()
 * Purpose		: to free the handles
 * Parameters		:
 * 	henv		: Environment handle
 * 	hdbc		: Connection Handle
 * 	hstmt		: Statement Handle
 * Returns		: Nothing.
 * -------------------------------------------------------------------------
 */ 
void Disconnect(SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt)
{
        SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
        SQLDisconnect(hdbc);
        SQLFreeHandle(SQL_HANDLE_DBC, hdbc);		
        SQLFreeHandle(SQL_HANDLE_ENV, henv);
}



/*##***********************************************************************
 *
 *                       acc_start_server()
 * 
 *       to start the server
 * 
 * Parameters		 :
 *         testcase_name : name of this program.
 *
 * Returns		 : nothing.
 * 
 */
void acc_start_server(char *testcase_name)
{
        /* Number of arguments (pgm name, catalog, user, password) */
        int g_argc = 4;
        /* This will hold the arguments. */
        char** g_argv = NULL;
        /* return code from function calls */
        int rc = 0;
        /* Temporary space when composing strings. 2 chars for "-U" +
         * 80 chars for user name + 1 char for string terminator.
         */
        char l_username[2 + 80 + 1] = "";
        char l_password[2 + 80 + 1] = "";

        /* Allocate enough string pointers for the params we'll pass. */
        g_argv = (char **) malloc(g_argc * sizeof(char *));
        assert(g_argv != NULL);

        g_argv[0] = testcase_name;

        /* catalog name (which we've hard-coded to "dba") */
        g_argv[1] = "-Cdba";

        /* user name */
        strcpy(l_username, "-U");
        strcat(l_username, username);
        g_argv[2] = l_username;

		/* password */
        strcpy( (char *) l_password, "-P");
        strcat(l_password, password);
        g_argv[3] = l_password;

        rc = SSCStartDisklessServer(g_argc, g_argv, &serverHandle,
                                    SSC_STATE_OPEN,
                                    lic_string, ini_string);
        assert(rc == SSC_SUCCESS);
        
		if (g_argv != NULL)  {
            free(g_argv);
        }        
}



/* -------------------------------------------------------------------------
* Function Name	: acc_stop_server()
* Purpose       : to stop the server
* Parameters    : None
* Returns       : Nothing
* --------------------------------------------------------------------------
*/
void acc_stop_server(void)
{
        SSCStopServer(serverHandle, TRUE);
}


/*##***********************************************************************\
 *
 *               main
 *
 */

#ifdef main
#  undef main
#endif

#if defined(SS_NT)
#  ifndef SS_CDECL
#    define SS_CDECL __cdecl
#  endif
#elif defined(SS_VXW)
#  define SS_CDECL
#  define main embed_main
#else
#  define SS_CDECL
#endif

int SS_CDECL main(int argc, char * argv[])
{
        /* Points to a string that contains an SQL command, e.g.
         * "SELECT col1 FROM table1" */
        char *command_ptr;
        /* This holds the Return Code from some functions. */
        int rc = 0;
        /* This holds a string that is returned from a query. */
#define MAX_STR_LEN 80
        char last_name[MAX_STR_LEN + 1];

        /* Start the server using SSCStartServer() */
        puts("Starting the server...");
        acc_start_server(argv[0]);
        puts("\n\n========================================================");

        /* Connect to the server (note that the application that starts an 
         * accelerator server does not automatically get a connection to that 
		 * server).  
         */
        puts("\nAttempting to connect to the server.");
        PrepareForConnect(&henvm, &hdbcm, &hstmtm);
        rc = Connect(&henvm, &hdbcm, &hstmtm, listenat, username, password);

        command_ptr = "CREATE TABLE employees (id INT, last_name CHAR(40), first_name CHAR(40))";
        execute_sql(hstmtm, command_ptr, SQL_NTS);

        command_ptr = "COMMIT WORK";
        execute_sql(hstmtm, command_ptr, SQL_NTS);

        command_ptr = "INSERT INTO employees (id, last_name, first_name) VALUES (1, 'Tan', 'Amy')";
        execute_sql(hstmtm, command_ptr, SQL_NTS);

        command_ptr = "INSERT INTO employees (id, last_name, first_name) VALUES (2, 'Manchester', 'William')";
        execute_sql(hstmtm, command_ptr, SQL_NTS);

        command_ptr = "COMMIT WORK";
        execute_sql(hstmtm, command_ptr, SQL_NTS);
        
        command_ptr = "SELECT id, last_name FROM employees WHERE id = 2";
        puts(command_ptr);
        execute_sql_and_return_one_string(hstmtm, command_ptr, SQL_NTS, 
                                          last_name,
                                          (unsigned int) MAX_STR_LEN);
        puts(last_name);
        if (strcmp(last_name, "Manchester") != 0)   {
            puts("ERROR: Did not retrieve the expected last_name!");
        }
        
        /* Disconnect from the server */
        puts("Disconnecting from the server...");
        Disconnect(henvm, hdbcm, hstmtm);
		
        /* Shut down the server. */
        puts("Shutting down the server...");
        acc_stop_server();
        
        return 0;      
}

/* EOF */
