import java.sql.*;
import java.util.Random;

// import the Java Accelerator classes (these are stored
// inside the SolidDriver jar-file)
import solid.ssc.*;

//
// SJASample UserThread class
//
class UserThread extends Thread {

        //
        // The user thread needs to access SolidServerControl
        // for registering/unregistering itself with the Solid
        // Accelerator server
        // 
        private SolidServerControl myServer = null;

        
        public UserThread( SolidServerControl ssc ){
                myServer = ssc;
        }

        
        public void run(){
                
                if (myServer != null){

                    Random rnd        = new Random();
                    String sql        = null;
                    Connection myConn = null;
                    Statement myStmt  = null;

                    //
                    // This is the first thing you must do inside the user
                    // thread communicating with the Solid Accelerator Server.
                    // You must register this thread explicitly !
                    //
                    long retval = myServer.registerThread();
                    myServer.getOutStream().println("SJASample::UserThread: "+
                                                    " call to SolidServerControl:registerThread()"+
                                                    " returned 0x"+Long.toHexString(retval));

                    try {
                        
                        //
                        // Init a new JDBC connection for this thread
                        //
                        myConn = DriverManager.getConnection(
                                "jdbc:solid://localserver:0",
                                "dba",
                                "dba" );

                        myConn.setAutoCommit( false );
                        
                        // Insert 10..30 lines into table SJATEST
                        int MAX_REPEATS = 10*(1+rnd.nextInt(2));
                        String thrname = Thread.currentThread().getName();
                        
                        myServer.getOutStream().println("\nSJASample::UserThread('"+
                                                        thrname +
                                                        "'): inserting "+ 
                                                        MAX_REPEATS +
                                                        " lines into table SJATEST");
                        
                        for (int i=0; i < MAX_REPEATS; i++){
                            
                            myStmt = myConn.createStatement();
                            sql = "insert into SJATEST values('"+
                                thrname + ", message number " + i + "')";
                            myStmt.executeUpdate( sql );
                            myStmt.close();
                        }

                        //
                        // Commit the work and close the connection
                        //
                        myConn.commit();
                        myConn.close();
                            
                    } catch (Exception e){
                        myServer.getOutStream().println("SJASample::UserThread: Exception caught while "+
                                           "processing JDBC statements.");
                        e.printStackTrace();
                    }

                    //
                    // Before you let the user thread communicating with Solid Accelerator
                    // server die, you must unregister this (registered) thread explicitly !
                    //
                    retval = myServer.unregisterThread();
                    myServer.getOutStream().println("SJASample::UserThread: "+
                                                    " call to SolidServerControl:unregisterThread()"+
                                                    " returned 0x"+Long.toHexString(retval));
                    
                } else {
                    System.out.println("SJASample::UserThread: myServer is null!");
                }
        }
        
}

//
// SJASample main class
//
public class SJASample {

        public static void main( String args[] ){

                // SolidServerControl class instance holds
                // a reference to the Solid database server
                SolidServerControl ssc = null;

                // SolidDriver contains Solid JDBC driver
                String soliddrivername = "solid.jdbc.SolidDriver";

                // URL for the JDBC accelerator server is given as follows
                String url = "jdbc:solid://localserver:0";
                
                String login = "dba";
                String pwd = "dba";
                long rc;

                // Parameters to be passed to the startServer method
                // -c <working directory>
                //    ! solid license file must be in the directory pointed by
                //      this argument!
                // -U <username>
                // -P <password>
                // -C <catalog name>
                String av[] = {
                    "-c.",
                    "-Udba",
                    "-Pdba",
                    "-Cdba"
                };
                    
                Driver dri = null;
                Connection conn = null;
                Statement stmt = null;
                ResultSet rset = null;

                // - Instantiate a SolidServerControl object
                // - Start the Solid server with argument vector av
                //   and SSC_STATE_OPEN
                try {

                    ssc = SolidServerControl.instance( System.out );
                    rc=ssc.startServer( av, SolidServerControl.SSC_STATE_OPEN );
                    if(SolidServerControl.SSC_SUCCESS != rc) {
                       ssc.getOutStream().println("Could not start server rc:" + rc);
                       System.exit(0);
                    }

                    ssc.getOutStream().println("\n\n");
                    ssc.getOutStream().println("***************************************************");
                    ssc.getOutStream().println(" This is Solid Java Accelerator Server Sample 0.1a");
                    ssc.getOutStream().println("  If you are able to see this text, it means that");
                    ssc.getOutStream().println("    you have been successful in starting the");
                    ssc.getOutStream().println("            Solid Accelerator Server.");
                    ssc.getOutStream().println("***************************************************");
                    ssc.getOutStream().println("\n\n");

                } catch ( SolidServerInitializationError ssie ){
                    ssc.getOutStream().println("Could not start server:"+ssie.toString());
                    System.exit(0);
                }

                //
                // Now we are ready to operate with JDBC 
                //
                
                try {
                    int i=0;

                    //
                    // Initiate the JDBC driver and start a new connection
                    // Note, that in the main thread (this is the only user thread
                    // the moment, you don't need to (mustn't) register the thread with
                    // ssc.registerThread() method call to the Solid Accelerator
                    // server !
                    //
                    
                    dri = (Driver)Class.forName(soliddrivername).newInstance();
                    conn = DriverManager.getConnection(url, login, pwd);
                    conn.setAutoCommit( false );

                    //
                    // Create table SJATEST
                    //
                    ssc.getOutStream().println("SJASample: Creating table"+
                                                    " SJATEST( MESSAGE VARCHAR )");
                    stmt = conn.createStatement();
                    stmt.executeUpdate("CREATE TABLE SJATEST( MESSAGE VARCHAR )");
                    stmt.close();
                    conn.commit();


                    //
                    // Create two new user threads for inserting data into the
                    // SJATEST table
                    //
                    
                    UserThread producer = new UserThread( ssc );
                    UserThread consumer = new UserThread( ssc );
                    
                    producer.start();
                    consumer.start();

                    while (producer.isAlive() || consumer.isAlive() ){}


                    //
                    // Print the contents of table SJATEST
                    // 
                    ssc.getOutStream().println("SJASample: Printing table"+
                                                    " SJATEST contents");
                    stmt = conn.createStatement();
                    rset = stmt.executeQuery("SELECT * FROM SJATEST");
                    
                    ssc.getOutStream().println("\nThe contents of table SJATEST\n");
                    ssc.getOutStream().println("\n********************************************");
                    while( rset.next() ){
                        i++;
                        ssc.getOutStream().println("Row "+i+":'"+rset.getString(1)+"'");
                    }
                    ssc.getOutStream().println("********************************************\n");
                    stmt.close();
                    rset.close();

                    //
                    // Drop table SJATEST
                    //
                    ssc.getOutStream().println("SJASample: Dropping table SJATEST");
                    stmt = conn.createStatement();
                    stmt.executeUpdate("DROP TABLE SJATEST");
                    stmt.close();

                    //
                    // We're finished with this sample...
                    //
                    
                    // Close connection
                    conn.close();
                    
                } catch (Exception e){
                    ssc.getOutStream().println("Exception occurred: "+e.toString());
                }

                //
                // Stop the server.
                // 
                // If you omit calling stopServer, the server 
                // will not stop before JVM starts finalizing
                // SolidServerControl or the program ends. 
                // In these cases serverStop will be called implicitly
                // by the JVM.
                //
                
                if (ssc != null){
                    ssc.stopServer( ~SolidServerControl.SSC_STATE_OPEN|SolidServerControl.SSC_STATE_OPEN );
                }
                
        }
}


// Copyright (c) 2002, Solid Information Technology.  All Rights Reserved.
