/*************************************************************************\
**  source       * clusterdbclient.c
**  directory    * 
**  description  * Sample 'client' program for solidDB Cluster; testing  
**               * lookup and update speed of solidDB.
**               * 
**               * Note that this sample's purpose is to demonstrate the load 
**               * balancing functionality built into the Solid ODBC driver & 
**               * DB server (and not any failover functionality).
**               * 
**               * (C) Copyright Solid Information Technology Ltd 1992-2007
\*************************************************************************/

/* Common system headers */
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>

#include <math.h>  /* random number calculation */

/* windows specific stuff */
#ifdef WIN32
#  include <windows.h>
#  include <winbase.h>
#  include <process.h>
#  include <sys/timeb.h>  /* for time in milliseconds */
#  define MAIN_RETURN_TYPE    __cdecl
#  define START_TIMING()  QueryPerformanceCounter(&counter1);
#  define END_TIMING()    QueryPerformanceCounter(&counter2); 
#  define CALC_DURATION() tx_duration = (double)(counter2.LowPart \
                               - counter1.LowPart) / (double)freq.LowPart;
#  define ss_sleep(t)     Sleep((t)*1000)

#  ifdef _UINTPTR_T_DEFINED
typedef uintptr_t thrhandle_t;
#  else
typedef unsigned long thrhandle_t;  /* uintptr_t is not a long, but 
                                     * _beginthread returns one. */
#  endif

#endif

#ifdef SS_UNIX
#  include <unistd.h>
#  include <sys/time.h>
#  include <sys/timeb.h>
#  include <pthread.h>

struct timeval starttime;
struct timeval endtime;

#  define MAIN_RETURN_TYPE int SS_CDECL
#  define start_timems    ((long double)starttime.tv_sec*1000.0 \
                            + (long double)starttime.tv_usec/1000.0)
#  define end_timems      ((long double)endtime.tv_sec*1000.0 \
                            + (long double)endtime.tv_usec/1000.0)
#  define START_TIMING()  gettimeofday(&starttime, 0); 
#  define END_TIMING()    gettimeofday(&endtime, 0);
#  define CALC_DURATION() tx_duration = (double)(end_timems - start_timems)/1000.0;
#  define ss_sleep sleep

typedef pthread_t thrhandle_t;

#endif

#include <solidodbc3.h>

#include "c.h" 

/* Utility macros - very crude error handling */
#define CHECK(r)  \
        if ( ((r) != SQL_SUCCESS) && ((r) != SQL_SUCCESS_WITH_INFO)    \
              && ((r) != SQL_NO_DATA) )                                \
            print_error((r), henv, hdbc, hstmt);                       \
        assert( ((r) == SQL_SUCCESS) || ((r) == SQL_SUCCESS_WITH_INFO) \
                || ((r) == SQL_NO_DATA) )
#define CHECK_STR(s)  assert((s)[sizeof(s)-1] == '\0' || !"buffer too short")
#define SU(_r)        assert(_r == SQL_SUCCESS)
#define SUH(_r, _h)   if (_r != SQL_SUCCESS) {outerror(_h); assert(0);}

typedef struct server_info_st {
        SQLHENV  si_henv;
        SQLHDBC  si_hdbc;
        SQLCHAR* si_connstr;
} server_info_t;

/* generic database variables */
SQLHENV   henv  = (SQLHENV)NULL;
SQLHDBC   hdbc  = (SQLHDBC)NULL;
SQLHSTMT  hstmt = (SQLHSTMT)NULL;

SQLCHAR connstr[300];
SQLCHAR connstr1[100];
SQLCHAR connstr2[100];
UCHAR* uid = "dba";
UCHAR* pwd = "dba";

#define SQLSTATE_LEN    6
#define PMON_INTERVAL_S 5
#define MAX_NAME_LEN    50
#define MAX_STR_LEN     128
#define THREADS_MAX     50

static int   nthreads = 1;
char         cmdbuf[2048];

int          debugp  = 0; /* should we output some debug information */
int          nopause = 0; /* don't ask for keypresses to continue */
volatile int pmon_running = 0;

typedef struct tagTXSTATS {
        int    thr_nr;
        long   rowcount;
        long   testrounds;
        long   readtxcount;
        long   updatetxcount;
        double readtotaltime;
        double updatetotaltime;
} TXSTATS, *PTXSTATS;

static void connect_to(server_info_t* sip);
static void disconnect(server_info_t* sip);
static int retrieve_results(
        SQLHSTMT*       hstmt_ptr,
        char*           retcode,
        char*           text);
int execute_single_statement(
        SQLHSTMT*       hstmt_ptr,
        char*           statement,
        char*           retcode,
        char*           text);

/*#***********************************************************************\
 *
 *      outerror
 *
 * Print out error message
 *
 * Parameters:
 *      hstmt - in, use
 *          statement handle with the error
 *
 */
static void outerror(SQLHSTMT hstmt)
{
        SQLCHAR     szSqlState[6];
        SQLCHAR     szErrorMsg[256];
        SQLINTEGER  naterr;
        SQLSMALLINT length;
        UCHAR* stripped;
        int    i;

        SQLError(SQL_NULL_HENV, SQL_NULL_HDBC, hstmt, szSqlState,
                 &naterr, szErrorMsg, sizeof(szErrorMsg), &length);
        
        stripped = szErrorMsg;
        for (i = 0; i < 3; i++) {
            UCHAR* tmp;
            tmp = (UCHAR*)strchr((char*)stripped, ']');
            if (tmp != NULL) {
                stripped = tmp + 1;
            } else {
                break;
            }
        }
        
        printf("%s\n", stripped);
        if (debugp) {
            printf("naterr %d, sqlstate=%s\n", (int)naterr, szSqlState);
        }
}


/*#***********************************************************************\
 *
 *      print_error
 *
 * Print error message
 *
 * Parameters:
 *      rc - in
 *          return code containing the error
 *
 *      henv - in, use
 *          related henv
 *
 *      hdbc - in, use
 *          related hdbc
 *
 *      hstmt - in, use
 *          related hstmt
 *
 */
static void print_error(RETCODE rc, SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt)
{
        UCHAR   SQLstate[MAX_NAME_LEN];
        SDWORD  errorcode;
        UCHAR   errormsg[SQL_MAX_MESSAGE_LENGTH];
        SWORD   msglen;
        RETCODE r;

        printf("Error (%d) encountered!\n", rc);

        /* Print more error info, if available */
        if (henv || hdbc || hstmt) {
            r = SQLError(henv, hdbc, hstmt, SQLstate, &errorcode, errormsg,
                         SQL_MAX_MESSAGE_LENGTH - 1, &msglen);
            if (r != SQL_NO_DATA_FOUND) {
                printf("%s\n", errormsg);
            }
        }
}

/*#***********************************************************************\
 *
 *      hstmt_success
 *
 * Checks if error code from statement functions is success. If not,
 * prints error message.
 *
 * Parameters:
 *      hstmt - in, use
 *          statement to examine
 *
 *      rc - in
 *          return code of the statement
 *
 * Return value:
 *      1, if was successful
 *      0, if not.
 *
 */
static int hstmt_success(
        SQLHSTMT    hstmt,
        SQLRETURN   rc)
{
        UCHAR   szSqlState[SQLSTATE_LEN];
        UCHAR   szErrorMsg[255];
        SDWORD  naterr;
        SWORD   length;

        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {
            return 1;
        }

        rc = SQLError(SQL_NULL_HENV, SQL_NULL_HDBC, hstmt, szSqlState,
                      &naterr, szErrorMsg, sizeof(szErrorMsg), &length);
        printf("Error %s: %s\n", szSqlState, szErrorMsg);

        return 0;
}

/*##**********************************************************************\
 *
 *      myRand
 *
 * Return random number in range [x,y].
 *
 * Parameters:
 *      x - in
 *          minimum
 *
 *      y - in
 *          maximum
 *
 * Return value:
 *      a random number
 *
 */
int myRand(int x, int y)
{
        /* the basic no-good RNG */
        return x + (int) (y * (rand() / (RAND_MAX + 1.0)));
}

/*##**********************************************************************\
 *
 *      randInit
 *
 * Initialize random number generator
 *
 */
void randInit(void)
{
        srand((unsigned int) time(NULL));        
}


/*#***********************************************************************\
 *
 *      create_tables
 *
 * Create tables for the sample
 *
 * Parameters:
 *      tablestyle - in
 *          1 = disk tables, 2 = in-memory tables
 *
 */
static void create_tables(int tablestyle)
{
        SQLHSTMT  hstmt = (SQLHSTMT)NULL;
        RETCODE   rc;
        
        char* droptable1 = "drop table subscriber";
        char* creatememorytable1 = "\
            create table subscriber ( \
                s_id integer not null, \
                s_# varchar(15) not null, \
                s_data1 varchar(100) not null, \
                s_data2 integer, \
                s_data3 integer, \
                primary key (s_id)  \
            ) store memory";

        char* createdisktable1 = "\
            create table subscriber ( \
                s_id integer not null, \
                s_# varchar(15) not null, \
                s_data1 varchar(100) not null, \
                s_data2 integer, \
                s_data3 integer, \
                primary key (s_id)  \
            ) store disk";

        char* createtable1index = "\
            create unique index s_idx1 on subscriber (s_#)";


        /* Allocate a statement handle. */
        
        if (debugp) {
            printf("SQLAllocStmt 0x%lx '0x%lx' ...\n", (unsigned long)hdbc,
                   (unsigned long)&hstmt);
        }
        rc = SQLAllocStmt(hdbc, &hstmt);
        CHECK(rc);

            
        if (debugp) {
            printf("Dropping tables...\n");
        }

        rc = SQLExecDirect(hstmt, droptable1, SQL_NTS);
        rc = SQLTransact(henv, hdbc, SQL_COMMIT);
        CHECK(rc);

        if (debugp) {
            printf("Create tables...\n");
        }

        if (tablestyle == 2) {
            strcpy(cmdbuf, creatememorytable1);
        } else {
            strcpy(cmdbuf, createdisktable1);
        }
        
        if (debugp) {
            printf("\ntable1: %s\n", cmdbuf);
        }
        rc = SQLExecDirect(hstmt, cmdbuf, SQL_NTS);
        CHECK(rc);
        rc = SQLTransact(henv, hdbc, SQL_COMMIT);
        CHECK(rc);

        if (debugp) {
            printf("\ntable1index: %s\n", createtable1index);
        }
        rc = SQLExecDirect(hstmt, createtable1index, SQL_NTS);
        CHECK(rc);
        rc = SQLTransact(henv, hdbc, SQL_COMMIT);
        CHECK(rc);
        
        /* Release statement resource allocation */
        rc = SQLFreeStmt(hstmt, SQL_DROP);
        CHECK(rc);
        
        return;
}

/*#***********************************************************************\
 *
 *      populate_subscriber_odbc
 *
 * Populate tables
 *
 * Parameters:
 *      rownum - in
 *          number of rows to insert
 *
 *      commitblock - in
 *          number of rows per commit
 *
 */
static void populate_subscriber_odbc(long rownum, long commitblock) 
{

        SQLHSTMT hstmt = (SQLHSTMT)NULL;
        RETCODE  r;
        RETCODE  rc;
        int      i;
        double   tx_duration = 0.0;
        
        /* subsciber table variables */
        SQLINTEGER s_id;
        SQLCHAR    s_number[16];
        SQLCHAR    s_data1[100];
        SQLINTEGER s_data2=0;
        SQLINTEGER s_data3=0;
        
#ifdef WIN32
        LARGE_INTEGER counter1;
        LARGE_INTEGER counter2;
        LARGE_INTEGER freq;
#endif
#ifdef SS_UNIX
        struct timeval starttime;
        struct timeval endtime;
#endif
        
        /* Allocate a statement handle. */
        rc = SQLAllocStmt(hdbc, &hstmt);
        CHECK(rc);
        
        sprintf(cmdbuf,
                "INSERT INTO subscriber VALUES (?,?,?,?,?)");
        
        r = SQLPrepare(hstmt,cmdbuf,SQL_NTS);
        SUH(r, hstmt);
        
        rc = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, SQL_C_LONG,
                              SQL_INTEGER, 0, 0, &s_id, 0, NULL);
        CHECK(rc);
        rc = SQLBindParameter(hstmt, 2, SQL_PARAM_INPUT, SQL_C_CHAR,
                              SQL_VARCHAR, MAX_STR_LEN, 0, s_number, 0, NULL);
        CHECK(rc);
        rc = SQLBindParameter(hstmt, 3, SQL_PARAM_INPUT, SQL_C_CHAR,
                              SQL_VARCHAR, MAX_STR_LEN, 0, s_data1, 0, NULL);
        CHECK(rc);
        rc = SQLBindParameter(hstmt, 4, SQL_PARAM_INPUT, SQL_C_SLONG,
                              SQL_INTEGER, 0, 0, &s_data2, 0, NULL);
        CHECK(rc);
        rc = SQLBindParameter(hstmt, 5, SQL_PARAM_INPUT, SQL_C_SLONG,
                              SQL_INTEGER, 0, 0, &s_data3, 0, NULL);
        CHECK(rc);
        
#ifdef SS_UNIX
        gettimeofday(&starttime, 0);
#endif
#ifdef WIN32
        QueryPerformanceFrequency(&freq);
#endif
        START_TIMING(); 
        
        /* fill in s_data1 column with (fixed/'arbitrary') data */
        memset(s_data1, 'A', 99);
        s_data1[99] = 0;
        
        for (i = 0; i < rownum; i++) {
            
            s_id    = (SQLINTEGER) i;
            s_data2 = (SQLINTEGER) i;
            s_data3 = (SQLINTEGER) i;
            sprintf(s_number, "%d", i);
            
            r = SQLExecute(hstmt);
            SUH(r, hstmt);
            
            if (i != 0 && i % commitblock == 0) {
                printf("\r%d", i);
                r = SQLTransact(henv, hdbc, SQL_COMMIT);
                SU(r);
            }
        }
        
        r = SQLTransact(henv, hdbc, SQL_COMMIT);
        SU(r);
        
        END_TIMING(); 
        CALC_DURATION();
        printf("\rInserted %ld rows in %f s.  %1.1f rows/s\n",
               rownum, tx_duration, rownum / tx_duration);
        
        rc = SQLFreeStmt(hstmt, SQL_RESET_PARAMS);
        CHECK(rc);
        r = SQLFreeStmt(hstmt, SQL_DROP);
        
}

/*##**********************************************************************\
 *
 *      run_test_set
 *
 * Run the test set transactions
 *
 * Parameters:
 *      pData - in, use
 *          test set stats
 *
 * Return value - out:
 *      nothing or NULL.
 *
 */
#ifdef WIN32
void __cdecl run_test_set(PTXSTATS pData)
#endif
#ifdef SS_UNIX
void* run_test_set(void* ptr)
#endif
{
#ifdef SS_UNIX
        PTXSTATS pData = ptr;
#endif
        
        int  thread     = pData->thr_nr;
        long rownum     = pData->rowcount;
        long testRounds = pData->testrounds;
        RETCODE rc;
        RETCODE r;
        long testTx = 0;
        long i      = 0;
        long rNum   = 0;
        long rNum2  = 0;
        
        /* database connections */
        SQLHENV    henv = (SQLHENV)NULL;
        SQLHDBC    hdbc = (SQLHDBC)NULL;
        
        /* statement handles */
        SQLHSTMT   hstmt_readtx = (SQLHSTMT)NULL;
        SQLHSTMT   hstmt_updatetx = (SQLHSTMT)NULL;
        
        /* subsciber table variables */
        SQLINTEGER s_id;
        SQLCHAR    s_number[16];
        SQLCHAR    s_data1[100];
        SQLINTEGER s_data2=0;
        SQLINTEGER s_data3=0;
        
        /* variables used for ODBC query execution */
        SQLINTEGER search_id;
        SQLINTEGER location;
        
        /* transaction result counters */
        double tx_duration=0;
#ifdef WIN32
        LARGE_INTEGER   counter1;
        LARGE_INTEGER   counter2;
        LARGE_INTEGER   freq;
#endif
#ifdef SS_UNIX
        struct timeval  starttime;
        struct timeval  endtime;
#endif
        
        printf("Test thread %d starting ...\n", thread);
        
        /*
         * get database connections
         */
        rc = SQLAllocEnv(&henv);
        CHECK(rc);
        rc = SQLAllocConnect(henv, &hdbc);
        CHECK(rc);
        rc = SQLConnect(hdbc, connstr, SQL_NTS, uid, SQL_NTS, pwd, SQL_NTS);
        CHECK(rc);
        rc = SQLSetConnectAttr(hdbc, SQL_ATTR_AUTOCOMMIT,
                               SQL_AUTOCOMMIT_OFF, 0);
        CHECK(rc);
        
        /*
         * initialize test result counters
         */
        pData->readtxcount     = 0;
        pData->updatetxcount   = 0;
        pData->readtotaltime   = 0.0;
        pData->updatetotaltime = 0.0;
        
        /*
         * prepare test SQL statements
         */
        
        /* read statement */
        
        rc = SQLAllocStmt(hdbc, &hstmt_readtx);
        CHECK(rc);
        
        r = SQLPrepare(hstmt_readtx,
                       "SELECT * FROM subscriber WHERE s_id = ?",
                       SQL_NTS);
        SUH(r, hstmt_readtx);
        
        rc = SQLBindParameter(hstmt_readtx, 1, SQL_PARAM_INPUT, SQL_C_LONG,
                              SQL_INTEGER, 0, 0, &search_id, 0, NULL);
        CHECK(rc);
        r = SQLBindCol(hstmt_readtx, 1, SQL_C_LONG, &s_id,
                       sizeof(s_id), NULL);
        SU(r);
        r = SQLBindCol(hstmt_readtx, 2, SQL_C_CHAR, s_number,
                       sizeof(s_number), NULL);
        SU(r);
        r = SQLBindCol(hstmt_readtx, 3, SQL_C_CHAR, s_data1,
                       sizeof(s_data1), NULL);
        SU(r);
        r = SQLBindCol(hstmt_readtx, 4, SQL_C_LONG, &s_data2,
                       sizeof(s_data2), NULL);
        SU(r);
        r = SQLBindCol(hstmt_readtx, 5, SQL_C_LONG, &s_data3,
                       sizeof(s_data3), NULL);
        SU(r);
        
        /* update statement */
        
        rc = SQLAllocStmt(hdbc, &hstmt_updatetx);
        CHECK(rc);
        
        r = SQLPrepare(hstmt_updatetx,
                       "UPDATE subscriber SET s_data3 = ? WHERE s_# = ?",
                       SQL_NTS);
        SUH(r, hstmt_updatetx);
        
        rc = SQLBindParameter(hstmt_updatetx, 1, SQL_PARAM_INPUT, SQL_C_LONG,
                              SQL_INTEGER, 0, 0, &location, 0, NULL);
        CHECK(rc);
        rc = SQLBindParameter(hstmt_updatetx, 2, SQL_PARAM_INPUT, SQL_C_CHAR,
                              SQL_VARCHAR, MAX_NAME_LEN, 0, s_number, 0, NULL);
        CHECK(rc);
        
        /*
         * run the test
         */
        
#ifdef WIN32
        QueryPerformanceFrequency(&freq);
#endif
        
        for (i = 1; i <= testRounds; i++) {
            
            /* each time select which test case to run */
            testTx = myRand(1, 100);
            
            if (testTx <= 80) {
                /* target is 80% reads, 'get Subcscriber data' */
                rNum = myRand(0, rownum-1);
                START_TIMING();

                search_id = rNum;
                
                r = SQLExecute(hstmt_readtx);
                SUH(r, hstmt_readtx);
                
                r = SQLFetch(hstmt_readtx);
                CHECK(r);
                
                assert(s_id == search_id);
                
                r = SQLTransact(henv, hdbc, SQL_COMMIT);
                CHECK(r);
                
                rc = SQLFreeStmt(hstmt_readtx, SQL_CLOSE); 
                SUH(rc, hstmt_readtx);
                
                END_TIMING();
                CALC_DURATION();
                pData->readtxcount++;
                pData->readtotaltime = pData->readtotaltime + tx_duration;
                
            } else {
                /* target is 20% for writes, 'update Subscriber data */
                rNum = myRand(0, 2147483647);  /* 2 to power 32 - 1 */
                rNum2 = myRand(0, rownum-1);
                
                START_TIMING();
                location = (SQLINTEGER) rNum;
                sprintf(s_number, "%ld", rNum2);
                
                r = SQLExecute(hstmt_updatetx);
                SUH(r, hstmt_updatetx);
                
                r = SQLTransact(henv, hdbc, SQL_COMMIT);
                SU(r);
                
                rc = SQLFreeStmt(hstmt_updatetx, SQL_CLOSE);
                SUH(rc, hstmt_updatetx);
                
                END_TIMING();
                CALC_DURATION();
                pData->updatetxcount++;
                pData->updatetotaltime = pData->updatetotaltime + tx_duration;
                
            }
        }
        
        /*
         * drop statements
         */
        r = SQLFreeStmt(hstmt_readtx, SQL_DROP);
        r = SQLFreeStmt(hstmt_updatetx, SQL_DROP);
        
        /*
         * disconnect from db
         */
        r = SQLDisconnect(hdbc);
        SU(r);
        r = SQLFreeConnect(hdbc);
        SU(r);
        r = SQLFreeEnv(henv);
        SU(r);

#ifdef SS_UNIX
        return NULL;
#endif        
}

/*##**********************************************************************\
 *
 *      pmon
 *
 * Do a pmon query and print out the selected results
 *
 * Parameters:
 *      sip - in, use
 *          server info
 *
 *      desc - in
 *          optional text to print out, can be NULL.
 *
 */
void pmon(server_info_t* sip, char* desc)
{
        char* cmd_hsb_state = "ADMIN COMMAND 'hsb state'";
        char* cmd_pmon      = "ADMIN COMMAND 'pmon'";
        RETCODE rc;
        SQLHSTMT hstmt;
        int rval;
        char text[1024];
        int i;
        char* mstr;
        /* must be in the same order as in pmon output */
        static char* matches[] = { 
            "Time (sec)",
            "SQL execute",
            "DBE insert",
            "DBE update",
            "DBE fetch",
            NULL
        };
                                   
        rc = SQLAllocStmt(sip->si_hdbc, &hstmt);
        SUH(rc, hstmt);

        printf("-----------------------------------------------------------------------------\n");

        if (desc != NULL) {
            printf("%s - ", desc);
        }
        
        /* hsb state output */

        rval = execute_single_statement(&hstmt, cmd_hsb_state, NULL, text);
        assert(rval == 0);
        printf("%s\n", text);
        
        /* read pmon output */

        rc = SQLBindCol(hstmt, 2, SQL_C_CHAR, text, sizeof(text), NULL);
        SUH(rc, hstmt);
        
        rc = SQLExecDirect(hstmt, cmd_pmon, SQL_NTS);
        SUH(rc, hstmt);

        mstr = matches[0];
        i = 0;
        while (mstr != NULL) {
            rc = SQLFetch(hstmt);
            if (rc == SQL_NO_DATA) {
                break;
            }
            SUH(rc, hstmt);

            if (strstr(text, mstr) != NULL) {
                /* "reformat" output to fit 80 cols. We'll just skip a
                 * block since we can't seek ':' as Time row doesn't
                 * have it.
                 */
                printf("%-12s %s\n", mstr, text + 30);
                mstr = matches[++i];
            }
        }
        
        rc = SQLFreeStmt(hstmt, SQL_DROP);
        SUH(rc, hstmt);
}

/*##**********************************************************************\
 *
 *      run_pmon
 *
 * monitoring thread, runs pmon queries.
 *
 * Parameters:
 *      ptr - notused
 *
 * Return value - out:
 *      nothing or NULL.
 *
 */
#ifdef WIN32
void __cdecl run_pmon(void* ptr)
#endif
#ifdef SS_UNIX
void* run_pmon(void* ptr)
#endif
{
        server_info_t srv_a = {(SQLHENV)NULL, (SQLHDBC)NULL, connstr1 };
        server_info_t srv_b = {(SQLHENV)NULL, (SQLHDBC)NULL, connstr2 };
        (void) ptr; /* kill unused variable warning */
        
        connect_to(&srv_a);
        connect_to(&srv_b);

        while (pmon_running) {
            pmon(&srv_a, "Server A");
            pmon(&srv_b, "Server B");
            ss_sleep(PMON_INTERVAL_S);
            printf("\n\n");
        }
        disconnect(&srv_a);
        disconnect(&srv_b);
        
#ifdef SS_UNIX
        return NULL;
#endif
}

/*#***********************************************************************\
 *
 *      connect_to_db
 *
 * Connect to DB using the cluster connection string.
 *
 */
static void connect_to_db(void) 
{
        RETCODE rc;
        
        rc = SQLAllocEnv(&henv);
        CHECK(rc);
        rc = SQLAllocConnect(henv, &hdbc);
        CHECK(rc);
        printf("Solid connect string: %s\n", connstr);
        rc = SQLConnect(hdbc, connstr, SQL_NTS, uid, SQL_NTS,  pwd, SQL_NTS);
        CHECK(rc);
        rc = SQLSetConnectAttr(hdbc, SQL_ATTR_AUTOCOMMIT,
                               SQL_AUTOCOMMIT_OFF, 0);
        CHECK(rc);
}

/*#***********************************************************************\
 *
 *      connect_to
 *
 * Connects to server
 *
 * Parameters:
 *      sip - in, use
 *          server info of the desired server
 *
 */
static void connect_to(server_info_t* sip)
{
        RETCODE rc;
        SQLHENV henv = sip->si_henv;
        SQLHDBC hdbc = sip->si_hdbc;
        
        if (debugp) {
            printf("connect_to: %s\n", sip->si_connstr);
        }
        
        rc = SQLAllocEnv(&henv);
        CHECK(rc);
        rc = SQLAllocConnect(henv, &hdbc);
        CHECK(rc);
        rc = SQLConnect(hdbc, sip->si_connstr, SQL_NTS, uid, SQL_NTS,
                        pwd, SQL_NTS);
        CHECK(rc);
        rc = SQLSetConnectAttr(hdbc, SQL_ATTR_AUTOCOMMIT,
                               SQL_AUTOCOMMIT_OFF, 0);
        CHECK(rc);
        
        sip->si_henv = henv;
        sip->si_hdbc = hdbc;
}

/*##**********************************************************************\
 *
 *      do_netcopy
 *
 * Start netcopy from chosen server, must be connected.
 *
 * Parameters:
 *      sip - in, use
 *          server info
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
void do_netcopy(server_info_t* sip)
{
        char* cmd_netcopy     = "ADMIN COMMAND 'hsb netcopy'";
        char* cmd_status_copy = "ADMIN COMMAND 'hsb status copy'";
        RETCODE  rc;
        SQLHSTMT hstmt;
        int  rval;
        char retcode[1024];
        char text[1024];
        
        rc = SQLAllocStmt(sip->si_hdbc, &hstmt);
        SUH(rc, hstmt);
        rval = execute_single_statement(&hstmt, cmd_netcopy,
                                        retcode, text);
        if (rval) {
            printf("Error: %s: %s\n", retcode, text);
            assert(0);
        }

        while (1) {
            rval = execute_single_statement(&hstmt, cmd_status_copy,
                                            retcode, text);
            assert(!rval);
            rval = atoi(retcode);
            printf("netcopy: status copy: %s: %s\n", retcode, text);
            /* 14003: active
             * 0:     success
             */
            if (rval == 0) {
                break;
            }
            ss_sleep(2);
        }
        
        rc = SQLFreeStmt(hstmt, SQL_DROP);
        SUH(rc, hstmt);
}

/*##**********************************************************************\
 *
 *      do_hsb_connect
 *
 * Connect servers to have HA.
 *
 * Parameters:
 *      sip - in, use
 *          server info, must be connected already
 *
 */
void do_hsb_connect(server_info_t* sip)
{
        char* cmd_connect        = "ADMIN COMMAND 'hsb connect'";
        char* cmd_status_connect = "ADMIN COMMAND 'hsb status connect'";
        RETCODE rc;
        SQLHSTMT hstmt;
        int rval;
        char retcode[1024];
        char text[1024];
        
        rc = SQLAllocStmt(sip->si_hdbc, &hstmt);
        SUH(rc, hstmt);
        rval = execute_single_statement(&hstmt, cmd_connect,
                                        retcode, text);
        if (rval) {
            printf("Error: %s: %s\n", retcode, text);
            assert(0);
        }

        while (1) {
            rval = execute_single_statement(&hstmt, cmd_status_connect,
                                            retcode, text);
            assert(!rval);
            rval = atoi(retcode);
            printf("HSB connect: status connect: %s: %s\n", retcode, text);
            /* 14008: catchup
             * 0: connected
             */
            if (rval == 0) {
                break;
            }
            ss_sleep(2);
        }
        
        rc = SQLFreeStmt(hstmt, SQL_DROP);
        SUH(rc, hstmt);
}

/*##**********************************************************************\
 *
 *      execute_single_statement
 *
 * Name says it all.
 *
 * Parameters:
 *      hstmt_ptr - in, use
 *          hstmt
 *
 *      statement - in
 *          statement to execute
 *
 *      retcode - in, out, use
 *          buffer for return code, can be NULL.
 *
 *      text - in, out, use
 *          buffer for text output of the statement, can be NULL.
 *
 * retcode and text can't both be NULL.
 *
 * Return value:
 *      0, if ok
 *     -1, if failed
 *
 */
int execute_single_statement(
        SQLHSTMT*       hstmt_ptr,
        char*           statement,
        char*           retcode,
        char*           text)
{
        SQLRETURN   rc;
        
        /* Execute the SQL statement. */
        rc = SQLExecDirect(*hstmt_ptr, (SQLCHAR *) statement, SQL_NTS);
        if (rc == SQL_ERROR || rc == SQL_SUCCESS_WITH_INFO) { /* failed */
            retrieve_results(hstmt_ptr, retcode, text);
            return -1;
        }
                    
        if (rc == SQL_SUCCESS
            && retrieve_results(hstmt_ptr, retcode, text) == 0) {

            return 0;
        }

        return -1;
}

#define MAX_BUFLEN 1024
/*#***********************************************************************\
 *
 *      retrieve_results
 *
 * Retrieve results of a statement
 *
 * Parameters:
 *      hstmt_ptr - in, use
 *          hstmt
 *
 *      retcode - in, out, use
 *          buffer for return code, can be NULL.
 *
 *      text - in, out, use
 *          buffer for text output, can be NULL.
 *
 * retcode and text can't both be NULL.
 *
 * Return value:
 *      0, if ok
 *     -1, if failed
 *
 */
static int retrieve_results(
        SQLHSTMT*       hstmt_ptr,
        char*           retcode,
        char*           text)
{
        int             i;
        SQLUSMALLINT    colnum;
        SQLHSTMT        hstmt = *hstmt_ptr;
        SQLRETURN       rc = 0;
        int             ret = -1; /* return value */

        char    colname[32];
        SWORD   coltype;
        SWORD   colnamelen;
        SWORD   nullable;
        SQLULEN collen[2];
        SWORD   scale;
        SQLLEN  outlen[2];
        SWORD   nresultcols;

        char    RC_buf[MAX_BUFLEN];
        char    text_buf[MAX_BUFLEN];
        char*   RC_ptr;
        char*   text_ptr;

        RC_ptr      = (retcode == NULL) ? RC_buf: retcode;
        text_ptr    = (text == NULL)    ? text_buf: text;
        RC_ptr[0]   = '\0';
        text_ptr[0] = '\0';

        if (retcode == NULL && text == NULL) {
            /* there's no point in having both NULL */
            goto return_rr;
        }

        rc = SQLNumResultCols(hstmt, &nresultcols);
        if (!hstmt_success(hstmt, rc)) {
            goto return_rr;
        }

        if (nresultcols != 2) {
            goto return_rr;
        }

        for (i = 0; i < nresultcols; i++) {
            colnum = (SQLUSMALLINT) (i + 1);

            rc = SQLDescribeCol(hstmt, colnum, (SQLCHAR *) colname,
                                (SWORD) sizeof(colname), &colnamelen,
                                &coltype, &collen[i], &scale,
                                &nullable);
            if (!hstmt_success(hstmt, rc)) {
                goto return_rr;
            }

            if (strcmp(colname, "RC") == 0) {
                rc = SQLBindCol(hstmt, colnum, SQL_C_CHAR, RC_ptr,
                                collen[i], &outlen[i]);
                if (!hstmt_success(hstmt, rc)) {
                    goto return_rr;
                }
            } else if (strcmp(colname, "TEXT") == 0) {
                rc = SQLBindCol(hstmt, colnum, SQL_C_CHAR, text_ptr,
                                collen[i], &outlen[i]);
                if (!hstmt_success(hstmt, rc)) {
                    goto return_rr;
                }
            } else { /* SUCCESS */
                ret = 0;
                goto return_rr;
            }
        } /* end for */

        rc = SQLFetch(hstmt);
        
        if (!hstmt_success(hstmt, rc)) {
            goto return_rr;
        }

        ret = 0;
        
return_rr:
        return ret;
}

/*##**********************************************************************\
 *
 *      check_hsb_state
 *
 * Check that the server is ACTIVE in HSB, therefore, they are connected.
 *
 * Parameters:
 *      sip - in, use
 *          server info, must be connected.
 *
 * Return value:
 *      1, if connected.
 *      0, if not.
 *
 */
int check_hsb_state(server_info_t* sip)
{
        char* cmd_state = "ADMIN COMMAND 'hsb state'";
        RETCODE  rc;
        SQLHENV  henv = sip->si_henv;
        SQLHDBC  hdbc = sip->si_hdbc;
        SQLHSTMT hstmt = (SQLHSTMT)NULL;
        int  rval;
        char text[1024];
        int  res = 0;
        
        rc = SQLAllocStmt(hdbc, &hstmt);
        CHECK(rc);
        rval = execute_single_statement(&hstmt, cmd_state,
                                        NULL, text);
        assert(rval == 0);
        printf("HSB state: %s\n", text);
        if (strstr(text, "ACTIVE") != NULL) {
            res = 1;
        }
        
        rc = SQLFreeStmt(hstmt, SQL_DROP);
        SUH(rc, hstmt);
        sip->si_hdbc = hdbc;

        return res;
}

/*#***********************************************************************\
 *
 *      disconnect_db
 *
 * Disconnect from the clustered DB.
 *
 */
static void disconnect_db(void)
{
        RETCODE r;
        
        r = SQLDisconnect(hdbc);
        SU(r);
        r = SQLFreeConnect(hdbc);
        SU(r);
        r = SQLFreeEnv(henv);
        SU(r);
        hdbc = (SQLHDBC)NULL;
        henv = (SQLHENV)NULL;
}

/*#***********************************************************************\
 *
 *      disconnect
 *
 * Disconnect from a certain DB.
 *
 * Parameters:
 *      sip - in, use
 *          server info
 *
 */
static void disconnect(server_info_t* sip)
{
        RETCODE r;
        SQLHENV henv = sip->si_henv;
        SQLHDBC hdbc = sip->si_hdbc;

        r = SQLDisconnect(hdbc);
        SU(r);
        r = SQLFreeConnect(hdbc);
        SU(r);
        r = SQLFreeEnv(henv);
        SU(r);

        sip->si_henv = (SQLHENV)NULL;
        sip->si_hdbc = (SQLHDBC)NULL;
}

static void usage()
{
        printf("Usage: clusterDBclient \"connstr1\" \"connstr2\" [options]\n");
        printf("Options:\n");
        printf("    rowcount=n      - number of subscriber rows inserted (default 50000)\n");
        printf("    commitblock=n   - size of the data insert commitblock\n");
        printf("    threads=n       - number of client threads\n");
        printf("    populate        - create test table and populate test data\n");
        printf("    dbe             - use BE disk tables\n");
        printf("    mme             - use BE MME tables\n");
        printf("    no_lb           - no load balancing, use db1=connstr 1 db for all transactions\n");
        printf("    debug\n");
        printf("    nopause         - don't wait for keypresses\n");
        exit(1);
}

/*##**********************************************************************\
 *
 *      main
 *
 */
MAIN_RETURN_TYPE main(int argc, char* argv[])
{
        long rowcount    = 50000;
        int  commitblock = 2000;
        long testrounds  = 10000;
        int  populate    = 0;   /* should we create tables and
                                 * populate test data, default=no */
        int  tablestyle  = 2;   /* 1 = dbe, 2 = mme */
        int  noloadbalance = 0; /* use single=server1 server for all
                                 * transactions */
        long tmplong;
        int  i;

        long   totalTestRounds = 0;
        double totalTestTime = 0;
        char   userAnswer;
        SQLHSTMT hstmt = (SQLHSTMT)NULL;
        thrhandle_t t_rc[THREADS_MAX];
        TXSTATS ThreadData[THREADS_MAX];
        thrhandle_t t_pmon;
        
        server_info_t srv_a = {(SQLHENV)NULL, (SQLHDBC)NULL, connstr1 };
        server_info_t srv_b = {(SQLHENV)NULL, (SQLHDBC)NULL, connstr2 };
        RETCODE rc;
#ifdef SS_UNIX
        pthread_t t_id[THREADS_MAX];
        pthread_attr_t attr;
#endif
        long   readtxcount      = 0;
        long   updatetxcount    = 0;
        double readtotaltime    = 0;
        double updatetotaltime  = 0;        
        double txtotalpersecond = 0;
        
        /* - Get the connection strings to the databases.
         * - Add the load balancing connection string syntax.
         * NOTE: we default to dba dba as uname & pwd
         */

        if (argc<3) {
            usage();
        }
        strncpy(connstr1, argv[1], sizeof(connstr1));
        CHECK_STR(connstr1);        
        strncpy(connstr2, argv[2], sizeof(connstr2));
        CHECK_STR(connstr2);
        sprintf(connstr, "TF_LEVEL=SESSION PREFERRED_ACCESS=READ_MOSTLY SERVERS=%s,%s", connstr1, connstr2);

        /* process args */
        for (i = 1; i < argc; i++) {
            char* p;
            if ((p = strstr(argv[i], "rowcount=")) != NULL) {
                p += 9;
                tmplong = atol(p);
                if (tmplong > 0) {
                    rowcount = tmplong;
                }
            } else if ((p = strstr(argv[i], "commitblock=")) != NULL) {
                p += 12;
                tmplong = atol(p);
                if (tmplong > 0) {
                    commitblock = tmplong;
                }
            } else if ((p = strstr(argv[i], "threads=")) != NULL) {
                p += 8;
                tmplong = atol(p);
                if (tmplong > 0) {
                    nthreads = tmplong;
                }
            } else if ((p = strstr(argv[i], "testrounds=")) != NULL) {
                p += 11;
                tmplong = atol(p);
                if (tmplong > 0) {
                    testrounds = tmplong;
                }
            } else if (strcmp(argv[i], "dbe") == 0) {
                tablestyle = 1;
            } else if (strcmp(argv[i], "mme") == 0) {
                tablestyle = 2;
            } else if (strcmp(argv[i], "populate") == 0) {
                populate = 1;
            } else if (strcmp(argv[i], "no_lb") == 0) {
                noloadbalance = 1;
                strcpy(connstr,argv[1]);
            } else if (strcmp(argv[i], "debug") == 0) {
                debugp = 1;
            } else if (strcmp(argv[i], "nopause") == 0) {
                nopause = 1;
            } else if (strstr(argv[i], "help")) {
                usage();
            }
        }
        
        if (nthreads >= THREADS_MAX) {
            printf("Max amount of threads is %d, exiting\n", THREADS_MAX);
            exit(1);
        }
        
#ifdef SS_UNIX
        pthread_attr_init(&attr);
#  define STACKSIZE 50000
        pthread_attr_setstacksize(&attr, STACKSIZE);
#endif
        
        printf("===============\n");
        printf("  subscriber rowcount    = %ld\n", rowcount);
        printf("  client threads         = %d\n",  nthreads);
        printf("  testrounds per thread  = %ld\n", testrounds);
        if (populate) {
            printf("  insert commitblock     = %d\n", commitblock);
            if (tablestyle == 1) {
                printf("  DBE tables\n"); 
            } else {
                printf("  MME tables\n");
            }
        }
        printf("--\n");

        randInit();
        
        if (populate) {
            /* set up HSB */
            char  retcode[10];
            int   rval;
            char* cmd_switch_sec = "ADMIN COMMAND 'hsb switch secondary'";
            char* cmd_switch_pri = "ADMIN COMMAND 'hsb switch primary'";
            
            /* make B secondary just in case */
            connect_to(&srv_b);
            rc = SQLAllocStmt(srv_b.si_hdbc, &hstmt);
            SUH(rc, hstmt);
            rval = execute_single_statement(&hstmt, cmd_switch_sec,
                                            retcode, NULL);
            rc = SQLFreeStmt(hstmt, SQL_DROP);
            SUH(rc, hstmt);
            disconnect(&srv_b);

            connect_to(&srv_a);
            rc = SQLAllocStmt(srv_a.si_hdbc, &hstmt);
            SUH(rc, hstmt);
            rval = execute_single_statement(&hstmt, cmd_switch_pri,
                                            retcode, NULL);
            rc = SQLFreeStmt(hstmt, SQL_DROP);
            SUH(rc, hstmt);

            do_netcopy(&srv_a);
            do_hsb_connect(&srv_a);
            disconnect(&srv_a);
            
            connect_to_db();
            printf("create tables\n");
            create_tables(tablestyle);
            printf("populate subscriber data\n");
            populate_subscriber_odbc(rowcount, commitblock);
            disconnect_db();
        } else {
            printf("No populate, use existing db\n");
            printf("Solid connect string: %s\n", connstr);
        }
        
        if (noloadbalance) {
            printf("No load balancing\n");
        }

        if (!nopause) {
            userAnswer = 'n';
            while ((userAnswer != 'Y') && (userAnswer != 'y')) {
                int hsb_ok;
                int has_failed = 0;
                printf("\nMake sure the solidDB instances are in states:\n    Primary Active - Secondary Active.\n");

                do {
                    printf("Checking HSB state...\n");

                    connect_to(&srv_a);
                    hsb_ok = check_hsb_state(&srv_a);
                    if (!hsb_ok) {
                        printf("Error: HSB state is incorrect -- please correct it or stop the test\n  and restart servers\n");
                        has_failed = 1;
                    }                    
                    disconnect(&srv_a);

                    /* we ask at the first time and then if HSB is broken. once
                     * fixed, we go ahead automatically */
                    if (!has_failed || !hsb_ok) {
                        printf("Ready to start the test (Y/N)? ");
                        scanf("%c", &userAnswer);
                        while (getchar() != '\n') {
                            ; /* eat rest of the line */
                        }
                    }
                } while (!hsb_ok);
            }
            printf("\n");
        } else {
            connect_to(&srv_a);
            printf("Checking HSB state...\n");
            assert(check_hsb_state(&srv_a));
            disconnect(&srv_a);
        }

        printf("Start pmon\n");
        pmon_running = 1;
#ifdef WIN32
        t_pmon = _beginthread(run_pmon, 0, NULL);
        assert(t_pmon != -1);
#endif
#ifdef SS_UNIX
        pthread_create(&t_pmon, &attr, run_pmon, NULL);
        (void) t_rc; /* kill unused variable warning */
#endif
        
        printf("Start test\n");
        
        /* start test threads */
        for (i = 0; i < nthreads; i++) {
            ThreadData[i].thr_nr = i;
            ThreadData[i].rowcount = rowcount;
            ThreadData[i].testrounds = testrounds;
#ifdef WIN32
            t_rc[i] = _beginthread(run_test_set, 0, &ThreadData[i]);
            assert(t_rc[i] != -1);
#endif
#ifdef SS_UNIX
            pthread_create(&t_id[i], &attr, run_test_set, &ThreadData[i]);
#endif
        }
        
#ifdef WIN32
        WaitForMultipleObjects(nthreads, (HANDLE *) t_rc, TRUE, INFINITE);
        if (pmon_running) {
            pmon_running = 0;
            WaitForSingleObject((HANDLE *) t_pmon, INFINITE);
        }
#endif
#ifdef SS_UNIX
        for (i = 0; i < nthreads; i++) {
            pthread_join(t_id[i], NULL);
        }
        if (pmon_running) {
            pmon_running = 0;
            pthread_join(t_pmon, NULL);
        }
#endif

        /* sum up stats */
        for (i = 0; i < nthreads; i++) {
            readtxcount     += ThreadData[i].readtxcount;
            updatetxcount   += ThreadData[i].updatetxcount;
            readtotaltime   += ThreadData[i].readtotaltime;
            updatetotaltime += ThreadData[i].updatetotaltime;
        }
        
        totalTestRounds  = readtxcount + updatetxcount;
        totalTestTime    = (double) (readtotaltime + updatetotaltime);
        txtotalpersecond = (double) totalTestRounds / totalTestTime * nthreads;

        printf("\nTest results:\n");
        printf("Reads (Get-Subscriber-Data):\n");
        printf("txs: %ld (%1.1f%%), avg latency: %f ms, tx/s: %.2f\n",
               readtxcount,
               (double) readtxcount / totalTestRounds * 100.0,
               readtotaltime / readtxcount * 1000.0,
               readtxcount / readtotaltime * nthreads);
        printf("Writes (Update-Subscriber-Data):\n");
        printf("txs: %ld (%1.1f%%), avg latency: %f ms, tx/s: %.2f\n",
               updatetxcount,
               (double) updatetxcount / totalTestRounds * 100.0,
               updatetotaltime / updatetxcount * 1000.0,
               updatetxcount / updatetotaltime * nthreads);
        printf("txs total    : %ld \n", totalTestRounds);
        printf("txs total/sec: %f \n", (float)txtotalpersecond);
        printf("\n");
        printf("End test\n");
        
        return 0;
}

