/*************************************************************************
**  source      * hsb.c
**              *
**              *
**  Assumption  * Primary and Secondary servers are Up and Running
**              * with seed database.
**              *
**  description * Solid HotStandBy server usage sample.
**              * Following scenario is covered in the hsb.c
**              * 1. Connection to Primary and Secondary running databases
**              *    are made by providing the Primary DSN (Data Source Name),
**              *    Secondary DSN, Username and Password in the Command 
**              *    Line Arguments.
**              * 2. Primary is connected to Secondary.
**              * 3. At the Primary, create the Table and Procedure,
**              *    where procedure inserts records into the created table.
**              * 4. Call the Procedure atr Primary and COMMIT it.
**              * 5. Select records from the table at both Primary and 
**              *    Secondary.
**              *
**              * (C) 2005, Solid Information Technology.
**		* All Rights Reserved. 
\*************************************************************************/


/* Common system headers */
#ifdef SS_NT
#  include <windows.h>
#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include <solidodbc3.h>

/* Solid samples portability header */
#include "c.h"

/* macro definitions */
#define MAXCOLS         10
#define DATALEN         255
#define MAX_NAME_LEN    80
#ifndef bool
#  define bool            int
#endif
#ifndef TRUE
#  define TRUE            1
#endif
#ifndef FALSE
#  define FALSE           0
#endif

/* Utility macro for error handling */
#define CHECKHstmt(r, hstmt)  assert(check_hstmt(r, hstmt))

#define CHECKDbc(r, hdbc) assert(check_hdbc(r, hdbc))

#define CHECKEnv(r, henv) assert(check_henv(r, henv))

/* Connecting to datasource  */
/* Global variables */
static SQLHDBC     hdbcs, hdbcp;
static SQLHENV     henvs, henvp;
static SQLHSTMT    hstmts, hstmtp;
static SQLCHAR     primaryat[MAX_NAME_LEN]; /* Primary DSN (Data Source Name)*/
static SQLCHAR     secondaryat[MAX_NAME_LEN]; /* Secondary DSN */
static SQLCHAR     username[MAX_NAME_LEN];
static SQLCHAR     password[MAX_NAME_LEN];


/*=============================================================
Function Name    : print_error()
Purpose          : to print the error messages
Parameters       :
    rc           : Return code
    handle_type  : Handle type identifier
    handle       : handle
Returns          : none
==============================================================*/
static void print_error(SQLRETURN rc, SQLSMALLINT handle_type,
                        SQLHANDLE handle)
{
        SQLCHAR      SQLstate[MAX_NAME_LEN];
        SQLINTEGER   errorcode = 0;
        SQLCHAR      errormsg[SQL_MAX_MESSAGE_LENGTH];
        SQLSMALLINT  msglen = 0;
        SQLRETURN    r = 0;
        
        printf("Error (%d) encountered!\n", rc);
        
        /* Print more error info, if available */
        r = SQLGetDiagRec(handle_type, handle, 1, (SQLCHAR *)SQLstate,
                          (SQLINTEGER *)&errorcode, (SQLCHAR *)errormsg,
                          SQL_MAX_MESSAGE_LENGTH - 1, &msglen);
        if (r != SQL_NO_DATA_FOUND) {
            printf("%s\n", errormsg);
        }
}


static bool check_hstmt(SQLRETURN r, SQLHSTMT hstmt)
{
        if (r != SQL_SUCCESS && r != SQL_SUCCESS_WITH_INFO) {
            print_error(r, SQL_HANDLE_STMT, hstmt);
        }
        return (r == SQL_SUCCESS || r == SQL_SUCCESS_WITH_INFO);
}


static bool check_hdbc(SQLRETURN r, SQLHDBC hdbc)
{
        if (r != SQL_SUCCESS && r != SQL_SUCCESS_WITH_INFO) {
            print_error(r, SQL_HANDLE_DBC, hdbc);
        }
        return (r == SQL_SUCCESS || r == SQL_SUCCESS_WITH_INFO);
}


static bool check_henv(SQLRETURN r, SQLHENV henv)
{
        if (r != SQL_SUCCESS && r != SQL_SUCCESS_WITH_INFO) {
            print_error(r, SQL_HANDLE_ENV, henv);
        }
        return (r == SQL_SUCCESS || r == SQL_SUCCESS_WITH_INFO);
}


/* Connecting to datasource
 */
static void connect_and_alloc_stmt(SQLHENV* p_henv,
                                   SQLHDBC* p_hdbc,
                                   SQLHSTMT* p_hstmt,
                                   char* sname,
                                   char* uname,
                                   char* pwd)
{
        SQLRETURN r = 0;
        
        r = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, p_henv);
        CHECKEnv(r, *p_henv);
        r = SQLSetEnvAttr(*p_henv, SQL_ATTR_ODBC_VERSION,
                          (void *)SQL_OV_ODBC3, SQL_NTS);
        CHECKEnv(r, *p_henv);
        r = SQLAllocHandle(SQL_HANDLE_DBC, *p_henv, p_hdbc);
        CHECKDbc(r, *p_hdbc);
        r = SQLConnect(*p_hdbc, sname, SQL_NTS, uname, SQL_NTS, pwd, SQL_NTS);
        CHECKDbc(r, *p_hdbc);
        SQLSetConnectAttr(*p_hdbc, SQL_ATTR_AUTOCOMMIT, SQL_AUTOCOMMIT_OFF, 
                          SQL_IS_INTEGER);
        r = SQLAllocHandle(SQL_HANDLE_STMT, *p_hdbc, p_hstmt);
        CHECKDbc(r, *p_hdbc);
}


/*=============================================================
Function Name    : commit()
Purpose          : to commit the transactions
Parameters       :
    handleType   : Handle type identifier
    handle       : handle
Returns          : none
==============================================================*/

static void commit(SQLSMALLINT handleType, SQLHANDLE handle)
{
        SQLRETURN r;
        
        r = SQLEndTran(handleType, handle, SQL_COMMIT);
        CHECKDbc(r, handle);
}


/*=============================================================
Function Name   : sql_execute()
Purpose         : to execute the query
Parameters      :
    hstmt       : Statement Handle
    sql         : query to be executed
    cbsql       : Length of query
Returns         : none
==============================================================*/
static int sql_execute(SQLHSTMT hstmt, SQLCHAR *sql, SQLINTEGER cbsql)
{
        SQLRETURN       rc, rc1, rc2;
        SQLSMALLINT     ColCount = 0;
        SQLCHAR         ColName[MAX_NAME_LEN];
        SQLLEN          Len[MAXCOLS];
        SQLCHAR         Data[MAXCOLS][DATALEN];
        SQLINTEGER      i = 0, RowCount = 0;
        SQLSMALLINT     iCol = 0;
        
        /* Check that pointer to SQL command is not NULL. */
        assert(sql);
        
        printf("%s\n", sql);
        
        rc = SQLExecDirect(hstmt, sql, cbsql);
        
        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO)
        {
            rc1 = SQLNumResultCols(hstmt, &ColCount);
            if(ColCount > 0)
            {
                /* Print column names. */
                for (i = 0; i < ColCount; i++)
                {
                    /* Column numbers start with 1, not 0. */
                    iCol = i + 1;
                    rc1 = SQLDescribeCol(hstmt, iCol, ColName, sizeof(ColName),
                                         NULL, NULL, NULL, NULL, NULL);
                    
                    assert(rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO);
                    
                    printf("%-12s ", ColName);
                    
                    rc1 = SQLBindCol(hstmt,iCol,SQL_C_CHAR,
                                     Data[iCol-1],
                                     sizeof(Data[iCol-1]),
                                     &Len[iCol-1]);
                }
                
                printf("%s","\n");
                
                /* Print the data values returned by the query. */
                rc2 = SQLFetch(hstmt);
                while (rc2 != SQL_NO_DATA_FOUND &&
                       rc2 != SQL_ERROR)
                {
                    RowCount++;
                    for (i = 0; i < ColCount; i++)
                    {
                        if(Len[i] == SQL_NULL_DATA)
                        {
                            strcpy(Data[i],"NULL");
                        }
                        printf("%-12s ",Data[i]);
                    }
                    
                    printf("%s","\n");
                    rc2 = SQLFetch(hstmt);
                }
                
                SQLFreeStmt(hstmt, SQL_RESET_PARAMS);
                SQLFreeStmt(hstmt, SQL_UNBIND);
                SQLFreeStmt(hstmt, SQL_CLOSE);
            }
        }
        return(rc);
}



/*=============================================================
Function Name   : free_hstmt_and_disconnect()
Purpose         : to free the handles
Parameters      :
    henv        : Environment handle
    hdbc        : Connection Handle
    hstmt       : Statement Handle
Returns         : Boolean
==============================================================*/
static void free_hstmt_and_disconnect(SQLHENV henv,
                                      SQLHDBC hdbc,
                                      SQLHSTMT hstmt)
{
        SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
        SQLDisconnect(hdbc);
        SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
        SQLFreeHandle(SQL_HANDLE_ENV, henv);
}


/***********************************************************************\
 *
 *          main
 *
 * SQL query execution program main function.
 *
 * Return value - give :
 *       0      = Success return
 *      -1      = Error occurred
 */

#if defined(SS_NT)
#  ifndef SS_CDECL
#    define SS_CDECL __cdecl
#  endif
#elif defined(SS_VXW)
#  define SS_CDECL
#  define main embed_main
#else
#  define SS_CDECL
#endif

int SS_CDECL main(int argc, char * argv[])
{
        SQLRETURN   r;
        bool        b1 = TRUE;
        
        if (argc != 5)
        {
            printf(" Wrong number of arguments \n");
            printf(" **** Usage *****: \n");
            printf(" 'Primary DSN'  'Secondary DSN' 'Username' 'Password' \n");
            exit(0);
        }
        
        strcpy(primaryat, argv[1]);
        strcpy(secondaryat, argv[2]);
        strcpy(username, argv[3]);
        strcpy(password, argv[4]);
        
        /* Connect to the primary server */
        connect_and_alloc_stmt(&henvp, &hdbcp, &hstmtp, primaryat,
                               username, password);
        
        /* Connect to the secondary server */
        connect_and_alloc_stmt(&henvs, &hdbcs, &hstmts, secondaryat,
                               username, password);
        
        printf(" Perform in primary: ");
        r = sql_execute(hstmtp, "ADMIN COMMAND 'hsb state'", SQL_NTS);
        CHECKHstmt(r, hstmtp);
        
        printf(" Perform in primary: ");
        r = sql_execute(hstmtp, "ADMIN COMMAND 'hsb status connect'", SQL_NTS);
        CHECKHstmt(r, hstmtp);
        
        printf(" Perform in secondary: ");
        r = sql_execute(hstmts, "ADMIN COMMAND 'hsb state'", SQL_NTS);
        CHECKHstmt(r, hstmts);
        
        printf(" Perform in secondary: ");
        r = sql_execute(hstmts, "SELECT * FROM hsb_seedtable", SQL_NTS);
        CHECKHstmt(r, hstmts);
        
        printf(" Perform in primary: ");
        r = sql_execute(hstmtp, "ADMIN COMMAND 'hsb state'", SQL_NTS);
        CHECKHstmt(r, hstmtp);
        
        printf(" Perform in primary: ");
        r = sql_execute(hstmtp, "CALL hsb_seedproc(2315,'sys3')", SQL_NTS);
        CHECKHstmt(r, hstmtp);
        commit(SQL_HANDLE_DBC, hdbcp);
        
        printf(" Perform in primary: ");
        printf("primary status\n");
        r = sql_execute(hstmtp, "ADMIN COMMAND 'hsb state'", SQL_NTS);
        CHECKHstmt(r, hstmtp);
        
        printf(" Perform in primary: ");
        r = sql_execute(hstmtp, "SELECT * from hsb_seedtable", SQL_NTS);
        CHECKHstmt(r, hstmts);
        
        
        printf(" Perform in secondary: ");
        commit(SQL_HANDLE_DBC, hdbcs);
        
        r = sql_execute(hstmts, "SELECT * from hsb_seedtable", SQL_NTS);
        CHECKHstmt(r, hstmts);
        
        /* free hstmt and disconnect from the server */
        free_hstmt_and_disconnect(henvp, hdbcp, hstmtp);
        free_hstmt_and_disconnect(henvs, hdbcs, hstmts);
        
        return 0;
}

/* EOF */
