/*************************************************************************\
 **  file         * watchdog.c
 **               *
 **  description  * SOLID WATCHDOG for HotStandby G2 sample:
 **               *
 **               * Watchdog for Hotstandby
 **               *
 **              (C) Copyright Solid Information Technology Ltd 2000-2003
\*************************************************************************/

/*
 * How to make the Watchdog (WD) react faster:
 *   - use pessimistic mode. (cmdln: '-p' or solid.ini: 'Pessimistic = Yes')
 *     (assumes server crash is more probable than network failure.)
 *
 * Change PingInterval to a smaller number. At 20ms ('PingInterval = 20') it is
 * quite quick.
 *
 */

/* if WD is to print out timestamps; say "make timing" instead of changing */
#if 0
#  define WD_TIMING 
#endif

/* Common system headers */
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#ifdef SS_UNIX
#  include <sys/time.h>
#else
#  ifdef SS_NT
#    include <windows.h>
#    include <sys/timeb.h>
#  endif /* SS_NT */
#endif /* !SS_UNIX */

#include <solidodbc3.h>

/* Solid samples portability header */
#include "c.h"

#ifdef SS_NT
#  define ftime _ftime
#endif

/* Constants */
#define MAX_STRLEN 255
#define MAX_BUFLEN 1000

#define NORMAL_MODE   0
#define FAILURE_MODE  1

#define APA_INTERVAL         10  /* AutoPrimaryAlone check (s) */
#define MAX_DISPLAY_INTERVAL  5  /* Status display interval (s) */

/* for selecting servers */
#define SRV_FIRST   (0x01)
#define SRV_SECOND  (0x02)
#define SRV_BOTH    (SRV_FIRST|SRV_SECOND)

typedef enum {
    STATUS_CONNECTED,
    STATUS_CATCHUP,
    STATUS_BROKEN,
    STATUS_NO_RESPONSES,
    STATUS_OUT_OF_SYNC  /* unused at the moment */
} HSB_STATUS;

typedef enum {
    HSB_PRIMARY,
    HSB_SECONDARY,
    HSB_STANDALONE,
    HSB_NO_RESPONSES
} HSB_ROLE;

typedef enum {
    STATE_PRIMARY_ACTIVE,
    STATE_STANDALONE,
    STATE_PRIMARY_UNCERTAIN,
    STATE_PRIMARY_ALONE,
    STATE_SECONDARY_ACTIVE,
    STATE_SECONDARY_ALONE,
    STATE_NOT_HSB_SERVER,     /* returned by a server not configured for HSB */
    STATE_UNSET               /* not a real state */
} HSB_STATE;

typedef enum {
    SRV_ERR_NONE,
    SRV_ERR_HOST_NOT_FOUND
} SRV_ERROR;

typedef struct server_info_st {
        uchar       connect[MAX_STRLEN];
        uchar       username[MAX_STRLEN];
        uchar       password[MAX_STRLEN];
        bool        isconnect;
        HENV        henv;
        HDBC        hdbc;
        HSTMT       hstmt;
} server_info_t;

typedef struct server_status_st {
        HSB_ROLE    role;
        HSB_STATUS  status;
        HSB_STATE   state;
        SRV_ERROR   error;
} server_status_t;

typedef struct hsb_servers_st {
        server_info_t   srv_info[2];
        server_status_t srv_status[2];
        int primary;
        int secondary;
} hsb_servers_t;

#define SRV1_CONNECT(p)      ((p)->srv_info[0].connect)
#define SRV2_CONNECT(p)      ((p)->srv_info[1].connect)

#define SRV1_ROLE(p)         ((p)->srv_status[0].role)
#define SRV2_ROLE(p)         ((p)->srv_status[1].role)

#define SRV1_STATE(p)        ((p)->srv_status[0].state)
#define SRV2_STATE(p)        ((p)->srv_status[1].state)

#define SRV1_STATUS(p)       ((p)->srv_status[0].status)
#define SRV2_STATUS(p)       ((p)->srv_status[1].status)

#define PRI_CONNECT(p)       ((p)->srv_info[(p)->primary].connect)
#define SEC_CONNECT(p)       ((p)->srv_info[(p)->secondary].connect)

#define SRV1_SERVER(p)       (&((p)->srv_info[0]))
#define SRV2_SERVER(p)       (&((p)->srv_info[1]))

#define PRIMARY_SERVER(p)    ((p)->srv_info + ((p)->primary))
#define SECONDARY_SERVER(p)  ((p)->srv_info + ((p)->secondary))

#define PRIMARY_STATUS(p)    ((p)->srv_status[(p)->primary].status)
#define SECONDARY_STATUS(p)  ((p)->srv_status[(p)->secondary].status)

typedef enum {
    NOACTION = 0,
    SET_PRIMARY_ALONE,
    SWITCH_SEC_TO_PRI,
    BOTH_SERVERS_DOWN,
    BOTH_SERVERS_SECONDARY,
    PRI_CONN_DOWN,    /* connection between primary and watchdog is down */
    SEC_CONN_DOWN,    /* connection between secondary and watchdog is down */
    HSB_CONN_DOWN,    /* connection between primary and secondary is down */
    NON_VALID_HSB,    /* servers are in a nonvalid mode combination */
    IMPOSSIBLE        /* should not happen */
} WDACTION;

static char* wdaction_strings[] = {
    "No action", "Set Primary alone", "Switch Secondary to Primary",
    "Both servers are down", "Both servers are secondary",
    "Connection to Primary is down", "Connection to Secondary is down",
    "HotStandby connection is down", "Not valid HSB", "Impossible"
};

typedef struct config_st {
        int   NumRetry;
        int   PingInterval; /* (ms) */
        bool  AutoSwitch;  /* Automatically switch:
                            *   - Primary uncertain -> Primary alone
                            *   - 1 down, 1 secondary -> 1 primary.
                            */
        bool  DualSecAutoSwitch; /* 2 secondaries -> 1 primary, 1 secondary */ 

        bool  ObserveOnly; /* watchdog makes no actions, just observes
                            * the situation. */
        bool  Pessimistic; /* if watchdog expects the worst (ie. server down
                            * instead of link down). */
        bool  LoneTrigger; /* quit when only one server is up and it is
                            * PRIMARY ALONE. */
        uchar logfilename[MAX_STRLEN];

} config_t;

/* true, 1, if an error occured */
#define CHECK(rc)   ((rc) != SQL_SUCCESS && (rc) != SQL_SUCCESS_WITH_INFO) 

/* runtime flags */

static int debug = 0;
static int warn  = 0;

/* module variables */

static config_t config;

/* Function prototypes */

/* utility */
static char* skip_spaces(char* s);
static void  trim_spaces(char* s);
static void  print_wdaction(WDACTION wa, FILE* out);

/* Layer 2 */
static int initialize_connection(server_info_t* srv_info);
static int execute_single_statement(server_info_t* srv_info, uchar* statement, 
                                    uchar* retcode, uchar* text);
static int retrieve_results(server_info_t* srv_info, uchar* retcode,
                            uchar* text);
static void terminate_connection(server_info_t* srv_info);
static int exec_admin_command(server_info_t* srv_info, uchar* cmd,
                              uchar* retcode, uchar* text);
static bool server_isautoprimaryalone(server_info_t* srv_info);
static int get_role(server_info_t* srv_info, server_status_t* srv_status,
                    uchar* state);
static int get_state(server_info_t* srv_info, server_status_t* srv_status,
                     uchar* state);
static int get_connect_status(server_info_t* srv_info,
                              server_status_t* srv_status);
static int set_primary_alone(hsb_servers_t* hsb_servers);
static int wait_for_switch(server_info_t* srv_info);
static int switch_sec_to_pri(hsb_servers_t* hsb_servers);
static int wait_for_connect(server_info_t* srv_info);
static int hsb_connect(hsb_servers_t* hsb_servers, uchar* text);

/* Layer 3 */
static void init_hsb_servers(hsb_servers_t *hsb_servers, 
                             uchar *connect1, uchar *username1,
                             uchar *password1, 
                             uchar *connect2, uchar *username2,
                             uchar *password2);
static void print_server_status(server_info_t *srv_info,
                                server_status_t *srv_status,
                                FILE *out);
static void print_hsb_servers_status(hsb_servers_t *hsb_servers, FILE *out);
static void get_server_status(server_info_t *srv_info,
                              server_status_t *srv_status);
static void get_hsb_servers_status(hsb_servers_t *hsb_servers, int which);

/* Layer 4 */
static void mysleep(int millisec);
static void init_config(config_t *config);
static int  process_line(uchar *buf, uchar *section,
                         uchar *keyname, uchar *keyvalue);
static int  read_inifile(hsb_servers_t *hsb_servers, config_t *config);
static WDACTION check_hsb_servers(hsb_servers_t *hsb_servers, uchar *err_msg,
                                  int checked_servers);
static bool server_alone(hsb_servers_t *hsb_servers);

/* watchdog actions */
static int watchdog_set_pri_alone(hsb_servers_t* hsb_servers, bool autoswitch, 
                                  FILE* outfile);
static int watchdog_switch_sec_to_pri(hsb_servers_t* hsb_servers,
                                      bool autoswitch, FILE* outfile);
static int watchdog_switch_one_to_pri(hsb_servers_t* hsb_servers,
                                      bool autoswitch, FILE* outfile);
static int watchdog_hsb_connect(hsb_servers_t* hsb_servers, FILE* outfile);

#ifdef WD_TIMING
/*##**********************************************************************\
 *
 *      print_timestamp
 *
 * Prints current time in seconds with precision to microseconds.
 *
 * On Linux you can get the same with "date +%s.%N" and
 * a UNIX alternative would be "date +%T.%N". 
 *
 * Parameters:
 *      None.
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
void print_timestamp(void)
{
        char* ts_fmt = "[Watchdog timestamp]: %ld.%06ld\n";

#if defined(SS_UNIX)
        struct timeval tv;
        
        gettimeofday(&tv, NULL);
        printf(ts_fmt, tv.tv_sec, tv.tv_usec); 
#elif defined(SS_WINDOWS) || defined(SS_NT)
        struct _timeb t; /* with or without the '_'? */

        ftime(&t);
        printf(ts_fmt, (long) t.time, (long) t.millitm * 1000);
#endif
        
}
#endif /* WD_TIMING */

/*#***********************************************************************\
 *
 *      server_alone
 *
 * Checks if there is only one server standing and it is in PRIMARY ALONE state
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby servers
 *
 * Return value:
 *      TRUE, if so.
 *      FALSE, if not.
 *
 * Limitations:
 *
 * Globals used:
 */
static bool server_alone(hsb_servers_t* hsb_servers)
{
        if ((SRV1_STATUS(hsb_servers) == STATUS_NO_RESPONSES)
            && (SRV2_STATE(hsb_servers) == STATE_PRIMARY_ALONE)) {
            return TRUE;
        }
        if ((SRV2_STATUS(hsb_servers) == STATUS_NO_RESPONSES)
            && (SRV1_STATE(hsb_servers) == STATE_PRIMARY_ALONE)) {
            return TRUE;
        }
            
        return FALSE;
}        

/*#***********************************************************************\
 *
 *      skip_spaces
 *
 * Skips spaces from the start of the string (aka ltrim).
 *
 * Parameters:
 *      buf - in, use
 *          the string
 *
 * Return value:
 *      Pointer to the new starting position of the string.
 *
 * Limitations:
 *
 * Globals used:
 */
static char* skip_spaces(char* buf)
{
        while (*buf != '\0' && isspace(*buf)) {
            buf++;
        }
        return (buf);
}

/*#***********************************************************************\
 *
 *      trim_spaces
 *
 * Trim spaces from the end of the string (aka rtrim).
 *
 * Parameters:
 *      buf - in, use
 *          the string
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void trim_spaces(char* buf)
{
        int i;

        i = strlen(buf);
        while (i > 0 && isspace(buf[i-1])) {
            i--;
        }
        buf[i] = '\0';
}

/*#***********************************************************************\
 *
 *      hstmt_success
 *
 * Checks if error code from statement functions is success. If not,
 * prints error message.
 *
 * Parameters:
 *      hstmt - in, use
 *          statement to examine
 *
 *      rc - in
 *          return code of the statement
 *
 * Return value:
 *      1, if was successful
 *      0, if not.
 *
 * Limitations:
 *
 * Globals used:
 */
static int hstmt_success(HSTMT hstmt, RETCODE rc)
{
        UCHAR   szSqlState[6];
        UCHAR   szErrorMsg[255];
        SDWORD  naterr;
        SWORD   length;

        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {
            return(1);
        }

        rc = SQLError(SQL_NULL_HENV, SQL_NULL_HDBC, hstmt, szSqlState,
                      &naterr, szErrorMsg, sizeof(szErrorMsg), &length);
        printf("Error %s : %s \n", szSqlState, szErrorMsg);

        return(0);
}

/*#***********************************************************************\
 *
 *      initialize_connection
 *
 * Allocate and initialize connection elements and establish connection.
 *
 * Parameters:
 *
 *      srv_info - in, use
 *          server info.
 *
 * Return value:
 *      0, if successful or already connected.
 *      -1, on general error.
 *      -2, if host not found.
 *
 * Limitations:
 *
 * Globals used:
 */
static int initialize_connection(server_info_t *srv_info)
{
        RETCODE rc;
        HENV  henv  = NULL;   /* local variables - for error handling macros */
        HDBC  hdbc  = NULL;
        HSTMT hstmt = NULL;
        int status = -1;      /* return code */

        if (debug >= 2) {
            printf("%s:initialize_connection\n", srv_info->connect);
        }

        if (srv_info->isconnect == TRUE) {
            if (debug >= 2) {
                printf("%s:initialize_connection:srv_info->isconnect=1\n",
                       srv_info->connect);
            }
            return 0; /* already connected */
        }

        /* Allocate environment and connection handles. */
        rc = SQLAllocEnv(&henv);
        if (CHECK(rc) == 1) {
            if (debug) {
                assert(0);
            }
            return -1; /* fail */
        }

        rc = SQLAllocConnect(henv, &hdbc);
        if (CHECK(rc) == 1) {
            if (debug) {
                assert(0);
            }
            SQLFreeEnv(henv);
            return -1; /* fail */
        }

        rc = SQLConnect(hdbc, srv_info->connect, SQL_NTS,
                        srv_info->username, SQL_NTS, 
                        srv_info->password, SQL_NTS);
        if (CHECK(rc) == 1) { /* failed */
            SQLSMALLINT recno = 1;
            SQLCHAR     sqlstate[6];
            SQLINTEGER  naterr;
            SQLCHAR     msgtext[MAX_STRLEN];
            SQLSMALLINT len;
            
            if (debug >= 2) {
                printf("%s:initialize_connection:SQLConnect failed\n",
                       srv_info->connect);
            }
            while (1) { /* get error data */
                rc = SQLGetDiagRec(SQL_HANDLE_DBC, hdbc, recno, sqlstate,
                                   &naterr, msgtext, sizeof(msgtext), &len);
                if (CHECK(rc) == 1) { /* failed or no more data */
                    break;
                }
                if (debug >= 2) {
                    printf("Error %s: %s, %ld\n", sqlstate, msgtext,
                           (long) naterr);
                }
                if ((long) naterr == 21322) { /* host not found */
                    status = -2;
                }

                recno++;
            }
        } else { /* we are now connected */
            rc = SQLAllocStmt(hdbc, &hstmt);
            if (CHECK(rc) == 1) {
                if (debug) {
                    assert(0);
                }
                SQLDisconnect(hdbc);
            
            } else {
                /* all is ok */
                srv_info->henv      = henv;
                srv_info->hdbc      = hdbc;
                srv_info->hstmt     = hstmt;
                srv_info->isconnect = TRUE;
                
                return 0;
            }
        }
        
        /* free resources, connection could not be made */
        SQLFreeConnect(hdbc);
        SQLFreeEnv(henv);
        
        return status; /* -1 probably */
}

/*#***********************************************************************\
 *
 *      execute_single_statement
 *
 * Execute a single sql statement.
 *
 * Parameters:
 *      srv_info - in, use
 *          server info
 *
 *      statement - in, use
 *          sql statement
 *
 *      retcode - out, use
 *          return code from the statement.
 *
 *      text - out, use
 *          text reply from server.
 *
 * Return value:
 *      0, if ok
 *      -1, if an error occured
 *
 * Limitations:
 *
 * Globals used:
 */
static int execute_single_statement(server_info_t* srv_info, 
                                    uchar* statement,
                                    uchar* retcode, uchar* text)
{
        RETCODE rc;
        HENV henv = srv_info->henv; 
        HDBC hdbc = srv_info->hdbc;
        HSTMT hstmt = srv_info->hstmt;

        if (debug >= 2) {
            printf(">>> exec stmt: '%s'\n", statement);
        }
        
        /* Execute the SQL statement. */
        rc = SQLExecDirect(hstmt, statement, SQL_NTS);
        if (!hstmt_success(hstmt, rc)) {

            if (debug >= 2) {
                printf("%s:execute_single_statement:SQLExecDirect(%s) failed\n",
                       srv_info->connect, statement);
            }
                
            /* Close any possibly open cursors */
            rc = SQLFreeStmt(hstmt, SQL_CLOSE );
            if (!hstmt_success(hstmt, rc)) {
                if (debug) {
                    assert(0);
                }
                return -1;
            }
            SQLTransact(henv, hdbc, SQL_ROLLBACK);
            return -1;
        }

        if (retrieve_results(srv_info, retcode, text) != 0) {
            return -1;
        }
        return 0;
}

/*#***********************************************************************\
 *
 *      retrieve_results
 *
 * Fetch the result of the executed statement
 *
 * Parameters:
 *      srv_info - in, use
 *          Server we executed on.
 *
 *      retcode - in out, use
 *          Space for return code -> return code of the sql statement.
 *          Can be NULL.
 *
 *      text - in out, use
 *          Space for text reply -> output text of the sql statement.
 *          Can be NULL.
 *
 * Return value:
 *      0, if OK
 *      -1, if error occured
 *
 * Limitations:
 *      Both retcode and text should not be NULL -> error (Why would you
 *      want to do ita anyway?).
 *
 * Globals used:
 */
static int retrieve_results(server_info_t* srv_info, uchar* retcode,
                            uchar* text) 
{
        int     i;
        SQLUSMALLINT colnum;
        HENV    henv = srv_info->henv; 
        HDBC    hdbc = srv_info->hdbc;
        HSTMT   hstmt = srv_info->hstmt;
        RETCODE rc = 0;

        UCHAR   colname[32];
        SWORD   coltype;
        SWORD   colnamelen;
        SWORD   nullable;
        SQLLEN  collen[2];
        SWORD   scale;
        SQLLEN  outlen[2];
        SWORD   nresultcols;

        uchar   RC_buf[MAX_BUFLEN];
        uchar   text_buf[MAX_BUFLEN];
        uchar*  RC_ptr;
        uchar*  text_ptr;
        
        RC_ptr      = (retcode == NULL) ? RC_buf: retcode;
        text_ptr    = (text == NULL)    ? text_buf: text;
        RC_ptr[0]   = '\0';
        text_ptr[0] = '\0';

        if (retcode == NULL && text == NULL) {
            /* there's no point in having both NULL */
            return -1;
        }
        
        rc = SQLNumResultCols(hstmt, &nresultcols);
        if (!hstmt_success(hstmt, rc)) {
            if (debug) {
                assert(0);
            }
            return -1;
        }

        if (nresultcols != 2) {
            if (debug) {
                assert(0);
            }
            return -1;
        }

        for (i = 0; i < nresultcols; i++) {
            colnum = (SQLUSMALLINT) (i + 1);

            rc = SQLDescribeCol(hstmt, colnum, colname,
                                (SWORD) sizeof(colname), &colnamelen,
                                &coltype, &collen[i], &scale, 
                                &nullable);
            if (!hstmt_success(hstmt, rc)) {
                if (debug) {
                    assert(0);
                }
                return -1;
            }

            if (strcmp(colname, "RC") == 0) {
                rc = SQLBindCol(hstmt, colnum, SQL_C_CHAR, RC_ptr,
                                collen[i], &outlen[i]);
                if (!hstmt_success(hstmt, rc)) {
                    if (debug) {
                        assert(0);
                    }
                    return -1;
                }
            } else if (strcmp(colname, "TEXT") == 0) {
                rc = SQLBindCol(hstmt, colnum, SQL_C_CHAR, text_ptr,
                                collen[i], &outlen[i]);
                if (!hstmt_success(hstmt, rc)) {
                    if (debug) {
                        assert(0);
                    }
                    return -1;
                }
            } else {
                if (debug) {
                    assert(0);
                }
                return -1;
            }
        } /* end for */

        rc = SQLFetch(hstmt);
        if (!hstmt_success(hstmt, rc)) {
            if (debug) {
                assert(0);
            }
            return -1;
        }

        /* Close open cursors */
        rc = SQLFreeStmt(hstmt, SQL_CLOSE );
        if (!hstmt_success(hstmt, rc)) {
            if (debug) {
                assert(0);
            }
            return -1;
        }

        /* Unbind variables */
        rc = SQLFreeStmt(hstmt, SQL_UNBIND );
        if (!hstmt_success(hstmt, rc)) {
            if (debug) {
                assert(0);
            }
            return -1;
        }

        rc = SQLTransact(henv, hdbc, SQL_COMMIT);
        if (CHECK(rc)) {
            return -1;
        }

        /* trim the trailing spaces */
        trim_spaces(RC_ptr);
        trim_spaces(text_ptr);

        return 0;
}

/*#***********************************************************************\
 *
 *      terminate_connection
 *
 * Disconnect and free the connection element.
 *
 * Parameters:
 *      srv_info - in, use
 *          server info
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void terminate_connection(server_info_t* srv_info)
{
        HENV henv   = srv_info->henv; 
        HDBC hdbc   = srv_info->hdbc;
        HSTMT hstmt = srv_info->hstmt;

        if (srv_info->isconnect == FALSE) {
            return;
        }

        /* Free the statement handle.       */
        /* Free the connection handle.      */
        /* Free the environment handle.     */

        SQLFreeStmt(hstmt, SQL_DROP);
        SQLFreeConnect(hdbc);
        SQLFreeEnv(henv);
        srv_info->henv      = NULL;
        srv_info->hdbc      = NULL;
        srv_info->hstmt     = NULL;
        srv_info->isconnect = FALSE;
}

/*************************************************************************\
 *
 *              exec_admin_command
 *
 * Execute an admin command.
 *
 * Parameters - input:
 *      srv_info: server info
 *      cmd: admin command to execute
 *
 * Parameters - output:
 *      retcode: return code of the admin command
 *      text: output text of the admin command
 *
 * Returns:
 *      0, if okay.
 *      -1, if an error occured.
 *      -2, if host not found.
 */
static int exec_admin_command(server_info_t* srv_info, uchar* cmd,
                              uchar* retcode, uchar* text)
{
        uchar admincmd[MAX_BUFLEN];
        int rc = 0;

        sprintf(admincmd, "admin command '%s'", cmd);

        if (   ((rc = initialize_connection(srv_info)) != 0)
            || ((rc = execute_single_statement(srv_info, admincmd,
                                               retcode, text)) != 0) ) {

            terminate_connection(srv_info);
            return rc;
        }

        return 0;
}

/*##**********************************************************************\
 *
 *      server_isautoprimaryalone
 *
 * Checks if a server has AutoPrimaryAlone turned on. That is BAD.
 *
 * Parameters:
 *      srv_info - in, use
 *          server data.
 *
 * Return value:
 *      TRUE, if server's AutoPrimaryAlone is ON/TRUE.
 *      FALSE, if not. False is good.
 *
 * Limitations:
 *
 * Globals used:
 */
static bool server_isautoprimaryalone(server_info_t* srv_info)
{
        uchar text[MAX_BUFLEN];
        char* p;
        int rc;

        rc = exec_admin_command(srv_info, "hsb parameter AutoPrimaryAlone",
                                NULL, text);
        if (rc != 0) {
            return FALSE;
        }

        /* reply as: 'HotStandby AutoPrimaryAlone <current> <ini> <factory>' */

        /* skip section and param name */
        for (p = text; *p != ' ' && *p; p++)
            ;
        for (p++; *p != ' ' && *p; p++)
            ;
        p++;

        if (p[0] == 'Y' || p[0] == 'y') {
            if (debug) {
                printf("Warning! AutoPrimaryAlone = TRUE on '%s'!\n",
                       srv_info->connect);
            }
            return TRUE;
        }
        if (!(p[0] == 'N' || p[0] == 'n')) {
            printf("Watchdog: Error: AutoPrimaryAlone query result is not bool: '%s'\n",
                   p);
        }
        
        return FALSE;
}

/*#***********************************************************************\
 *
 *      get_role
 *
 * Get the hotstandby role of the server.
 *
 * Parameters:
 *      srv_info - in, use
 *          server info
 *
 *      srv_status - out, use
 *          server status
 *
 *      state - in
 *          state text of the server, or NULL. If not NULL, function gets the
 *          server role from the state text.
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *
 * Limitations:
 *
 * Globals used:
 */
static int get_role(server_info_t* srv_info, server_status_t* srv_status,
                    uchar* state)
{
        uchar text[MAX_BUFLEN];
        int rc;
        uchar* ptext = state;

        if (state == NULL) {
            rc = exec_admin_command(srv_info, "hsb role", NULL, text);
            if (rc == -1) {
                return -1;
            } else if (rc == -2) {
                srv_status->error = SRV_ERR_HOST_NOT_FOUND;
                return -1;
            }
            if (debug >= 2) {
                printf("%s:get_role:%s\n", srv_info->connect, text);
            }
            ptext = text;
        }

        srv_status->error = SRV_ERR_NONE;
        
        if (strncmp(ptext, "PRIMARY", 7) == 0) {
            srv_status->role = HSB_PRIMARY;

        } else if (strncmp(ptext, "SECONDARY", 9) == 0) {
            srv_status->role = HSB_SECONDARY;

        } else {
            srv_status->role = HSB_STANDALONE;
        }
        return 0;
}

/*#***********************************************************************\
 *
 *      get_state
 *
 * Get the HotStandby state of the server
 *
 * Parameters:
 *      srv_info - in, use
 *          server info
 *
 *      srv_status - out, use
 *          server status
 *
 *      state - out, use
 *          server's state text reply, may be NULL.
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *      -2, if bad command (server is G1 or non-HSB).
 *
 * Limitations:
 *
 * Globals used:
 */
static int get_state(server_info_t* srv_info, server_status_t* srv_status,
                     uchar* state)
{
        uchar text[MAX_BUFLEN];
        uchar retcode[MAX_BUFLEN];

        if (exec_admin_command(srv_info, "hsb state", retcode, text) != 0) {
            return -1;
        }

        if (debug >= 2) {
            printf("%s:get_state:%s\n", srv_info->connect, text);
        }

        if (state != NULL) {
            strncpy(state, text, MAX_BUFLEN-1);
            state[MAX_BUFLEN-1] = '\0';
        }

        if (strcmp(text, "PRIMARY ACTIVE") == 0) {
            srv_status->state = STATE_PRIMARY_ACTIVE;
        } else if (strcmp(text, "STANDALONE") == 0) {
            srv_status->state = STATE_STANDALONE;
        } else if (strcmp(text, "PRIMARY UNCERTAIN") == 0) {
            srv_status->state = STATE_PRIMARY_UNCERTAIN;
        } else if (strcmp(text, "PRIMARY ALONE") == 0) {
            srv_status->state = STATE_PRIMARY_ALONE;
        } else if (strcmp(text, "SECONDARY ACTIVE") == 0) {
            srv_status->state = STATE_SECONDARY_ACTIVE;
        } else if (strcmp(text, "SECONDARY ALONE") == 0) {
            srv_status->state = STATE_SECONDARY_ALONE;
        } else if (strcmp(text, "This is a non-HotStandby server.") == 0) {
            srv_status->state = STATE_NOT_HSB_SERVER;
            return -2;
        } else if (strcmp(retcode, "14526") == 0) {
            /* "Bad Hot Standby command", which means that server
             * is HSB G1 or not HSB at all. */
            return -2;
        } else {
            return -1;
        }
        return 0;
}

/*#***********************************************************************\
 *
 *      get_connect_status
 *
 * Get the connect status of the hotstandby server.
 *
 * Parameters:
 *      srv_info - in, use
 *          server info
 *
 *      srv_status - out, use
 *          server status.
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *
 * Limitations:
 *
 * Globals used:
 */
static int get_connect_status(server_info_t* srv_info,
                              server_status_t* srv_status)
{
        uchar text[MAX_BUFLEN];

        if (exec_admin_command(srv_info, "hsb status connect",
                               NULL, text) != 0) {
            return -1;
        }

        if (debug >= 2) {
            printf("%s:get_connect_status:%s\n", srv_info->connect, text);
        }

        if (strcmp(text, "CONNECTED") == 0) {
            srv_status->status = STATUS_CONNECTED;
        } else if (strcmp(text, "CATCHUP") == 0) {
            srv_status->status = STATUS_CATCHUP;
        } else {
            srv_status->status = STATUS_BROKEN;
        }

        return 0;
}

/*#***********************************************************************\
 *
 *      set_primary_alone
 *
 * Connect and set Primary server to Primary Alone mode.
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby server info.
 *
 * Return value:
 *      0, if OK
 *      -1, if some WD error
 *      -2, if errorcode conversion failed.
 *      else the errorcode
 *
 * Limitations:
 *      If there are more than one primary (which isn't a legal HSB state),
 *      you can't choose which to set alone.  You may go around this by
 *      setting the value on primary from hsb_servers_t.
 *
 * Globals used:
 */
static int set_primary_alone(hsb_servers_t* hsb_servers)
{
        uchar text[MAX_BUFLEN];
        uchar retcode[MAX_BUFLEN];
        int rv;
        
        if (exec_admin_command(PRIMARY_SERVER(hsb_servers), 
                               "hsb set primary alone", retcode, text)) {
            return -1;
        }

        /* check reply for success */
        if (strcmp(retcode, "0") == 0) { /* OK */
            return 0;
        }
        
        if (debug) {
            printf("set_primary_alone: failed: %s: %s\n", retcode, text);
        }

        rv = atoi(retcode);
        if (rv == 0) {
            rv = -2;
        }
        
        return rv;
}

/*#***********************************************************************\
 *
 *      wait_for_switch
 *
 * Monitor the server until the switch is successful or an error has occured.
 *
 * Parameters:
 *      srv_info - in, use
 *          server info
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *
 * Limitations:
 *
 * Globals used:
 */
static int wait_for_switch(server_info_t* srv_info)
{
        uchar text[MAX_BUFLEN];

        while (1) {

            if (exec_admin_command(srv_info, "hsb status switch",
                                   NULL, text)) {
                return -1;
            }

            if (debug >= 2) {
                printf("%s:wait_for_switch:%s\n", srv_info->connect, text);
            }

            if (strcmp(text, "SUCCESS") == 0) {
                break;
            } else if (strcmp(text, "ACTIVE") == 0) {
                mysleep(config.PingInterval);
                continue;
            } else {
                return -1; 
            }
        }
        
        return 0;
}

/*#***********************************************************************\
 *
 *      switch_sec_to_pri
 *
 * Switch the secondary server to primary server.
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby server info
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *
 * Limitations:
 *
 * Globals used:
 */
static int switch_sec_to_pri(hsb_servers_t* hsb_servers)
{
        uchar text[MAX_BUFLEN];
        uchar retcode[MAX_BUFLEN];
        int oldprimary;
        int rv;

        if (exec_admin_command(SECONDARY_SERVER(hsb_servers), 
                               "hsb switch primary", retcode, text) != 0) {
            return -1;
        }

        if (debug >= 2) {
            printf("switch_sec_to_pri: rc=%s, text=\"%s\"\n", retcode, text);
        }
        
        if (strcmp(retcode, "0") != 0) { /* not OK */
            rv = atoi(retcode);
            if (rv == 0) {
                rv = -2;
            }
            return rv;
        }
        
        wait_for_switch(SECONDARY_SERVER(hsb_servers));

        oldprimary = hsb_servers->primary;
        hsb_servers->primary   = hsb_servers->secondary;
        hsb_servers->secondary = oldprimary;

        return 0;
}

/*#***********************************************************************\
 *
 *      wait_for_connect
 *
 * Monitor the server until HSB connection is made or error has occured.
 *
 * Parameters:
 *      srv_info - in, use
 *          server info
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *
 * Limitations:
 *
 * Globals used:
 */
static int wait_for_connect(server_info_t* srv_info)
{
        uchar text[MAX_BUFLEN];
        uchar retcode[MAX_BUFLEN];
        
        while (1) {

            if (exec_admin_command(srv_info, "hsb status connect",
                                   retcode, text)) {
                return -1;
            }

            if (debug >= 2) {
                printf("%s:wait_for_connect: %s: %s\n", srv_info->connect,
                       retcode, text);
            }

            if ((strcmp(retcode, "14007") == 0) /* "CONNECTING" */
                || (strcmp(retcode, "14008") == 0)) { /* "CATCHUP" */
                mysleep(config.PingInterval);
                continue;
            }

            if (strcmp(retcode, "0") == 0) { /* OK */
                break;
            }            
            else if (strcmp(retcode, "14537") == 0) { /* BROKEN */
                /* timeout, could not connect */
                return -1;
            } else {
                /* 14700 & 14701 for example ... */
                return -1;
            }
            
        }
        return 0;
}

/*#***********************************************************************\
 *
 *      hsb_connect
 *
 * Connect the primary server to secondary server.
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby server info
 *
 *      t - in, use
 *          buffer to return command's text in
 *
 * Return value:
 *      0, if OK
 *      -1, if some WD specific error was encountered
 *      else the error code of the command.
 *
 * Limitations:
 *
 * Globals used:
 */
static int hsb_connect(hsb_servers_t* hsb_servers, uchar* t)
{
        uchar btext[MAX_BUFLEN];
        uchar retcode[MAX_BUFLEN];
        uchar* text = btext;

        if (t != NULL) {
            text = t;
        }
        
        if (exec_admin_command(PRIMARY_SERVER(hsb_servers), 
                               "hsb connect", retcode, text) != 0) {
            return -1;
        }

        if (strcmp(retcode, "0") == 0) { /* OK */
            /* wait for connection */
            wait_for_connect(PRIMARY_SERVER(hsb_servers));

            return 0;
        } else {
            if (debug >= 1) {
                printf("hsb_connect: failed: %s: %s\n", retcode, text);
            }
        }
        
        return atoi(retcode);
}

/*#***********************************************************************\
 *
 *      init_hsb_servers
 *
 * Initialize the hotstandby server structure.
 *
 * Parameters:
 *      hsb_servers - out
 *          hotstandby server info
 *
 *      connect1 - in, use
 *          connection string for the first server.
 *
 *      username1 - in, use
 *          username for the first server.
 *
 *      password1 - in, use
 *          password for the first server.
 *
 *      connect2 - in, use
 *          connection string for the second server.
 *
 *      username2 - in, use
 *          username for the second server.
 *
 *      password2 - in, use
 *          password for the second server.
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void init_hsb_servers(
        hsb_servers_t* hsb_servers, 
        uchar* connect1, uchar* username1, uchar* password1,
        uchar* connect2, uchar* username2, uchar* password2)
{
        strcpy (hsb_servers->srv_info[0].connect, connect1);
        strcpy (hsb_servers->srv_info[0].username, username1);
        strcpy (hsb_servers->srv_info[0].password, password1);
        hsb_servers->srv_info[0].henv      = NULL;
        hsb_servers->srv_info[0].hdbc      = NULL;
        hsb_servers->srv_info[0].hstmt     = NULL;
        hsb_servers->srv_info[0].isconnect = FALSE;
        
        strcpy (hsb_servers->srv_info[1].connect, connect2);
        strcpy (hsb_servers->srv_info[1].username, username2);
        strcpy (hsb_servers->srv_info[1].password, password2);
        hsb_servers->srv_info[1].henv      = NULL;
        hsb_servers->srv_info[1].hdbc      = NULL;
        hsb_servers->srv_info[1].hstmt     = NULL;
        hsb_servers->srv_info[1].isconnect = FALSE;

        hsb_servers->primary   = 0;
        hsb_servers->secondary = 1;
}

/*#***********************************************************************\
 *
 *      print_server_status
 *
 * Print the server status
 *
 * Parameters:
 *      srv_info - in, use
 *          server info
 *
 *      srv_status - in, use
 *          server status
 *
 *      out - out, use
 *          output file (eg. stdout).
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void print_server_status(server_info_t* srv_info,
                                server_status_t* srv_status,
                                FILE* out)
{
        switch (srv_status->role) {
            case HSB_PRIMARY: 
                fprintf(out, "\tPrimary server '%s'\n\t    Connect status:", 
                        srv_info->connect);
                break;
            case HSB_SECONDARY: 
                fprintf(out, "\tSecondary server '%s'\n\t    Connect status:", 
                        srv_info->connect);
                break;
            case HSB_STANDALONE: 
                fprintf(out, "\tServer '%s' is a Standalone server\n",
                        srv_info->connect);
                return;
                break;
            case HSB_NO_RESPONSES: 
                fprintf(out, "\tServer '%s' has no responses", 
                        srv_info->connect);
                switch (srv_status->error) {
                    case SRV_ERR_HOST_NOT_FOUND:
                        fprintf(out, ": host not found\n");
                        break;
                    default:
                        fprintf(out, "\n");
                        break;
                }
                return;

                break;
            default:
                break;
        }
        switch (srv_status->status) {
            case STATUS_CONNECTED: 
                fprintf(out, " Connected");
                break;
            case STATUS_CATCHUP: 
                fprintf(out, " Catchup");
                break;
            case STATUS_BROKEN: 
                fprintf(out, " Broken");
                break;
            case STATUS_NO_RESPONSES: 
                fprintf(out, " Unknown");
                break;
            default:
                break;
        }
        fprintf(out, ", State: ");
        switch (srv_status->state) {
            case STATE_PRIMARY_ACTIVE:
                fprintf(out, " (Primary active)");
                break;
            case STATE_STANDALONE:
                fprintf(out, " (Standalone)");
                break;
            case STATE_PRIMARY_UNCERTAIN:
                fprintf(out, " (Primary uncertain)");
                break;
            case STATE_PRIMARY_ALONE:
                fprintf(out, " (Primary alone)");
                break;
            case STATE_SECONDARY_ACTIVE:
                fprintf(out, " (Secondary active)");
                break;
            case STATE_SECONDARY_ALONE:
                fprintf(out, " (Secondary alone)");
                break;
            case STATE_NOT_HSB_SERVER:
                fprintf(out, " (This is a non-HotStandby server)");
                break;
            default:
                break;
        }
        fprintf(out, "\n");
}

/*#***********************************************************************\
 *
 *      print_hsb_servers_status
 *
 * Print the status of the hotstandby servers.
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby server info
 *
 *      out - out, use
 *          output file (eg. stdout)
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void print_hsb_servers_status(hsb_servers_t* hsb_servers, FILE* out)
{
        char* ct;
        time_t now;
        
        now = time(NULL);
        ct  = ctime(&now);
        fprintf(out, "Servers Status: %s", ct);
        print_server_status(&hsb_servers->srv_info[0], 
                            &hsb_servers->srv_status[0],
                            out);

        print_server_status(&hsb_servers->srv_info[1], 
                            &hsb_servers->srv_status[1],
                            out);
}

/*#***********************************************************************\
 *
 *      get_server_status
 *
 * get the server role, connect status and state of the server.
 *
 * Parameters:
 *      srv_info - in, use
 *          server info
 *
 *      srv_status - out
 *          server status
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void get_server_status(server_info_t* srv_info,
                              server_status_t* srv_status)
{
        int rv;
        uchar state[MAX_BUFLEN];
        
        srv_status->role   = HSB_NO_RESPONSES;
        srv_status->status = STATUS_NO_RESPONSES;
        srv_status->state  = STATE_NOT_HSB_SERVER;
        
        rv = get_state(srv_info, srv_status, state);
        if (rv == -1) {
            return;
        }

        /* rv == -2 -> state couldn't be read, ie. server is HSB G1
         * or non-HSB */
        if (get_role(srv_info, srv_status, (rv == -2) ? NULL: state) != 0) {
            return;
        }

        if (   (srv_status->role != HSB_PRIMARY)
            && (srv_status->role != HSB_SECONDARY)) {
            /* role == standalone -> HSB is not in use. */
            return;
        }

        if (get_connect_status(srv_info, srv_status) != 0) {
            return;
        }

        /* last */
        return;
}

/*#***********************************************************************\
 *
 *      get_hsb_servers_status
 *
 * get the server status, role and state of the hotstandby servers.
 *
 * Parameters:
 *      hsb_servers - in out, use
 *          hotstandby server info
 *
 *      which - in
 *          which servers' status to check
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void get_hsb_servers_status(hsb_servers_t* hsb_servers, int which)
{
        if (which & SRV_FIRST) {
            get_server_status(&hsb_servers->srv_info[0],
                              &hsb_servers->srv_status[0]);
        }
        if (which & SRV_SECOND) {
            get_server_status(&hsb_servers->srv_info[1],
                              &hsb_servers->srv_status[1]);
        }
        
        return;
}

#if defined(SS_CRX) || defined(SS_VXW)
#include <time.h>
#endif

/*#***********************************************************************\
 *
 *      mysleep
 *
 * Sleep for specified number of milliseconds.
 *
 * Parameters:
 *      millisec - in
 *          number of milliseconds to sleep.
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void mysleep(int millisec)
{
        if (millisec < 0) {
            return;
        }
#if defined(SS_WINDOWS) || defined(SS_NT)
        Sleep(millisec);
/*#elif defined(SS_CRX) || defined(SS_VXW)*/
#else
        {
            struct timespec req;
            struct timespec rem;
            int rv;

            req.tv_sec  =  millisec / 1000;
            req.tv_nsec = (millisec % 1000) * 1000000;
            do {
                rv = nanosleep(&req, &rem);
                req = rem;
            } while (rv == -1);
        }
#endif
}

/*#***********************************************************************\
 *
 *      print_wdaction
 *
 * Prints out WDACTION as text.
 *
 * Parameters:
 *      wa - in
 *          action to print
 *
 *      out - out, use
 *          output file (eg. stdout)
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void print_wdaction(WDACTION wa, FILE* out) {
        fprintf(out, "Watchdog action: ");
        switch (wa) {
            case NOACTION:
                fprintf(out, "None\n");
                break;
            case SET_PRIMARY_ALONE:
                fprintf(out, "Set primary alone\n");
                break;
            case SWITCH_SEC_TO_PRI:
                fprintf(out, "Switch secondary to primary\n");
                break;
            case BOTH_SERVERS_DOWN:
                fprintf(out, "Both down: do nothing\n");
                break;
            case BOTH_SERVERS_SECONDARY:
                fprintf(out, "Both are secondary; set one to primary\n");
                break;
            case PRI_CONN_DOWN:
                if (config.Pessimistic) {
                    fprintf(out, "Primary connection down: pessimistic mode; switch secondary to primary\n");
                } else {
                    fprintf(out, "Primary connection down: do nothing\n");
                }
                break;
            case SEC_CONN_DOWN:
                fprintf(out, "Secondary connection down: do nothing\n");
                break;
            case HSB_CONN_DOWN:
                fprintf(out, "HSB down: reconnect HSB\n");
                break;
            case NON_VALID_HSB:
                fprintf(out, "Non-valid HSB: no action\n");
                break;
            case IMPOSSIBLE:
                fprintf(out, "THIS SHOULD NOT HAPPEN!\n");
                break;
        }
}


/*#***********************************************************************\
 *
 *      check_hsb_servers
 *
 * Get the hotstandby server status and return the watchdog action.
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby server info
 *
 *      err_msg - out
 *          error message.
 *
 *      checked_servers - in
 *          which servers' status to check (SRV_BOTH only)
 *
 * Return value:
 *      WDACTION that should be performed by the watchdog.
 *
 * Limitations:
 *
 * Globals used:
 */
static WDACTION check_hsb_servers(hsb_servers_t* hsb_servers, uchar* err_msg,
                                  int checked_servers)
{
        int i = 0;
        err_msg[0] = '\0';
        
        get_hsb_servers_status(hsb_servers, checked_servers);
        for(i = 0; i < 3; i++){
            /* Try to get status of servers 3 times if no responses (was 5) */
            if (SRV1_ROLE(hsb_servers) == HSB_NO_RESPONSES && 
                SRV2_ROLE(hsb_servers) == HSB_NO_RESPONSES) {
                mysleep(5);
                if (debug >= 2) {
                    printf("check_hsb_servers: no responses: repeat\n");
                }
                get_hsb_servers_status(hsb_servers, checked_servers);
            } else {
                break;
            }
        }
    
        if (SRV1_ROLE(hsb_servers) == HSB_NO_RESPONSES && 
            SRV2_ROLE(hsb_servers) == HSB_NO_RESPONSES) {

            sprintf(err_msg, 
                    "Both servers '%s' and '%s' are down.",
                    SRV1_CONNECT(hsb_servers),
                    SRV2_CONNECT(hsb_servers));

            return BOTH_SERVERS_DOWN;

        } else if (SRV1_ROLE(hsb_servers) == HSB_STANDALONE) { 
            sprintf(err_msg, 
                    "Server '%s' is a Standalone server.\nWatchdog is used for HotStandby servers only.", 
                    SRV1_CONNECT(hsb_servers));

            return NON_VALID_HSB;

        } else if (SRV2_ROLE(hsb_servers) == HSB_STANDALONE) { 
            sprintf(err_msg, 
                    "Server '%s' is a Standalone server.\nWatchdog is used for HotStandby servers only.", 
                    SRV2_CONNECT(hsb_servers));

            return NON_VALID_HSB;

        } else if (SRV1_ROLE(hsb_servers) == HSB_PRIMARY &&
                   SRV2_ROLE(hsb_servers) == HSB_PRIMARY) {
            sprintf(err_msg, 
                    "Both servers '%s' and '%s' are Primary.\nSwitch one of the servers to Secondary.",
                    SRV1_CONNECT(hsb_servers),
                    SRV2_CONNECT(hsb_servers));

            {
                /* Check for PRIMARY UNCERTAINs.  One possible
                 * scenario is that primary doesn't crash, networks
                 * fail from Watchdog (WD) to primary and from
                 * secondary to primary.  Old primary goes to PRIMARY
                 * UNCERTAIN and WD sets secondary to be the new
                 * primary.  Then when old primary comes back
                 * online...
                 *
                 * Here WD could automatically set uncertain primary
                 * (old primary) to SECONDARY ALONE, but we'll leave
                 * it to the DBA.  Even though we DO handle the case
                 * of two secondaries.
                 *
                 * Availability Guide says that WD handles
                 * this.. Maybe it should.
                 */
                int res = 0;
                uchar tbuf[MAX_STRLEN];
                
                res = (SRV1_STATE(hsb_servers) == STATE_PRIMARY_UNCERTAIN)
                    ? 0x01: 0;
                res |= (SRV2_STATE(hsb_servers) == STATE_PRIMARY_UNCERTAIN)
                    ? 0x02: 0;
                switch (res) {
                    case 3:
                        sprintf(tbuf,
                                "  Both servers are PRIMARY UNCERTAIN.\n");
                        strcat(err_msg, tbuf);
                        break;
                    case 2:
                    case 1:
                        sprintf(tbuf, "  But server %d: '%s' is PRIMARY UNCERTAIN; Server %d is recommended for new primary.\n", res,
                                hsb_servers->srv_info[res-1].connect,
                                3 - res);
                        strcat(err_msg, tbuf);
                        break;
                    default:
                        assert(1);
                        break;
                }
                        
            }
                
            return NON_VALID_HSB;

        } else if (SRV1_ROLE(hsb_servers) == HSB_SECONDARY &&
                   SRV2_ROLE(hsb_servers) == HSB_SECONDARY) {
            sprintf(err_msg, 
                    "Both servers '%s' and '%s' are Secondary.\nSwitch one of the servers to Primary",
                    SRV1_CONNECT(hsb_servers),
                    SRV2_CONNECT(hsb_servers));

            return BOTH_SERVERS_SECONDARY;

        } else if (SRV1_ROLE(hsb_servers) == HSB_PRIMARY ||
                   SRV2_ROLE(hsb_servers) == HSB_SECONDARY) {
            hsb_servers->primary   = 0;
            hsb_servers->secondary = 1;
            /* OK, no return */
            
        } else if (SRV1_ROLE(hsb_servers) == HSB_SECONDARY ||
                   SRV2_ROLE(hsb_servers) == HSB_PRIMARY) {
            hsb_servers->primary   = 1;
            hsb_servers->secondary = 0;
            /* OK, no return */

        } else {
            /* shouldn't get here */
            return IMPOSSIBLE;
        }

        /* check the status */
        if (PRIMARY_STATUS(hsb_servers) == STATUS_CONNECTED && 
            SECONDARY_STATUS(hsb_servers) == STATUS_CONNECTED) {

            return NOACTION;

        } else if (PRIMARY_STATUS(hsb_servers) == STATUS_CATCHUP && 
                   SECONDARY_STATUS(hsb_servers) == STATUS_CATCHUP) {

            return NOACTION;

        } else if ((PRIMARY_STATUS(hsb_servers) == STATUS_CONNECTED ||
                    PRIMARY_STATUS(hsb_servers) == STATUS_CATCHUP) &&
                   SECONDARY_STATUS(hsb_servers) == STATUS_NO_RESPONSES) {

            sprintf(err_msg, 
                    "Connection between Secondary '%s' and Watchdog is broken.", 
                    SEC_CONNECT(hsb_servers));

            return SEC_CONN_DOWN;

        } else if ((SECONDARY_STATUS(hsb_servers) == STATUS_CONNECTED ||
                    SECONDARY_STATUS(hsb_servers) == STATUS_CATCHUP) &&
                   PRIMARY_STATUS(hsb_servers) == STATUS_NO_RESPONSES) {

            sprintf(err_msg, 
                    "Connection between Primary '%s' and Watchdog is broken.", 
                    PRI_CONNECT(hsb_servers));

            return PRI_CONN_DOWN;

        } else if (PRIMARY_STATUS(hsb_servers) == STATUS_BROKEN && 
                   SECONDARY_STATUS(hsb_servers) == STATUS_NO_RESPONSES
                   && (hsb_servers->srv_status[hsb_servers->primary].state
                       != STATE_PRIMARY_ALONE) ) {

            sprintf(err_msg, "Secondary '%s' is down.", 
                    SEC_CONNECT(hsb_servers));

            return SET_PRIMARY_ALONE;

        } else if (PRIMARY_STATUS(hsb_servers) == STATUS_NO_RESPONSES && 
                   SECONDARY_STATUS(hsb_servers) == STATUS_BROKEN) {

            sprintf(err_msg, "Primary '%s' is down.", 
                    PRI_CONNECT(hsb_servers));

            return SWITCH_SEC_TO_PRI;

        } else if (PRIMARY_STATUS(hsb_servers) == STATUS_BROKEN &&
                   SECONDARY_STATUS(hsb_servers) == STATUS_BROKEN) {
            /* both know that the connection is broken. if one is broken and
             * the other one is connected or catchup, then we wait for servers
             * to realise their condition (handled with NOACTION, at end of
             * function) */
            sprintf(err_msg, 
                    "Connection between Primary '%s' and Secondary '%s' is broken.", 
                    PRI_CONNECT(hsb_servers), 
                    SEC_CONNECT(hsb_servers));

            return HSB_CONN_DOWN;
        }
        
        return NOACTION;
}

/*#***********************************************************************\
 *
 *      process_line
 *
 * Extract the key and key value from a line of the ini file.
 *
 * Parameters:
 *      buf - in, use
 *          the line to process
 *
 *      section - out
 *          section of the config
 *
 *      keyname - out
 *          key of the config
 *
 *      keyvalue - out
 *          keyvalue of the config
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *
 * Limitations:
 *
 * Globals used:
 */
static int process_line(uchar* buf, uchar* section,
                        uchar* keyname, uchar* keyvalue)
{
        *keyname  = '\0';
        *keyvalue = '\0';

        buf = skip_spaces(buf);
        if (*buf == '\0' || *buf == '#' || *buf == ';') {
            /* comment or empty */
            return 0;
        }

        if (*buf == '[') { /* section ("[foo]") */
            buf++;
            while (*buf != '\0' && *buf != ']') {
                *section++ = *buf++;
            }
            *section = '\0';
            return 0;
        }

        /* it must be a key(name) then ("foo=bar") */
        while (*buf != '\0' && *buf != '=' && *buf != '#' && !isspace(*buf)) {
            *keyname++ = *buf++;
        }
        *keyname = '\0';

        buf = skip_spaces(buf);
        if (*buf == '=') {
            buf++;
        } else {
            return -1;
        }

        /* then the value */
        buf = skip_spaces(buf);
        while (*buf != '\0' && *buf != '\n' && *buf != '#') {
            *keyvalue++ = *buf++;
        }
        *keyvalue = '\0';

        return 0;
}

/*#***********************************************************************\
 *
 *      init_config
 *
 * Set the default calue for config of the watchdog.
 *
 * Parameters:
 *      config - out, use
 *          config of the watchdog.
 *
 * Return value:
 *
 * Limitations:
 *
 * Globals used:
 */
static void init_config(config_t* config)
{
        config->NumRetry     = 0;
        config->PingInterval = 1000;
        config->AutoSwitch   = TRUE;
        config->Pessimistic  = FALSE;
        strcpy(config->logfilename, "Watchdog.log");
        config->DualSecAutoSwitch = FALSE; /* TRUE is dangerous */
        config->ObserveOnly  = FALSE;
        /* not read from inifile yet */
        config->LoneTrigger  = FALSE;
}

/*#***********************************************************************\
 *
 *      strmatch
 *
 * Checks if two strings match. CASE-INSENSITIVE.
 *
 * Parameters:
 *      s1 - in, use
 *          First string
 *
 *      s2 - in, use
 *          Second string.
 *
 * Return value:
 *      1, if strings match.
 *      0, if they don't match.
 *
 * Limitations:
 *      Case-insensitive.
 *
 * Globals used:
 */
static int strmatch(uchar* s1, uchar* s2) 
{
        while (*s1 != '\0' && *s2 != '\0') {
            if (tolower(*s1++) != tolower(*s2++)) {
                return 0;
            }
        }

        if (*s1 == '\0' && *s2 == '\0') {
            return 1;
        } 
        return 0;
}

/*#***********************************************************************\
 *
 *      read_inifile
 *
 * Process the inifile and setup the hotstandby server info and watchdog config
 *
 * Parameters:
 *      hsb_servers - out
 *          hotstandby server info
 *
 *      config - out
 *          watchdog config
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *
 * Limitations:
 *
 * Globals used:
 */
static int read_inifile(hsb_servers_t* hsb_servers, config_t* config)
{
        FILE *infp = NULL;
        uchar buf[MAX_BUFLEN];
        uchar buf2[MAX_BUFLEN];
        uchar section[MAX_BUFLEN];
        uchar keyname[MAX_BUFLEN];
        uchar keyvalue[MAX_BUFLEN];

        uchar connect1[MAX_STRLEN];
        uchar username1[MAX_STRLEN];
        uchar password1[MAX_STRLEN];
        uchar connect2[MAX_STRLEN];
        uchar username2[MAX_STRLEN];
        uchar password2[MAX_STRLEN];

        connect1[0]  = '\0';
        username1[0] = '\0';
        password1[0] = '\0';
        connect2[0]  = '\0';
        username2[0] = '\0';
        password2[0] = '\0';
        init_config(config);

        infp = fopen("solid.ini", "r");
        if (infp == NULL) {
            printf("Cannot open file 'solid.ini'\n");
            return -1;
        }

        section[0] = '\0';
        while (fgets(buf, MAX_BUFLEN, infp)) {
            bool line_continues;
            int rc;
            int sl;
            
            do { /* handle continuing line ie. line ends with \ + enter */
                sl = strlen(buf) - 1;
                line_continues = FALSE;
                
                if (buf[sl] != '\n' && buf[sl] != '\r') { /* last line if any*/
                    break; /* can't continue */
                }
                while (sl > 0 && (buf[sl] == '\n' || buf[sl] == '\r')) { /* skip */
                    sl--;
                }
                if (buf[sl] == '\\') {
                    line_continues = TRUE;
                    buf[sl] = '\0';
                    if (fgets(buf2, MAX_BUFLEN, infp) == NULL) {
                        printf("Error: Inifile line continues but file ends.\n");
                        fclose(infp);
                        return -1;
                    }
                    if( (sl + strlen(buf2) + 1) > MAX_BUFLEN) {
                        printf("Error: Continued inifile line is too long (longer than %d).\n", MAX_BUFLEN);
                        fclose(infp);
                        return -1;
                    }
                    strcat(buf, buf2);
                }
            } while (line_continues);
                
            rc = process_line(buf, section, keyname, keyvalue);

            if (*keyname == '\0') {
                continue;
            }

            /* skip all the other sections */
            if (!strmatch(section, "Watchdog"))  {
                continue;
            }
            /* check if syntax in our section was ok */
            if (rc != 0) {
                buf[sl+1] = '\0'; /* kill newline */
                printf("Error: Invalid line '%s' in solid.ini.\n", buf);
                fclose(infp);
                return -1;
            }

            if (strmatch(keyname, "Connect1")) {
                strcpy(connect1, keyvalue);

            } else if (strmatch(keyname, "Username1")) {
                strcpy(username1, keyvalue);

            } else if (strmatch(keyname, "Password1")) {
                strcpy(password1, keyvalue);

            } else if (strmatch(keyname, "Connect2")) {
                strcpy(connect2, keyvalue);

            } else if (strmatch(keyname, "Username2")) {
                strcpy(username2, keyvalue);

            } else if (strmatch(keyname, "Password2")) {
                strcpy(password2, keyvalue);

            } else if (strmatch(keyname, "NumRetry")) {
                config->NumRetry = atoi(keyvalue);

            } else if (strmatch(keyname, "PingInterval")) {
                config->PingInterval = atoi(keyvalue);

            } else if (strmatch(keyname, "AutoSwitch")) {
                config->AutoSwitch = strmatch(keyvalue, "Yes") 
                    ? TRUE : FALSE;

            } else if (strmatch(keyname, "WatchdogLog")) {
                strcpy(config->logfilename, keyvalue);

            } else if (strmatch(keyname, "Pessimistic")) {
                config->Pessimistic = strmatch(keyvalue, "Yes") ? TRUE : FALSE;

            } else if (strmatch(keyname, "ObserveOnly")) {
                config->ObserveOnly = strmatch(keyvalue, "Yes") ? TRUE : FALSE;

            } else if (strmatch(keyname, "DualSecAutoSwitch")) {
                config->DualSecAutoSwitch = strmatch(keyvalue, "Yes")
                    ? TRUE : FALSE;

            } else {
                printf("Error: Unknown parameter %s in solid.ini.\n", keyname);
                fclose(infp); /* not very tolerant.. might warn only */
                return -1;
            }
        }
        fclose(infp);

        /* check required parameters */
        if (connect1[0] == '\0') {
            printf("Error: Parameter 'connect1' is not set in solid.ini.\n");
            return -1;
        }
        if (connect2[0] == '\0') {
            printf("Error: Parameter 'connect2' is not set in solid.ini.\n");
            return -1;
        }

        /* server 1 */
        if (username1[0] == '\0') {
            printf("Enter the username for server '%s':\n", connect1);
            fgets(username1, MAX_STRLEN - 1, stdin);
            trim_spaces(username1);
        }
        if (password1[0] == '\0') {
            printf("Enter the password for server '%s':\n", connect1);
            fgets(password1, MAX_STRLEN - 1, stdin);
            trim_spaces(password1);
        }
        /* server 2 */
        if (username2[0] == '\0') {
            printf("Enter the username for server '%s':\n", connect2);
            fgets(username2, MAX_STRLEN - 1, stdin);
            trim_spaces(username2);
        }
        if (password2[0] == '\0') {
            printf("Enter the password for server '%s':\n", connect2);
            fgets(password2, MAX_STRLEN - 1, stdin);
            trim_spaces(password2);
        }

        init_hsb_servers(hsb_servers, connect1, username1, password1, 
                         connect2, username2, password2);
        return 0;
}

/*#***********************************************************************\
 *
 *      watchdog_set_pri_alone
 *
 * Sets the primary to primary alone mode.
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby server info
 *
 *      autoswitch - in
 *          whether to autoswitch primary to alone, or just inform
 *          DBA about it.
 *
 *      outfile - out
 *          output file.
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *
 * Limitations:
 *
 * Globals used:
 */
static int watchdog_set_pri_alone(hsb_servers_t* hsb_servers, bool autoswitch, 
                                  FILE* outfile)
{
        int rc = 0;

        if (!autoswitch) {
            fprintf(outfile, 
                    "No Action from Watchdog (AutoSwitch is not set).\n"); 
            fprintf(outfile, 
                    "DBA Action: Set Primary '%s' to primary alone mode.\n", 
                    PRI_CONNECT(hsb_servers));
        } else { 
            fprintf(outfile, 
                    "Watchdog Action: Set Primary '%s' to primary alone mode...",
                    PRI_CONNECT(hsb_servers));
            fflush(outfile);

            rc = set_primary_alone(hsb_servers);
            fprintf(outfile, "%s (%d)\n", (rc != 0) ? "FAILED" : "SUCCESS",
                    rc);
        }
        return rc;
}

/*#***********************************************************************\
 *
 *      watchdog_switch_sec_to_pri
 *
 * Switch secondary to primary and set primary to primary alone mode.
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby server info
 *
 *      autoswitch - in
 *          whether to autoswitch secondary to primary or just inform DBA.
 *
 *      outfile - out
 *          output file.
 *
 * Return value:
 *      0, if OK
 *      -1, if failed.
 *
 * Limitations:
 *
 * Globals used:
 */
static int watchdog_switch_sec_to_pri(hsb_servers_t* hsb_servers,
                                      bool autoswitch, FILE* outfile)
{
        int rc = 0;

        if (!autoswitch) {

            fprintf(outfile, 
                    "No Action from Watchdog (AutoSwitch is not set).\n"); 
            fprintf(outfile, 
                    "DBA Action: Switch Secondary '%s' to primary AND ",
                    SEC_CONNECT(hsb_servers));
            fprintf(outfile, 
                    "Set the NEW Primary to primary alone mode.\n");
        } else {

            fprintf(outfile, 
                    "Watchdog Action: Switch Secondary '%s' to primary...",
                    SEC_CONNECT(hsb_servers));
            fflush(outfile);

            rc = switch_sec_to_pri(hsb_servers);
            fprintf(outfile, "%s (%d)\n", (rc) ? "FAILED" : "SUCCESS", rc);

        }

        return rc;
}

/*#***********************************************************************\
 *
 *      watchdog_switch_one_to_pri
 *
 * Switches one of many (2) secondaries to primary.
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby server info
 *
 *      autoswitch - in
 *          if we are to actually do the switch or just inform DBA.
 *
 *      outfile - in, use
 *          logfile
 *
 * Return value:
 *      0, if ok.
 *      -1, if an error occured.
 *
 * Limitations:
 *
 * Globals used:
 */
static int watchdog_switch_one_to_pri(hsb_servers_t* hsb_servers,
                                      bool autoswitch, FILE* outfile)
{
        int switch_roles = 0;
        int rc = 0;

        if (!autoswitch) {
            fprintf(outfile, 
                    "No Action from Watchdog (DualSecAutoSwitch is not set).\n"); 
            fprintf(outfile, 
                    "DBA Action: Switch one Secondary to primary.\n");
        } else {
            if (0) {
                /* In this case always switch old secondary to primary. We swap
                 * primary and secondary indices.
                 */              
                switch_roles = 1;
            } else {
                /* In this case we ask log positions from both
                 * servers and select one that has more transactions.
                 */
#if 0
                uchar rcstate[6];
                uchar local_logpos_text[MAX_STRLEN];
                uchar remote_logpos_text[MAX_STRLEN];
                rc = exec_admin_command(PRIMARY_SERVER(hsb_servers),
                                        "hsb logpos local", rcstate,
                                        local_logpos_text);
                if (rc == 0) {
                    rc = exec_admin_command(SECONDARY_SERVER(hsb_servers),
                                            "hsb logpos remote", rcstate,
                                            remote_logpos_text);
                }
                if (rc == 0) {
                    switch_roles = strcmp(local_logpos_text,
                                          remote_logpos_text) < 0;
                }

                if (!switch_roles) {
                    rc = exec_admin_command(PRIMARY_SERVER(hsb_servers),
                                            "hsb logpos remote", rcstate,
                                            remote_logpos_text);
                    if (rc == 0) {
                        rc = exec_admin_command(SECONDARY_SERVER(hsb_servers),
                                                "hsb logpos local", rcstate,
                                                local_logpos_text);
                    }
                    if (rc == 0) {
                        switch_roles = strcmp(local_logpos_text,
                                              remote_logpos_text) < 0;
                    }
                }
#else /* new logpos command */
                uchar pri_logpos[MAX_STRLEN];
                uchar sec_logpos[MAX_STRLEN];

                rc = exec_admin_command(PRIMARY_SERVER(hsb_servers),
                                        "hsb logpos", NULL, pri_logpos);
                if (rc == 0) {
                    rc = exec_admin_command(SECONDARY_SERVER(hsb_servers),
                                            "hsb logpos", NULL, sec_logpos);
                }

                if (rc == 0) {
                    switch_roles = strcmp(pri_logpos, sec_logpos) < 0;
                }
#endif /* ! old logpos */
            }
            
            fprintf(outfile, 
                    "Watchdog Action: Switch one Secondary, '%s', to primary...",
                    (switch_roles)? SEC_CONNECT(hsb_servers):
                    PRI_CONNECT(hsb_servers));
            fflush(outfile);

            if (switch_roles) {
                rc = switch_sec_to_pri(hsb_servers);
                fprintf(outfile, "%s (%d)\n",
                        (rc != 0) ? "FAILED" : "SUCCESS", rc);
            } else {
                rc = set_primary_alone(hsb_servers);
                fprintf(outfile, "%s (%d)\n",
                        (rc != 0) ? "FAILED" : "SUCCESS", rc);
            }                
        }

        return rc;
}

/*#***********************************************************************\
 *
 *      watchdog_hsb_connect
 *
 * connect hsb servers to each other.
 *
 * Parameters:
 *      hsb_servers - in, use
 *          hotstandby servers to connect.
 *
 *      outfile - in, use
 *          logfile
 *
 * Return value:
 *      0, if ok
 *      -1, if couldn't connect. 
 *      or error code of the connect call.
 *
 * Limitations:
 *
 * Globals used:
 */
static int watchdog_hsb_connect(hsb_servers_t* hsb_servers, FILE* outfile)
{
        int rc = 0;
        uchar text[MAX_BUFLEN] = "<none>";

        fprintf(outfile, 
                "Watchdog Action: Connect Primary '%s' to Secondary '%s'...",
                PRI_CONNECT(hsb_servers), 
                SEC_CONNECT(hsb_servers));
        fflush(outfile);

        rc = hsb_connect(hsb_servers, text);
        fprintf(outfile, "%s (%d: %s)\n", (rc != 0) ? "FAILED" : "SUCCESS",
                rc, text);

        return rc;
}

/*#***********************************************************************\
 *
 *      usage
 *
 * display usage.
 *
 * Parameters:
 *      argv - in, use
 *          argv given to main.
 *
 * Return value:
 *      Exits with a 1.
 *
 * Limitations:
 *
 * Globals used:
 */
static void usage(char* argv[])
{
        printf("Usage: %s [<option>]...\n", argv[0]);
        printf("\t-a           autoswitch: switch primary uncertain to primary alone\n");
        printf("\t             and a lonely secondary to primary alone.\n");
        printf("\t-s           autoswitch one of two secondaries to primary.\n");
        printf("\t-t <time>    set running time of watchdog in seconds.\n");
        printf("\t-o           observe only; take no action.\n");
        printf("\t-p           pessimistic mode; expect server crash to be the\n");
        printf("\t             cause of connection failure rather than the network.\n");
        printf("\n\tDebug flags:\n\n");
        printf("\t-d           debug on. Add d's to up debug level (eg. -dd).\n");
        printf("\t-l           lone trigger: quit when there's a lone primary alone\n");
        printf("\t             server. (for use with timestamping.)\n");
        printf("\n");
        exit(1);
}

#if defined(SS_VXW)

#  define main watchdog_main
#  define SS_CDECL 
int watchdog_main(uint argc, char** argv);
void watchdog(){
        char* args[2];

        args[0] = "watchdog";
        args[1] = NULL;
        watchdog_main(1, args);

        return;
}

#endif /* SS_VXW */

/*##**********************************************************************\
 *
 *      main
 *
 * Watchdog program.
 *
 * Parameters:
 *      argc - in
 *          number of arguments
 *
 *      argv - in, use
 *          commandline arguments.
 *
 * Return value:
 *      0, if OK.
 *      -1, on error.
 *
 * Limitations:
 *
 * Globals used:
 */
int SS_CDECL main(int argc, char* argv[])
{
        hsb_servers_t hsb_servers;
        server_status_t old_status[2];
        WDACTION wdaction;
        WDACTION lastwdaction = NOACTION;
        int num_of_tries = 0;/* how many times we have tried to connect/ping */
        uchar err_msg[MAX_BUFLEN];
        uchar* oldprimary;
        uchar* oldsecondary;
        int recheck = 0;  /* if states need to be checked immediately again */
        int i;
        int rc;
        bool doublecheck = FALSE; /* in some cases it is good to doublecheck
                                   * the state before action.  Reason is that
                                   * the state may change after one server has
                                   * been queried and then WD won't have a
                                   * correct picture of the situation.
                                   * eg. switch can lead to "secondary active -
                                   * secondary active" false state.
                                   *
                                   * This means that one should check that the
                                   * state of the first queried server is the
                                   * same on second query too.. (not done yet)
                                   *
                                   */
        
        FILE* outfile;         /* logfile */
        int mode;              /* watchdog mode */
        int runtime = 0;  /* time for watchdog to run (s). 0 is no_limit. */
        time_t start_time = time(NULL);
        time_t time_now;
        time_t last_apa_check    = 0; /* apa = AutoPrimaryAlone */
        time_t last_display_time = 0; /* of server states */
        bool status_changed;

        /* configure watchdog */
        
        if (read_inifile(&hsb_servers, &config) != 0) {
            return (-1);
        }

        for (i = 1; i < argc; i++) {
            if (strncmp(argv[i], "-d", 2) == 0)  {
                char* p;
                if (debug == 0) {
                    debug = 1; /* debug info */
                } else {
                    debug++;
                }
                /* multiple d's increase debug level. eg: '-ddd' -> debug = 3
                 * or '-d -d' -> debug = 2 */
                p = &(argv[i][2]);
                while (*p) {
                    if (*p == 'd') {
                        debug++;
                        warn = 1; /* debug implies warn */
                    } else {
                        usage(argv);
                    }
                    p++;
                }
            } else if (strcmp(argv[i], "-t") == 0) {
                /* set lifespan of watchdog in seconds */
                i++;
                if (argv[i] != NULL) {
                    runtime = atoi(argv[i]);
                } else {
                    usage(argv);
                }
            } else if (strcmp(argv[i], "-p") == 0) {
                /* be pessimistic; assume servers crash more often than
                 * network fails. Faster reaction to server failure. May
                 * cause false alarms and unnecessary switches. */
                config.Pessimistic = TRUE;
            } else if (strcmp(argv[i], "-a") == 0) {
                config.AutoSwitch = TRUE;
            } else if (strcmp(argv[i], "-s") == 0) {
                config.DualSecAutoSwitch = TRUE;
            } else if (strcmp(argv[i], "-o") == 0) {
                config.ObserveOnly = TRUE;
            } else if (strcmp(argv[i], "-l") == 0) {
                config.LoneTrigger = TRUE;
            } else {
                usage(argv);
            }
            /*
             * Old used options: -w (warn)
             */
        }

        outfile = fopen(config.logfilename, "w");
        if (outfile == NULL) {
            printf("Cannot open file '%s'.\n", config.logfilename);
            return (-1);
        }

        if (debug) { /* print config */
            char tmp[40];
            printf("Watchdog config:\n");
            /* this is getting long, will wrap soon */
            printf("\tFlags: Debug%s%s%s%s%s%s%s\n",
                   (config.Pessimistic)
                   ? ", Pessimistic" : "",
                   (runtime > 0)
                   ? (sprintf(tmp, ", Time: %ds", runtime), tmp): "",
                   (config.ObserveOnly)
                   ? ", ObserveOnly": "",
                   (config.AutoSwitch)
                   ? ", AutoSwitch": "",
                   (config.DualSecAutoSwitch)
                   ? ", DualSecAutoSwitch": "",
                   (config.LoneTrigger)
                   ? ", LoneTrigger": "",
#ifdef WD_TIMING
                   ", Timestamps"
#else
                   ""
#endif
                   );
            printf("\tPing Interval: %dms\n", config.PingInterval);
            printf("\n");
        }

        if (!config.AutoSwitch && config.Pessimistic) {
            printf("Warning: Pessimistic is set but AutoSwitch is not.\n");
            printf("\tMeaning: Pessimistic will not work.\n");
        }
        
        old_status[0].state = STATE_UNSET;
        
        mode = NORMAL_MODE;
        while (1) { /* main loop */
            time_now = time(NULL);

            /* Get the server status and watchdog action */
            wdaction = check_hsb_servers(&hsb_servers, err_msg, SRV_BOTH);

            if (config.AutoSwitch 
                && (time_now > (last_apa_check + APA_INTERVAL)) ) {

                /* AutoPrimaryAlone and AutoSwitch is dangerous. */
                last_apa_check = time_now;
                if ((SRV1_STATE(&hsb_servers) != STATE_NOT_HSB_SERVER) &&
                    server_isautoprimaryalone(&(hsb_servers.srv_info[0]))) {
                    printf("Warning! Server '%s' is in AutoPrimaryAlone mode\nDisabling AutoSwitch\n\n",
                           SRV1_CONNECT(&hsb_servers));
                    config.AutoSwitch = FALSE;
                }
                if ((SRV2_STATE(&hsb_servers) != STATE_NOT_HSB_SERVER) &&
                    server_isautoprimaryalone(&(hsb_servers.srv_info[1]))) {
                    printf("Warning! Server '%s' is in AutoPrimaryAlone mode\nDisabling AutoSwitch.\n\n",
                           SRV2_CONNECT(&hsb_servers));
                    config.AutoSwitch = FALSE;
                }
            }
            
            /* check if something has happened */
            status_changed = FALSE;
            if ((   old_status[0].state != hsb_servers.srv_status[0].state)
                || (old_status[0].role  != hsb_servers.srv_status[0].role)
                || (old_status[1].state != hsb_servers.srv_status[1].state)
                || (old_status[1].role  != hsb_servers.srv_status[1].role)) {

                status_changed = TRUE;

                old_status[0] = hsb_servers.srv_status[0];
                old_status[1] = hsb_servers.srv_status[1];                
            }
                
            if (status_changed ||
                (time_now > (last_display_time + MAX_DISPLAY_INTERVAL)) ) {

                printf("\n");
                print_hsb_servers_status(&hsb_servers, stdout);
                if (config.ObserveOnly) {
                    printf("Watchdog in ObserveOnly mode.\n");
                }
                if (debug >= 1) {
                    printf("\tWDACTION = %d: %s, mode = %s\n", wdaction,
                           wdaction_strings[wdaction],
                           mode == NORMAL_MODE ? "NORMAL": "FAILURE");
                }
                last_display_time = time_now;
                status_changed = TRUE; /* make everyone print now */
#ifdef WD_TIMING
                print_timestamp();
#endif            
            }

            if (!config.ObserveOnly && (wdaction != lastwdaction)
                && (mode == NORMAL_MODE)) {

                printf("\n");
                print_wdaction(wdaction, stdout);
                printf("\n");
            }

            if (config.LoneTrigger && server_alone(&hsb_servers)) {
                printf("\nWatchdog exit: lone trigger.\n");
                return 0;
            }

            if (!config.ObserveOnly) {
                if (wdaction == NOACTION) {
                    num_of_tries = 0;
                    /*
                     * watchdog will only switch back to normal after DBA has
                     * brought both servers up and connected.
                     */
                    if (mode == FAILURE_MODE) {
                        fprintf(outfile,
                                "\nWatchdog goes back to Normal mode.\n\n");
                        mode = NORMAL_MODE;
                    }
                } else if (mode == NORMAL_MODE) {
                    
                    /* Retry connecting if limit is not reached */
                    num_of_tries++;
                    if (num_of_tries <= config.NumRetry) {
                        mysleep(config.PingInterval);
                        printf("Retry #%d...\n", num_of_tries); 
                        continue; /* restart main loop */
                    }
                    
                    /* if action changed, print the server status to log */
                    /* this makes logs quite short. use status_changed
                     * to trigger more prints (might be more useful
                     * :-) )
                     */
                    if (wdaction != lastwdaction) {
                        print_hsb_servers_status(&hsb_servers, outfile);
                        fprintf(outfile, "%s\n", err_msg);
                    }
                    
                    switch (wdaction) {
                        case SET_PRIMARY_ALONE: /* secondary has died */
                            oldsecondary = SEC_CONNECT(&hsb_servers);
                            
                            watchdog_set_pri_alone(&hsb_servers, 
                                                   config.AutoSwitch, outfile);
                            
                            fprintf(outfile, 
                                    "DBA Action: Restart old Secondary '%s'.\n", 
                                    oldsecondary);
                            mode = FAILURE_MODE;
                            recheck = 1;
                            break;
                            
                        case SWITCH_SEC_TO_PRI: /* primary has died */
                            oldprimary = PRI_CONNECT(&hsb_servers);
                            
                            watchdog_switch_sec_to_pri(&hsb_servers, 
                                                       config.AutoSwitch,
                                                       outfile);
                            
                            fprintf(outfile, 
                                    "DBA Action: Restart old Primary '%s'.\n", 
                                    oldprimary);
                            mode = FAILURE_MODE;
                            break;
                            
                        case HSB_CONN_DOWN:
                            
                            /* 
                             * Set primary to primary alone first
                             * before reconnect because the primary
                             * can be in an unknown state.
                             *
                             * if determining of action was optimized
                             * then set_pri_alone could be eliminated:
                             * SET_PRIMARY_ALONE should have priority
                             * over HSB_CONN_DOWN.
                             *
                             */
                            watchdog_set_pri_alone(&hsb_servers, 
                                                   config.AutoSwitch, outfile);
                            
                            rc = watchdog_hsb_connect(&hsb_servers, outfile);
                            if (rc != 0) {                                
                                if (rc == 14524 || rc == 14525) {
                                    fprintf(outfile,
                                            "DBA Action: Databases are out of sync!\n");
                                } else {
                                    fprintf(outfile, 
                                            "DBA Action: Fix the network connection between Primary and Secondary.\n");
                                }
                                mode = FAILURE_MODE;
                            }
                            break;
                            
                        case PRI_CONN_DOWN:
                            
                            if (config.Pessimistic) {
                                /* assume primary has crashed even
                                 * though the cause may be network
                                 * failure.
                                 */
                                oldprimary = PRI_CONNECT(&hsb_servers);
                                fprintf(outfile,
                                        "Pessimistic mode: Primary may be down, switching servers\n");                            
                                watchdog_switch_sec_to_pri(&hsb_servers, 
                                                           config.AutoSwitch,
                                                           outfile);
                                /* maybe Pessimistic should imply AutoSwitch */
                                
                                fprintf(outfile, 
                                        "DBA Action: Check old Primary '%s'.\n", 
                                        oldprimary);
                                /*mode = FAILURE_MODE;*/
                                
                            } else if (wdaction != lastwdaction) {
                                /* we can't fix the connection, so we
                                 * do nothing */
                                fprintf(outfile,
                                        "No Action from the Watchdog.\n");
                                fprintf(outfile, 
                                        "DBA Action: Fix the network connection between Primary and Watchdog.\n\n");
                            }
                            break;
                            
                        case SEC_CONN_DOWN:
                            /* in pessimistic mode watchdog might want to
                             * set primary alone as quickly as possible.
                             * BUT since primary works just fine before it
                             * realizes that secondary is down, this is
                             * not needed.
                             */
                            if (wdaction != lastwdaction) {
                                fprintf(outfile,
                                        "No Action from the Watchdog.\n");
                                fprintf(outfile, 
                                        "DBA Action: Fix the network connection between Secondary and Watchdog.\n\n");
                            }
                            break;
                            
                        case BOTH_SERVERS_DOWN:
                            fprintf(outfile, "No Action from the Watchdog.\n");
                            fprintf(outfile, 
                                    "DBA Action: Restart both Primary and Secondary servers.\n");
                            mode = FAILURE_MODE;
                            break;
                            
                        case BOTH_SERVERS_SECONDARY:
                            if (config.DualSecAutoSwitch == FALSE) {
                                break; /* prevent short circuiting */
                            }
                            if (doublecheck) {
                                watchdog_switch_one_to_pri(&hsb_servers, 
                                                           config.DualSecAutoSwitch,
                                                           outfile);
                                doublecheck = FALSE;
                            } else {
                                doublecheck = TRUE;
                            }
                            recheck = 1;
                            break;
                        case NON_VALID_HSB:
                            fprintf(outfile, "No action from the Watchdog.\n");
                            fprintf(outfile, 
                                    "DBA Action: invalid HSB configuration. ");
                            fprintf(outfile,
                                    "Re-config one server to be Primary and the other to be Secondary.\n");
                            mode = FAILURE_MODE;
                            break;
                        case IMPOSSIBLE:
                            fprintf(outfile, "No action from the Watchdog.\n");
                            fprintf(outfile, "DBA Action: an impossible state happened in the Watchdog, contact Solid.\n");
                        default:
                            break;
                    }
                    
                    if (mode == FAILURE_MODE) { /* if mode was changed */
                        fprintf(outfile, "\nWatchdog switch to Failure mode.\n");
                        fprintf(outfile, "No more action from Watchdog until both servers are up.\n\n");
                    }
                    
                } else { /* in failure mode */
                    
                    /* in failure mode, watchdog won't execute any
                     * admin command except 'hsb connect' because it
                     * might interfere with what DBA is doing.
                     */
                    /*
                     * It will now: If both are secondary, sets one to
                     * primary if configured to do so.
                     */
                    if (wdaction == HSB_CONN_DOWN) {
                        /* if changed, print the server status to log */
                        if (wdaction != lastwdaction) {
                            print_hsb_servers_status(&hsb_servers, outfile);
                            fprintf(outfile, "%s\n", err_msg);
                        }
                        /* 
                         * Set primary to primary alone first before reconnect
                         * because the primary can be in an unknown state.
                         */
                        watchdog_set_pri_alone(&hsb_servers, 
                                               config.AutoSwitch, outfile);
                        rc = watchdog_hsb_connect(&hsb_servers, outfile);
                        if (rc == 0) {
                            recheck = 1;
                        } else if (rc == 14524 || rc == 14525) {
                            /* DBs are out of sync, won't fix it:
                             * automatic netcopy is not a good idea
                             */
                        } else {
                            ; /* failed */
                        }
                    } else if (wdaction == BOTH_SERVERS_SECONDARY
                               && config.DualSecAutoSwitch) {

                        /* without autoswitch check this may short circuit */
                        
                        if (wdaction != lastwdaction) {
                            print_hsb_servers_status(&hsb_servers, outfile);
                            fprintf(outfile, "%s\n", err_msg);
                        }
                        if (doublecheck) {
                            watchdog_switch_one_to_pri(&hsb_servers,
                                                       config.DualSecAutoSwitch,
                                                       outfile);
                            doublecheck = FALSE;
                        } else {
                            doublecheck = TRUE;
                        }
                        recheck = 1;
                    }
                }
            }
            
            if (!config.ObserveOnly && status_changed
                && wdaction != NOACTION) {

                printf("See '%s' for more information.\n\n",
                       config.logfilename); 
            }
            fflush(stdout); /* needed if stdout is redirected to a file */
            fflush(outfile);
            lastwdaction = wdaction;

            if (!recheck) {
                mysleep(config.PingInterval);
                doublecheck = FALSE; /* doublecheck must imply recheck */
            }
            
            recheck = 0;
            
            if (runtime != 0) {
                if ( time(NULL) > (start_time + runtime) ) {
                    printf("Watchdog exit: time ran out.\n");
                    break; /* main loop */
                }
            }
        }
        
        return (0);
}

/* EOF */
