/*************************************************************************
**  source       * embed.c
**               * 
**  description  * SOLID SQL API usage sample:
**               * 
**               * Embedded SQL statements translated to dynamic CLI
**               * function calls
**               * 
**               * (C) Copyright Solid Information Technology Ltd 2005
\*************************************************************************/

/* Common system headers */

#ifdef SS_NT
#  include <windows.h>
#endif

#include <stdio.h>
#include <assert.h>

#include <solidodbc3.h>

/* Solid samples portability header */
#include "c.h"

/* Constants */
#define MAX_NAME_LEN 80

/* Utility macros - very crude error handling */
#define CHECK(r)                                                        \
        if ( ((r) != SQL_SUCCESS) && ((r) != SQL_SUCCESS_WITH_INFO) ) { \
            print_error(r, henv, hdbc, hstmt);                          \
        }                                                               \
        assert( ((r) == SQL_SUCCESS) || ((r) == SQL_SUCCESS_WITH_INFO) )

      

/* Function prototypes */
static void usage(void);
static void welcome(void);
static void goodbye(SQLCHAR* network_name);
static void ask_connection_info(
        SQLCHAR*  network_name, SQLCHAR* username, SQLCHAR* password
);
static void embedded_to_dynamic_SQL_example(
        SQLCHAR * server, SQLCHAR * uid, SQLCHAR * pwd
);
static void print_error(
        RETCODE rc, SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt
);
static void error_exit(
        RETCODE rc, SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt
);
static SQLCHAR* read_line(SQLCHAR* string, SWORD maxlen);


/*##**********************************************************************
 *
 *              main
 *
 * ad hoc SQL query execution program main function.
 *
 * Parameters : none
 *
 * Return value - give :
 *       0      = Success return
 *      -1      = Error occurred
 *
 */
#if defined(SS_VXW)
#define SS_CDECL
#define main embed_main


void embed() {
        char* args[5];
        args[0] = "embed";

        /* ***** IMPORTANT ****
         *  The host address 255.255.255.255 is invalid and you have
         *  to edit it according to your network configuration.  It's
         *  not possible to connect to localhost 127.0.0.1, so Solid
         *  DatabaseEngine has to be run in another machine.
         */
        args[1] = "tcp 255.255.255.255 2315";
        args[2] = "dba";
        args[3] = "dba";
        args[4] = NULL;
        embed_main(4, args);
        return;
}
#endif

int SS_CDECL main(int argc, SQLCHAR* argv[])
{
        SQLCHAR network_name[MAX_NAME_LEN] = "TCP 127.0.0.1 2313";
        SQLCHAR username[MAX_NAME_LEN]     = "dba";
        SQLCHAR password[MAX_NAME_LEN]     = "dba";

        welcome();

#ifndef __EPOC32__

        if (argc > 1) {
            usage();
            return(-1);
        }
        ask_connection_info(network_name, username, password);
#endif
        embedded_to_dynamic_SQL_example(network_name, username, password);
		goodbye(network_name);        

        return(0);
}


/*##**********************************************************************
 *
 *              usage
 *
 * Print usage instructions on standard output.
 *
 */
static void usage(void)
{
        printf("\nUsage:  embed \n\n");
        welcome();
}


/*##**********************************************************************
 *
 *              welcome
 *
 * Print welcome texts on standard output.
 *
 */
static void welcome(void)
{
        printf("This program automatically executes a set of static SQL \n");
        printf("statements and checks their success.\n\n");
}


/*##**********************************************************************
 *
 *              ask_connection_info
 *
 * Prompt the user for Server network name, username and password.
 *
 * Parameters - output:
 *      network name, username, password: character strings
 *
 */
static void ask_connection_info(SQLCHAR* network_name,
                                SQLCHAR* username,
                                SQLCHAR* password)
{
        printf("Try connecting with the default network name ");
#ifdef SS_WINDOWS
        printf("'shmem solid'.\n");
#else
        printf("'tcp [host] 2315'.\n");
#endif

        printf("\nConnect to solidDB server with network name:\n");
        read_line(network_name, MAX_NAME_LEN);
        if (strcmp(network_name, "") ==0) {
            strcpy(network_name, "tcp 2315");
        }

        printf("\nUsername:\n");
        read_line(username, MAX_NAME_LEN);
        if (strcmp(username, "") ==0) {
            strcpy(username, "dba");
        }

        printf("\nPassword:\n");
        read_line(password, MAX_NAME_LEN);
        if (strcmp(password, "") ==0) {
            strcpy(password, "dba");
        }

}


/*##**********************************************************************
 *
 *              goodbye
 *
 * Output normal termination message.
 *
 */
static void goodbye(UCHAR* network_name)
{
        printf("Disconnected from '%s'.\n", network_name);
}


/*##**********************************************************************
 *
 *              embedded_to_dynamic_SQL_example
 *
 *  Example C function that constructs SQL statements 
 *  within the application. The example comments include
 *  equivalent embedded SQL calls for illustrative purposes. 
 *
 *  Please see Programmer's Guide and Reference
 *  for detailed information on functions and their parameters.
 *
 */
static void embedded_to_dynamic_SQL_example(SQLCHAR * server,
                                            SQLCHAR * uid,
                                            SQLCHAR * pwd)
{
        SQLHENV   henv  = (SQLHENV)NULL;
        SQLHDBC   hdbc  = (SQLHDBC)NULL;
        SQLHSTMT  hstmt = (SQLHSTMT)NULL;
        SQLCHAR   name[MAX_NAME_LEN+1];
        SDWORD    id;
        SQLLEN    namelen;
        SQLLEN    rowcount;
        RETCODE   rc;

        /********************************************************/

        printf("Connecting to server '%s' ...\n", server);

        /* EXEC SQL CONNECT TO :server USER :uid USING :pwd;
         *
         *   Allocate an environment handle.
         *   Allocate a connection handle. 
         *   Connect to a data source. 
         *   Allocate a statement handle.
         */

        rc = SQLAllocEnv(&henv);
        CHECK(rc);
        rc = SQLAllocConnect(henv, &hdbc);
        CHECK(rc);
        rc = SQLConnect(hdbc, server, SQL_NTS, uid, SQL_NTS,  pwd, SQL_NTS);
        CHECK(rc);
        rc = SQLAllocStmt(hdbc, &hstmt);
        CHECK(rc);

        /********************************************************/

        printf("Creating table 'NAMEID' ...\n");

        /* EXEC SQL CREATE TABLE NAMEID (ID integer, NAME varUCHAR(50));
         *
         *   Execute the SQL statement.
         */

        rc = SQLExecDirect(hstmt, (SQLCHAR *)"CREATE TABLE NAMEID (ID INTEGER, NAME VARCHAR(50))", SQL_NTS);
        CHECK(rc);

        /********************************************************/

        printf("Commit.\n");

        /* EXEC SQL COMMIT WORK;     
         *
         *   Commit the table creation.
         */

        /* Note that the default transaction mode for drivers 
         * that support SQLSetConnectOption is auto-commit    
         * and SQLTransact here has no effect. */

        rc = SQLTransact(henv, hdbc, SQL_COMMIT);
        CHECK(rc);

        /********************************************************/

        printf("Inserting a row to table 'NAMEID' (500, Babbage) ...\n");

        /* EXEC SQL INSERT INTO NAMEID VALUES ( :id, :name );
         *
         *   Show the use of the SQLPrepare/SQLExecute method: 
         *     Prepare the insertion and bind parameters. 
         *     Assign parameter values. 
         *     Execute the insertion.
         */

        rc = SQLPrepare(hstmt, (SQLCHAR *)"INSERT INTO NAMEID VALUES(?,?)",
                        SQL_NTS);
        CHECK(rc);
        rc = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, SQL_C_SLONG, 
                              SQL_INTEGER, 0, 0, &id, 0, NULL);
        CHECK(rc);
        rc = SQLBindParameter(hstmt, 2, SQL_PARAM_INPUT, SQL_C_CHAR, 
                              SQL_VARCHAR, MAX_NAME_LEN, 0, name, 0, NULL);
        CHECK(rc);

        id=500;
        strcpy(name, "Babbage");
        rc = SQLExecute(hstmt);
        CHECK(rc);

        /********************************************************/

        printf("Commit.\n");

        /* EXEC SQL COMMIT WORK;
         *
         *   Commit the insertion.
         */

        rc = SQLTransact(henv, hdbc, SQL_COMMIT);
        CHECK(rc);

        /* Close statement */
        rc = SQLFreeStmt(hstmt, SQL_CLOSE);
        CHECK(rc);

        /* Unbind Parameters */
        rc = SQLFreeStmt(hstmt, SQL_RESET_PARAMS);
        CHECK(rc);
        id = 0;
        name[0] = '\0';

        /********************************************************/

        printf("Retrieving data from table 'NAMEID'...\n");

        /* EXEC SQL DECLARE c1 CURSOR FOR SELECT ID, NAME FROM NAMEID;   
         * EXEC SQL OPEN c1; 
         *
         *   Show the use of the SQLExecDirect method:
         *     Execute the selection. 
         *
         *   Note that the application does not explicitly declare a cursor.
         */

        rc = SQLExecDirect(hstmt, (SQLCHAR *)"SELECT ID,NAME FROM NAMEID",
                           SQL_NTS);
        CHECK(rc);

        /* EXEC SQL FETCH c1 INTO :id, :name; 
         *
         *   Bind the columns of the result set with SQLBindCol.    
         *   Fetch the first row.
         */

        rc = SQLBindCol(hstmt, 1, SQL_C_SLONG, &id, 0, NULL);
        CHECK(rc);
        rc = SQLBindCol(hstmt, 2, SQL_C_CHAR, name,
                        (SDWORD)sizeof(name), &namelen);
        CHECK(rc);
        rc = SQLFetch(hstmt);
        CHECK(rc);

        printf("A row found (%d, %s).\n", id, name);

        rc = SQLFetch(hstmt);
        assert(rc == SQL_NO_DATA_FOUND);

        printf("End of result set reached.\n");

        /********************************************************/

        printf("Commit.\n");

        /* EXEC SQL COMMIT WORK;
         *
         *   Commit the transaction.
         */

        rc = SQLTransact(henv, hdbc, SQL_COMMIT);
        CHECK(rc);

        /********************************************************/

        printf("Closing cursor ...\n");

        /* EXEC SQL CLOSE c1;         
         *
         *   Free the statement handle.
         */

        /* Close statement */
        rc = SQLFreeStmt(hstmt, SQL_CLOSE);
        CHECK(rc);

        /* Unbind variables */
        rc = SQLFreeStmt(hstmt, SQL_UNBIND);
        CHECK(rc);

        /********************************************************/

        printf("Deleting all rows from table 'NAMEID' ...\n");

        /* EXEC SQL DELETE FROM NAMEID
         *
         *   Execute the SQL statement.
         *   Check from sqlca structure the number of rows deleted.
         */

        rc = SQLExecDirect(hstmt, (SQLCHAR *)"DELETE FROM NAMEID", SQL_NTS);
        CHECK(rc);

        rc = SQLRowCount(hstmt, &rowcount);
        CHECK(rc);
        assert( rowcount == (SDWORD)1 );
        printf("1 row deleted.\n");

        /********************************************************/

        printf("Commit.\n");

        /* EXEC SQL COMMIT WORK; */

        rc = SQLTransact(henv, hdbc, SQL_COMMIT);
        CHECK(rc);

        /********************************************************/

        printf("Dropping table 'NAMEID' ...\n");

        /* EXEC SQL DROP TABLE NAMEID
         *
         *   Execute the SQL statement.
         */

        rc = SQLExecDirect(hstmt, (SQLCHAR *)"DROP TABLE NAMEID", SQL_NTS);
        CHECK(rc);

        /********************************************************/

        printf("Commit.\n");

        /* EXEC SQL COMMIT WORK;      */

        rc = SQLTransact(henv, hdbc, SQL_COMMIT);
        CHECK(rc);

        /* Release statement resource allocation */
        rc = SQLFreeStmt(hstmt, SQL_DROP);
        CHECK(rc);

        /********************************************************/

        printf("Disconnecting from server ...\n");

        /* EXEC SQL DISCONNECT;            
         *
         *   Disconnect from the data source. 
         *   Free the connection handle.      
         *   Free the environment handle.
         */

        SQLDisconnect(hdbc);
        SQLFreeConnect(hdbc);
        SQLFreeEnv(henv);
}


/*##**********************************************************************
 *
 *              terminate_connection
 *
 * Close connection and free resources.
 *
 */
static void terminate_connection(SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt)
{
        /* Free the statement handle.  
         * Disconnect from the database.
         * Free the connection handle.  
         * Free the environment handle.
         */

        if (hstmt) {
            SQLFreeStmt(hstmt, SQL_DROP );
            SQLDisconnect(hdbc);
        }
        if (hdbc) {
            SQLFreeConnect(hdbc);
        }
        if (henv) {
            SQLFreeEnv(henv);
        }
}


/*##**********************************************************************
 *
 *              print_error
 *
 * Output error message and disconnect from database.
 *
 */
static void print_error(RETCODE rc, SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt)
{
        SQLCHAR      SQLstate[MAX_NAME_LEN];
        SQLINTEGER   errorcode;
        SQLCHAR      errormsg[SQL_MAX_MESSAGE_LENGTH];
        SQLSMALLINT  msglen;
        RETCODE      r;
        
        printf("Error (%d) encountered!\n", rc);
        
        /* Print more error info, if available */
        if (henv && hdbc && hstmt) {
            r = SQLError(henv, hdbc, hstmt, (SQLCHAR *)SQLstate,
                         (SQLINTEGER *)&errorcode, (SQLCHAR *)errormsg,
                         SQL_MAX_MESSAGE_LENGTH - 1, &msglen);
            if (r != SQL_NO_DATA_FOUND) {
                printf("%s\n", errormsg);
            }
        }
}            


/*##**********************************************************************
 *
 *              read_line
 *
 * Read one line from standard input into the parameter buffer.
 * Allows continuing to the following by preceding the end-of-line with '\'.
 *
 */
static SQLCHAR* read_line(SQLCHAR* string, SWORD maxlen)
{

        static SQLCHAR retbuf[MAX_NAME_LEN + 1];
        static SQLCHAR buf[MAX_NAME_LEN + 1];
        SQLCHAR* sp = NULL; 
        SQLCHAR* tp = NULL;
        bool loopend = FALSE;

        *string = '\0';
        retbuf[0] = '\0';
        retbuf[MAX_NAME_LEN] = '\0';
        buf[0] = '\0';
        buf[MAX_NAME_LEN] = '\0';

        while (!loopend) {

            sp = buf;

            /* read a line */
            fgets((char*)sp, MAX_NAME_LEN - 1, stdin);

            /* trim leading non-alpha UCHARacters */
            while (*sp && *sp != '\n' && isspace(*sp)) {
                sp++;
            }

            if (*sp == '\n') {
                *sp = '\0';
            }

            if (!*sp) {
                /* empty string */
                break;
            }

            /* trim trailing non-alpha UCHARacters */
            tp = sp + strlen(sp) - 1;
            while (tp > sp && *tp != '\\' && isspace(*tp) ) {
                tp--;
            }

            /* check if input is continued on the following line */
            if (*tp == '\\') {
                *tp = '\0';
            } else {
                *++tp = '\0';
                loopend = TRUE;
            }
            strcat(retbuf, sp);
        }

        /* check and enforce maxlen */
        if ((SWORD)strlen(retbuf) > maxlen) {
            retbuf[maxlen] = '\0';
        }

        strcpy(string, retbuf);
        return(string);
}

/* EOF */
