/*************************************************************************\
**  file         * sqled.c
**               * 
**  description  * SOLID SQL API usage sample:
**               * 
**               * SQL statement execution utility
**               * 
**               * (C) Copyright Solid Information Technology Ltd 2005
\*************************************************************************/

/* Common system headers */

#ifdef SS_NT
#  include <windows.h>
#endif

/* Common system headers */
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>

#include <solidodbc3.h>

/* Solid samples portability header */
#include "c.h"

#define TRUE 1
#define FALSE 0

typedef enum {
        PRINTTYPE_COLNAMES,
        PRINTTYPE_UNDERLINE,
        PRINTTYPE_COLVALUES
} row_print_type_t;


/* Global buffer for binding data from SQLFetch, and a variable where
 * the actual length of data is stored
 */
static char   bindbuf[(int)(32L * 1024L - 16L)];
static SQLLEN pcbValue;

static FILE* infp;
static char* outfilename;
static FILE* outfp;

static bool silent;
static bool append;
static bool stopwatch;

static void print_version_header(void);

#ifdef SS_VXW
#include <string.h>
char* strdup(const char* str){
        int len = strlen(str);
        char* ret = NULL;
        ret = (char*)malloc(len);
        if (!ret) {
            assert(0);
        }
        strncpy(str, ret, len);
        return ret;
}
#elif defined(H1X64)
extern char *strdup(const char *);
#endif

/*##**********************************************************************\
 * 
 *		ui_sqlinput
 * 
 * Reads SQL command from input file to be executed.
 *
 */
char* SS_CDECL sqlinput(FILE* fp, bool change_case,
                        int (SS_CDECL *outfunc)(char* format, ...))
{
        char* buf;
        int buf_len = 20;
        int str_len = 0;
        int c;

        buf = (char*)malloc(buf_len);

        while ((c = fgetc(fp)) != EOF) {
            if (c == '\n') {
                continue;
            }

            if (c == ';') {
                break;
            }
                
            if (str_len == buf_len-2) {
                buf = (char*)realloc(buf, buf_len*2);
                buf_len *= 2;
            }
            buf[str_len] = (change_case ? toupper(c) : c);
            str_len++;
        }
        buf[str_len] = '\0';
        return buf;
}
            
/*##**********************************************************************\
 * 
 *		sql_printf
 * 
 * Prints out like printf
 *
 */

static int SS_CDECL sql_printf(char* format, ...)
{
        va_list ap;
        int retval;

        va_start(ap, format);
        retval = vprintf(format, ap);
        va_end(ap);

        if (outfilename != NULL) {
            if (outfp == NULL) {
                outfp = fopen(outfilename, append ? "a+" : "w");
                if (outfp == NULL) {
                    sql_printf("Error: Failed to open output file '%s'\n", outfilename);
                    exit(1);
                }
            }
            va_start(ap, format);
            vfprintf(outfp, format, ap);
            va_end(ap);
        }

        return(retval);
}

/*##**********************************************************************\
 *
 *              outerror
 *
 * Prints out error message
 *
 */

static void outerror(SQLHSTMT hstmt)
{
        SQLCHAR     szSqlState[6];
        SQLCHAR     szErrorMsg[256];
        SQLINTEGER  naterr;
        SQLSMALLINT length;
        SQLCHAR*    stripped;
        int i;

        SQLError(SQL_NULL_HENV, SQL_NULL_HDBC, hstmt, szSqlState,
                 &naterr, szErrorMsg, sizeof(szErrorMsg), &length);

        stripped = szErrorMsg;
        for (i = 0; i < 3; i++) {
            UCHAR* tmp;
            tmp = (unsigned char*)strchr((char*)stripped, ']');
            if (tmp != NULL) {
                stripped = tmp + 1;
            } else {
                break;
            }
        }

        sql_printf("%s\n", stripped);
}

static void rtrim_spaces(char* p)
{
        int len;

        len = strlen(p);
        while (len > 0 && p[len - 1] == ' ') {
            p[len - 1] = '\0';
            len--;
        }
}

static bool suc(SQLHDBC hdbc, RETCODE rc)
{
        SQLCHAR   szSqlState[6];
        SQLCHAR   szErrorMsg[256];
        SDWORD    naterr;
        SWORD     length;

        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {
            return(TRUE);
        }

        SQLError(SQL_NULL_HENV, hdbc, SQL_NULL_HSTMT, szSqlState,
                 &naterr, szErrorMsg, sizeof(szErrorMsg), &length);
        sql_printf("Error %s: %s\n", szSqlState, szErrorMsg);
        return(FALSE);
}

static void row_print(SQLHSTMT hstmt, row_print_type_t type)
{
        uint  i;
        char* p;
        uint  l;
        SWORD col_c,ccount=0;
        UCHAR szColName[20];
        SWORD pcbColName;
        SWORD pfSQLType;
        SQLULEN pcbColDef;
        SWORD pibScale;
        SWORD pfNullable;
   
        SQLNumResultCols(hstmt, &col_c);

        for (i = 0; (SWORD)i < col_c; i++) {
            bool align_left;
            align_left = FALSE;
            ccount = i+1;
            SQLDescribeCol(hstmt, ccount, szColName, sizeof(szColName),
                           &pcbColName, &pfSQLType, &pcbColDef,
                           &pibScale, &pfNullable);

            switch (type) {
                case PRINTTYPE_COLNAMES:
                    p = (char*)szColName;
                    break;
                case PRINTTYPE_UNDERLINE:
                    p = strdup((char*)szColName);
                    memset(p, '-', strlen(p));
                    break;
                case PRINTTYPE_COLVALUES:
                    SQLGetData(hstmt, ccount, SQL_C_CHAR,
                               bindbuf, sizeof(bindbuf), &pcbValue);

                    if (pcbValue == SQL_NULL_DATA) {
                        p = "NULL";
                    } else {
                        p = bindbuf;
                    }
                    break;
                default:
                    assert(0);
            }
            switch (pfSQLType) {
                case SQL_INTEGER:
                    l = 9;
                    break;
                case SQL_SMALLINT:
                    l = 6;
                    break;
                case SQL_TINYINT:
                    l = 4;
                    break;
                case SQL_REAL:
                    l = 12;
                    break;
                case SQL_DOUBLE:
                case SQL_FLOAT:
                    l = 14;
                    break;
                case SQL_CHAR:
                case SQL_VARCHAR:
                case SQL_LONGVARCHAR:
                    align_left = TRUE;
                    l = 18;
                    break;
                case SQL_DATE:
                    l = 10;
                    break;
                case SQL_TIME:
                    l = 8;
                    break;
                case SQL_TIMESTAMP:
                    l = 19;
                    break;
                case SQL_NUMERIC:
                case SQL_DECIMAL:
                    l = 14;
                    break;
                default:
                    align_left = TRUE;
                    l = 14;
                    break;
            }
            rtrim_spaces(p);
            if (align_left) {
                sql_printf("%-*s ", l, p);
            } else {
                sql_printf("%*s ", l, p);
            }
            switch (type) {
                case PRINTTYPE_COLNAMES:
                    break;
                case PRINTTYPE_UNDERLINE:
                    free(p);
                    break;
                case PRINTTYPE_COLVALUES:
                    break;
                default:
                    assert(0);
            }
        }
        sql_printf("\n");
}

/*##**********************************************************************\
 *
 *              execute_sql
 *
 * Execute sql statement
 */

static void execute_sql(SQLHDBC hdbc, SQLCHAR* sqlstr)
{
        SQLHSTMT hstmt;
        uint row_c;
        RETCODE rc;
        SWORD pccol;
        SWORD colCount=0;
        static SWORD par_c;
#define PLISTSZ 100
        char* plist[PLISTSZ];
        int pcount = 0;
        
        rc = SQLAllocStmt(hdbc, &hstmt);
        if (!suc(hdbc, rc)) {
            exit(1);
        }
        
        rc = SQLPrepare(hstmt, sqlstr, SQL_NTS);
        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {
            rc = SQLNumParams(hstmt, &par_c);
            if (rc == SQL_SUCCESS && par_c > 0) {
                int i;
                char* str;
                for (i = 1; i <= par_c; i++) {
                    colCount = i;
                    sql_printf("Input param %d:\n", i);
                    str = sqlinput(infp, FALSE, NULL);
                    rc = SQLSetParam(
                            hstmt,
                            colCount,
                            SQL_C_CHAR,
                            SQL_VARCHAR,
                            0,
                            0,
                            str,
                            NULL);
                    if (pcount == PLISTSZ) {
                        sql_printf("Error: hardcoded limit of '?'-parameters (%d) exceeded.\n", PLISTSZ);
                        exit(1);
                    }
                    plist[pcount++] = str;                    
                    
                    if (rc != SQL_SUCCESS) {
                        break;
                    }
                }
            }
        }
        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {
            rc = SQLExecute(hstmt);
        }

        for (pcount--; pcount >= 0; pcount--) {
            free(plist[pcount]);
        }

        if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {
            SQLNumResultCols(hstmt, &pccol);

            if (pccol != 0) {

                if (!silent) {
                    row_print(hstmt, PRINTTYPE_COLNAMES);
                    row_print(hstmt, PRINTTYPE_UNDERLINE);
                }

                rc = SQL_NO_DATA_FOUND;
                row_c = 0;

                for (;;) {
                    rc = SQLFetch(hstmt);
                    if (rc != SQL_SUCCESS && rc != SQL_SUCCESS_WITH_INFO) {
                        if (rc != SQL_NO_DATA_FOUND) {
                            outerror(hstmt);
                        }
                        break;
                    }
                    row_print(hstmt, PRINTTYPE_COLVALUES);
                    row_c++;
                }
                if (!silent) {
                    sql_printf("%u rows fetched.\n", row_c);
                }

            } else {
                SQLLEN crow;

                rc = SQLRowCount(hstmt, &crow);
                if (rc == SQL_SUCCESS) {
                    if (!silent) {
                        sql_printf("Command completed successfully, %ld rows affected.\n", (long)crow);
                    }
                } else {
                    sql_printf("Command executed, row count unknown, error %d.\n", rc);
                    outerror(hstmt);
                }
            }

        } else {
            outerror(hstmt);
        }

        SQLFreeStmt(hstmt, SQL_DROP);

        sql_printf("\n");
}

static void print_version_header(void)
{
        printf("SQL Editor Example\n");
        printf("Executes SQL statements entered by the user or read from file.\n\n");
}

#if defined(SS_VXW)
#  define SS_CDECL
#  define main sqled_main
int sqled_main(int argc, char* argv[]);
void sqled(){

        char* args[5];
        args[0] = "sqled";

        /* *** IMPORTANT ****           
         * The host address 255.255.255.255 is invalid and you have to
         * edit it according to your network configuration.  It's not
         * possible to connect to localhost 127.0.0.1, so Solid
         * DatabaseEngine has to be run in another machine.
         */
        args[1] = "tcp 255.255.255.255 2315";
        args[2] = "dba";
        args[3] = "dba";
        args[4] = NULL;
        sqled_main(4, args);
        return;
}
#endif

int SS_CDECL main(int argc, char *argv[])
{
        SQLHDBC hdbc;
        SQLHENV henv;
        bool output_sql = FALSE;
        UCHAR* servername = (unsigned char*)"";
        UCHAR username[80];
        UCHAR password[80];
        bool autocommit = FALSE;
        bool username_given = FALSE;
        bool password_given = FALSE;
        RETCODE rc;
        char* infilename = NULL;
        char* sqlstr = NULL;
        int optind = 1;

        if (!silent) {
            print_version_header();
        }
        if (argc - optind == 1 || argc - optind >= 3) {
            /* servername || servername username password [filename] */
            servername = (unsigned char*)argv[optind++];
#ifdef SS_CRX
            servername="tcp 127.0.0.1 2315";
#endif
        }
        if (argc - optind >= 2) {
            /* ... username password [filename] */
            strncpy((char*)username, argv[optind++], sizeof(username)-1);
            username[sizeof(username)-1] = '\0';
            username_given = TRUE;
            strncpy((char*)password, argv[optind++], sizeof(password)-1);
            password[sizeof(password)-1] = '\0';
            password_given = TRUE;
        }
        if (optind < (int)argc) {
            /* ... tablename */
            infilename = argv[optind++];
#ifdef SS_CRX
            infilename = "/ppc/t.sql";
#endif
        }
        if (infilename != NULL) {
            infp = fopen(infilename, "r");
            if (infp == NULL) {
                sql_printf("Error: File '%s' can not be opened.\n",
                           infilename);
                return(1);
            }
        } else if (sqlstr == NULL) {
            infp = stdin;
        }

        if (!(username_given && password_given)) {
            sql_printf("Error: too few parameters entered! \n");
            sql_printf("Usage: sqled servername user password [inputfile] \n");
            sql_printf("  e.g. sqled \"tcp myhost 2315\" myid mypass \n");
            return(1);
        }
        
        rc = SQLAllocEnv(&henv);
        if (rc != SQL_SUCCESS && rc != SQL_SUCCESS_WITH_INFO) {
            sql_printf("Error: failed to allocate environment!\n");
            return(1);
        }

        rc = SQLAllocConnect(henv, &hdbc);
        if (rc != SQL_SUCCESS && rc != SQL_SUCCESS_WITH_INFO) {
            sql_printf("Error: failed to allocate connection!\n");
            return(1);
        }

        rc = SQLConnect(hdbc, servername, SQL_NTS, (UCHAR*)username, SQL_NTS,
                        password, SQL_NTS);
        if (!suc(hdbc, rc)) {
            /* if connection is not established, print error msg and exit */
            sql_printf("SQLConnect failed\n");
            exit(1);
        }
        
        rc = SQLSetConnectOption(hdbc, SQL_AUTOCOMMIT, (long)autocommit);
        if (!suc(hdbc, rc)) {
            sql_printf("SQLSetConnectOption failed\n");
            exit(1);
        }

        if (!silent) {
            if (strcmp((char*)servername, "") != 0) {
                printf("Connected to '%s'.\n", servername);
            } else {
                printf("Connected to default server.\n");
            }
            printf("Enter SQL statements terminated by a semicolon.\n");
            printf("Exit by giving command: quit;\n");        
        }

        if (sqlstr != NULL) {
            execute_sql(hdbc, (UCHAR*)sqlstr);

        } else {

            while (!feof(infp)) {
                char *sqlstr;

                sqlstr = sqlinput(infp, FALSE, silent ? NULL : sql_printf);

                if (sqlstr == NULL ||
                    strcmp(sqlstr, "quit") == 0 ||
                    strcmp(sqlstr, "exit") == 0 ||
                    strcmp(sqlstr, "bye") == 0)
                {
                    if (sqlstr != NULL) {
                        free(sqlstr);
                    }
                    break;

                } else if (strcmp(sqlstr, "help") == 0) {
                    sql_printf("SQL Editor Example executes SQL statements.\n");
                    sql_printf("SQL statements must be terminated by a semicolon (;).\n");
                    sql_printf("To quit give command quit terminated by a semicolon.\n");

                } else {
                    if (output_sql) {
                        sql_printf("%s\n", sqlstr);
                    }
                    execute_sql(hdbc, (UCHAR*)sqlstr);
                }
                free(sqlstr);
            }
        }

        if (!silent) {
            printf("SQL Editor exiting.\n");
        }

        if (outfp != NULL) {
            fclose(outfp);
            outfp = NULL;
            outfilename = NULL;
        }

        SQLDisconnect(hdbc);

        SQLFreeConnect(hdbc);

        SQLFreeEnv(henv);

        return(0);
}

/* EOF */
