-- STEP 03 (master and replica)  
-- Create the database schema and procedures.
-- Execute this script in both master and replica databases. 
 
CREATE TABLE customer (
	id 		INTEGER NOT NULL PRIMARY KEY, 
	status 		CHAR(3) NOT NULL, 
	name		VARCHAR NOT NULL, 
	area 		CHAR(3) NOT NULL, 
	address		VARCHAR NOT NULL, 
	available_credit	FLOAT NOT NULL, 
	account_number  	INTEGER NOT NULL, 
	updatetime 	TIMESTAMP NOT NULL, 
	updatenode 	VARCHAR	 NOT NULL) ; 

-- Enable the synchronization history for this table
ALTER TABLE customer SET SYNCHISTORY ;
COMMIT WORK ; 

GRANT select ON customer TO app_user ;
COMMIT WORK ; 

-- Create index so that customers can be queried by area and name
-- The index on area column is required by the CUSTOMERS_BY_AREA publication
CREATE INDEX customer_idx1 ON customer (area, name) ; 
COMMIT WORK ; 

 
CREATE TABLE order (
	customer_id	INTEGER NOT NULL, 
	id 		INTEGER NOT NULL, 
	status		CHAR(3), 
	product		VARCHAR NOT NULL, 
	value		FLOAT NOT NULL, 
	updatetime 	TIMESTAMP NOT NULL, 
	updatenode 	VARCHAR	 NOT NULL, 
	PRIMARY KEY 	(customer_id, id)) ; 

ALTER TABLE order SET SYNCHISTORY ;
COMMIT WORK ;  

GRANT select ON order TO app_user ; 
COMMIT WORK ; 

CREATE INDEX order_idx1 ON order (id) ; 
COMMIT WORK ; 


-- Accounts can contain different types of financial information about business activities.
-- For instance, there can be sales accounts that summarize sales transactions. 
CREATE TABLE account (
	number		INTEGER NOT NULL PRIMARY KEY, 
	type		CHAR(3) NOT NULL, 
	balance		FLOAT NOT NULL, 
	updatetime	TIMESTAMP NOT NULL, 
	updatenode	VARCHAR NOT NULL) ;

ALTER TABLE account SET SYNCHISTORY ; 
COMMIT WORK ; 

GRANT select ON account TO app_user ;
COMMIT WORK ; 

-- Any errors that are detected by the Intelligent Transactions are logged to the ERR_LOG table
CREATE TABLE ERR_LOG (
	createtime 	TIMESTAMP NOT NULL PRIMARY KEY, 
	message		VARCHAR NOT NULL) ; 

ALTER TABLE err_log SET SYNCHISTORY ; 
COMMIT WORK ; 

GRANT SELECT ON err_log TO app_user ; 
COMMIT WORK ;

-- Create a procedure that inserts a new row to the ORDER table
"CREATE PROCEDURE insert_order (
	customer_id	INTEGER, 
	id 		INTEGER, 
	product		VARCHAR, 
	order_value	FLOAT) 
RETURNS
	(p_updatetime	TIMESTAMP, 
	p_status		CHAR(3), 
	p_message	VARCHAR)
BEGIN 

-- Validation rules for a new order:
-- 1) Customer with status 'OK' must exist  
-- 2) The customer's credit limit must not be exceeded by this order

-- Status of a successful order is 'OK'.
-- If any of the validation rules fail, the order status becomes 'INV' - invalid
-- and parameter ('ORDER', 'INV') is put to the bulletin board 
-- to inform any other parts of the transaction about the failed order 

	DECLARE p_available_credit 	FLOAT ; 
	DECLARE p_updatenode 	VARCHAR ; 
	DECLARE p_dummy 	VARCHAR ; 

	EXEC SQL WHENEVER SQLERROR ABORT ; 

	p_available_credit := 0 ; 
	p_status := 'OK' ; 
	p_updatetime := now() ; 
	p_updatenode := get_param('NODE_NAME') ; 
	p_message := '' ; 

-- Perform the validation 
-- 1 & 2: Check that the customer with sufficient credit exists

	EXEC SQL PREPARE ins_order_check_customer 
		SELECT 	available_credit 
		FROM 	customer
		WHERE 	id = ? AND 
			status = 'OK' ; 
	EXEC SQL EXECUTE ins_order_check_customer 
		USING (customer_id) INTO (p_available_credit) ; 
	EXEC SQL FETCH ins_order_check_customer ; 
	
	IF (NOT SQLSUCCESS) OR (p_available_credit < order_value)  
	THEN 
-- Leave the ('ORDER', 'INV') parameter to the bulletin board to inform the subsequent parts of the transaction
-- about the failed order. 
		p_dummy := put_param('ORDER', 'INV') ; 
		p_status := 'INV' ; 

-- Create an entry to the error log table. 
		p_message := 'Order ' + convert_char(id) + ' for customer ' + convert_char(customer_id)  
			+ ' failed: Customer does not exist or credit limit exceeded ' ; 

		EXEC SQL PREPARE ins_order_make_log_entry 
			INSERT INTO err_log VALUES (?,?) ; 
		EXEC SQL EXECUTE ins_order_make_log_entry 
			USING (p_updatetime, p_message) ; 
		EXEC SQL CLOSE ins_order_make_log_entry ; 
		EXEC SQL DROP ins_order_make_log_entry ; 
	END IF ; 
	EXEC SQL CLOSE ins_order_check_customer ; 
	EXEC SQL DROP ins_order_check_customer ; 

-- After validation, do the actual insert with proper status and updatetime values.

	EXEC SQL PREPARE ins_order_do_insert
		INSERT INTO order	(customer_id, id, status, product, value, updatetime, updatenode) 
		VALUES 	(?,?,?,?,?,?,?) ; 
	EXEC SQL EXECUTE ins_order_do_insert 	
		USING (customer_id, id, p_status, product, order_value, p_updatetime, p_updatenode) ;  
	EXEC SQL CLOSE ins_order_do_insert ; 
	EXEC SQL DROP ins_order_do_insert ; 

-- Do NOT commit transaction inside procedure. 
-- The transaction is committed or rolled back by application (in replica) 
-- or Solid DatabaseEngine (in master). 

END" ; 
COMMIT WORK ; 

GRANT execute ON insert_order TO app_user ; 
COMMIT WORK ; 


-- Update the used_credit column of the given customer
"CREATE PROCEDURE update_customer_credit  
	(customer_id 	INTEGER, 
	credit_change 	FLOAT) 
RETURNS 
	(p_updatetime	TIMESTAMP, 
	p_status 		CHAR(3)) 
BEGIN 

-- Validation rules for this operation:
-- 1) Customer with status 'OK' must exist  
-- 2) The parameter bulletin board must not contain parameter ('ORDER', 'INV') 

	DECLARE p_updatenode VARCHAR ; 
	DECLARE p_dummy VARCHAR ; 

	EXEC SQL WHENEVER SQLERROR ABORT ; 

-- Check if inserting the order has succeeded. 
	p_status := get_param('ORDER') ; 
	p_updatenode := get_param('NODE_NAME') ; 
	p_updatetime := now() ; 
	IF p_status IS NULL 
	THEN 	

-- If the parameter does not exist in the bulletin board, the validation of the order has been successful.
		p_status := 'OK' ; 

-- Update the customer credit. Here we assume that the customer existence has already been checked 
-- earlier in this transaction. 
		EXEC SQL PREPARE upd_credit_do_update
			UPDATE 	customer 
			SET 	available_credit = available_credit + ?, 
				updatetime = ?, 
				updatenode = ?
			WHERE 	id = ? AND status = 'OK' ; 
		EXEC SQL EXECUTE upd_credit_do_update
			USING 	(credit_change, p_updatetime, p_updatenode, customer_id)  ; 

-- If no rows were affected by the operation, then something is fatally wrong in the database
-- In this case, force the Solid DatabaseEngine perform rollback and halt the message execution.
-- The error code is returned to replica as the error code of the MESSAGE FORWARD or 
-- MESSAGE GET REPLY statement. 
		IF SQLROWCOUNT = 0 THEN 			
			p_dummy := put_param('SYS_ROLLBACK', 'YES') ;
			p_dummy := put_param('SYS_ERROR_CODE','99901') ;
			p_dummy := put_param('SYS_ERROR_TEXT','Updating customer credit limit failed - Customer not found') ;
		END IF; 
	ELSE 		

-- Placing the order had failed because of a validation error earlier in this transaction.  
-- Customer credit must not be updated. Just return 'INV' as the status of the operation. 
		p_status := 'INV' ; 
	END IF ; 

END" ; 
COMMIT WORK ; 

GRANT execute ON update_customer_credit TO app_user ; 
COMMIT WORK ; 


-- Create a procedure that updates the account balance of a given account. 
"CREATE PROCEDURE update_account_balance
	(acc_number 	INTEGER, 
	balance_change 	FLOAT) 
RETURNS (p_updatetime 	TIMESTAMP, 
	p_status 		CHAR(3), 
	p_message 	VARCHAR)
BEGIN 
-- Validation rules for the operation:
-- 1) Account with status 'OK' must exist  
-- 2) The parameter bulletin board must not contain parameter ('ORDER', 'INV') 

	DECLARE p_updatenode VARCHAR ; 

	EXEC SQL WHENEVER SQLERROR ABORT ; 

-- Check if inserting the order has succeeded. 
	p_updatetime := now() ;
	p_updatenode := get_param('NODE_NAME') ; 
	p_status := get_param('ORDER') ; 
	IF p_status IS NULL 
	THEN 	

-- If the parameter does not exist in the bulletin board, the validation of the order has been successful.
		p_status := 'OK' ; 

		EXEC SQL PREPARE upd_account_do_update
			UPDATE 	account  
			SET 	balance = balance + ?, 	
				updatetime = ?, 
				updatenode = ?
			WHERE 	number = ? ; 
		EXEC SQL EXECUTE upd_account_do_update
			USING 	(balance_change, p_updatetime, p_updatenode, acc_number)  ; 

-- If no rows were affected, the update has failed. 
		IF SQLROWCOUNT = 0 THEN 

			p_status := 'INV' ; 

-- Create an entry to the error log table about the missing account. 
			p_message := 'Updating account ' + convert_char(acc_number) + 
				' failed: Account not found. Updated amount: ' + 
				convert_char(balance_change) ;
	
			EXEC SQL PREPARE upd_account_log_error 
				INSERT INTO err_log VALUES (?,?) ; 
			EXEC SQL EXECUTE upd_account_log_error
				USING (p_updatetime, p_message) ; 
			EXEC SQL CLOSE upd_account_log_error ; 
			EXEC SQL DROP upd_account_log_error ; 
		END IF ; 
		EXEC SQL CLOSE 	upd_account_log_error ; 
		EXEC SQL DROP 	upd_account_log_error ; 

		EXEC SQL CLOSE 	upd_account_do_update ; 
		EXEC SQL DROP 	upd_account_do_update ; 		 
	END IF ; 
END" ; 
COMMIT WORK ; 

GRANT execute ON update_account_balance TO app_user ; 
COMMIT WORK ; 
