-- STEP 06 (replica)
-- Create procedures that do the data synchronization 

-- These procedures together implement a sample synchronization process. 
-- Create them only in the replica databases.
-- The synchronization is started by calling the run_sync procedure. 

message reg_publ begin;
message reg_publ append register publication customers_by_area;
message reg_publ end;
commit work;
message reg_publ forward timeout forever;
commit work;

 
"create procedure run_sync 
	returns  (overall_success integer) 
begin 

-- This is the main procedure. It does following things: 
--    - check whether there are messages with 'unsent' status in replica
--    - check whether there are messages with 'reply not received' status
--    - create and send a new synch message and wait for the reply message 

-- Return values: 
--    0 = all ok, 
--    1 = new message sent successfully, problems with earlier ones
--   -1 = failed to send the new message

	declare success integer ;

	exec sql whenever sqlerror abort ; 
	
-- Check whether there are no unsent messages in the replica database.
-- If there are, then send those messages now. 
	overall_success := 0 ; 

	exec sql prepare sync_crsr_1 call check_unsent_sync_msgs;
	exec sql execute sync_crsr_1 into (success);
	exec sql fetch sync_crsr_1;
	if success = 0 then 
		overall_success := 1 ;
	end if ; 
	exec sql close sync_crsr_1;
	exec sql drop sync_crsr_1;

-- Check that there are no sent messages in the replica database
-- that haven't received a reply message yet. If such message exist, 
-- try to get the reply messages for them now.

	exec sql prepare sync_crsr_2 call check_unreceived_sync_msgs;
	exec sql execute sync_crsr_2 into (success);
	exec sql fetch sync_crsr_2;
	if success = 0 then 
		overall_success := 1 ; 
	end if ; 
	exec sql close sync_crsr_2;
	exec sql drop sync_crsr_2;

-- Finally, build a new synchronization message, send it and wait for reply

	exec sql prepare sync_crsr_3 call send_new_syncmsg;
	exec sql execute sync_crsr_3 into (success);
	exec sql fetch sync_crsr_3;
	if success = 0 then 
		overall_success := -1 ; 
	end if ; 
	exec sql close sync_crsr_3;
	exec sql drop sync_crsr_3;
end"; 
commit work ; 

GRANT execute ON run_sync TO app_user ; 
commit work ; 


"create procedure check_unsent_sync_msgs
	returns (success integer)
begin

	declare msgname varchar ; 
	declare tmpstr varchar ; 

	exec sql whenever sqlerror abort ; 

-- Check from sys_sync_replica_msginfo table that whether there are messages with status  
-- 22 = R_SAVED (- replica has saved an outgoing message but message has not been sent yet)

	success := 1 ; 

	exec sql prepare sync_us_crsr1 
		select 	MSG_NAME 
		from 	SYS_SYNC_REPLICA_MSGINFO 
		where	STATE = 22 ; 

	exec sql execute sync_us_crsr1 into (msgname) ; 
	exec sql fetch sync_us_crsr1 ; 
	while sqlsuccess loop 

-- try to send all unsent messages to master
		tmpstr := 'MESSAGE ' + msgname + ' FORWARD TIMEOUT 60' ; 
		exec sql prepare sync_us_crsr2 tmpstr ; 		
	        exec sql commit work ;
		exec sql execute sync_us_crsr2 ; 
		if not sqlsuccess then 
			success := 0 ;
		end if ; 
		exec sql close sync_us_crsr2 ; 
		exec sql drop sync_us_crsr2 ; 

		exec sql fetch sync_us_crsr1 ; 
	end loop ; 

	exec sql close sync_us_crsr1 ; 
	exec sql drop sync_us_crsr1 ;
	exec sql commit work ;

end"; 
commit work ; 

GRANT execute ON check_unsent_sync_msgs TO app_user ; 
commit work ; 

 
"create procedure check_unreceived_sync_msgs
	returns (success integer)
begin
	declare msgname varchar ; 
	declare tmpstr varchar ; 

	exec sql whenever sqlerror abort ; 

-- check from sys_sync_replica_msginfo table that whether there are messages with status  
-- 23 = R_SENT (- replica has sent a message to master but has not yet received 
-- the reply message)

	success := 1 ; 
	
	exec sql prepare sync_ur_crsr1 
		select 	MSG_NAME 
		from 	SYS_SYNC_REPLICA_MSGINFO 
		where	STATE = 23 ; 

	exec sql execute sync_ur_crsr1 into (msgname) ; 
	exec sql fetch sync_ur_crsr1 ; 
	while sqlsuccess loop 

-- try to get all unreceived reply messages to the replica 
		tmpstr := 'MESSAGE ' + msgname + ' GET REPLY TIMEOUT 60' ; 
		exec sql prepare sync_ur_crsr2 tmpstr ; 
		exec sql execute sync_ur_crsr2 ; 
		if not sqlsuccess then 
			success := 0 ;
		end if ; 
		exec sql close sync_ur_crsr2 ; 
		exec sql drop sync_ur_crsr2 ; 

		exec sql fetch sync_ur_crsr1 ; 
	end loop ; 

	exec sql close sync_ur_crsr1 ; 
	exec sql drop sync_ur_crsr1 ;
	exec sql commit work ;

end"; 
commit work ; 

GRANT execute ON check_unreceived_sync_msgs TO app_user ;
COMMIT WORK ; 

-- create a sequence that assigns a new number for the names of new synchronization messages 
create sequence syncmsgnumber ;
commit work ; 

  
"create procedure send_new_syncmsg 
	returns(success integer) 
begin 
	declare msgnumber integer ; 
	declare tmpstr varchar ; 

-- assign a new unique name to the message
-- it can be done easiest using a sequencer object 
	exec sql whenever sqlerror abort; 

	success := 1 ; 
	exec sequence syncmsgnumber.next into msgnumber ; 

	tmpstr := 'MESSAGE MSG' + convert_char(msgnumber) + ' BEGIN' ; 
	exec  sql prepare sync_nm_crsr1 tmpstr ; 
	tmpstr := 'MESSAGE MSG' + convert_char(msgnumber) + ' APPEND PROPAGATE TRANSACTIONS' ; 
	exec  sql prepare sync_nm_crsr2 tmpstr ; 
-- these rows need to be modified to match the publications of the application
	tmpstr := 'MESSAGE MSG' + convert_char(msgnumber) + 
		' APPEND REFRESH CUSTOMERS_BY_AREA (''NOR'')' ; 
	exec  sql prepare sync_nm_crsr3 tmpstr ; 		
	tmpstr := 'MESSAGE MSG' + convert_char(msgnumber) + ' END' ; 
	exec  sql prepare sync_nm_crsr4 tmpstr ; 
	tmpstr := 'MESSAGE MSG' + convert_char(msgnumber) + ' FORWARD TIMEOUT 60' ; 
	exec  sql prepare sync_nm_crsr5 tmpstr; 
		
	exec sql execute sync_nm_crsr1 ; 
	if not sqlsuccess then 
		success := 0 ;
	end if ;
	exec sql execute sync_nm_crsr2 ; 
	if not sqlsuccess then 
		success := 0 ;
	end if ;
	exec sql execute sync_nm_crsr3 ; 
	if not sqlsuccess then 
		success := 0 ;
	end if ;
	exec sql execute sync_nm_crsr4 ; 
	if not sqlsuccess then 
		success := 0 ;
	end if ;
	exec sql commit work ; 
	exec sql execute sync_nm_crsr5 ; 
	if not sqlsuccess then 
		success := 0 ;
	end if ;
	exec sql commit work ; 

	exec sql drop sync_nm_crsr1; 	
	exec sql drop sync_nm_crsr2; 
	exec sql drop sync_nm_crsr3; 
	exec sql drop sync_nm_crsr4; 
	exec sql drop sync_nm_crsr5; 
end"; 

COMMIT WORK ;

GRANT execute ON send_new_syncmsg TO app_user ; 
COMMIT WORK ; 


