/*************************************************************************\
**  source       * saembed.c
**               * 
**  description  * Solid SA-Interface usage sample:
**               * 
**               * Embedded SQL statements translated to dynamic CLI
**               * (sa calls) function calls
**               * 
**               * (C) Copyright Solid Information Technology Ltd 2005
\*************************************************************************/

/* Common system headers */
#ifdef SS_NT
#  include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <solidodbc3.h>

/* SOLID portability header */
#include "c.h"

/* SOLID SA interface portablity header */
#include "sa.h"

/* Constants */
#define MAX_NAME_LEN 80

/* Function prototypes */
static void usage(void);
static void welcome(void);
static void ask_connection_info(char *, char*, char*);
static void embedded_to_dynamic_SQL_example(char *, char *, char *);
static char* read_line(char* , unsigned int);
static void t_scon_error(SaConnectT *,int,char *,int);
static void t_scur_error(SaCursorT *,int,char *,int);
static void print_assert_info(char *,int);
static int fetch_data(SaConnectT *);

/* Utility macros - very crude error handling */
#define t_con_check(scon,expr) t_scon_error(scon,expr,__FILE__,__LINE__);    
#define t_cur_check(scur,expr) t_scur_error(scur,expr,__FILE__,__LINE__);    

/*=============================================================
Function Name	: main()
Purpose			: ad hoc SQL query execution program main function.
Parameters		: none
Returns			: '0' on success,-1 on failure
==============================================================*/

#ifdef main
#  undef main
#endif

#if defined(SS_NT)
#  ifndef SS_CDECL
#    define SS_CDECL __cdecl
#  endif
#elif defined(SS_VXW)
#define SS_CDECL
#define main embed_main


void embed(){
        char* args[5];
        args[0] = "embed";

        /* **** IMPORTANT ****
         *  The host address 255.255.255.255 is invalid and you have to edit it
         *  according to your network configuration.
         *  It's not possible to connect to localhost 127.0.0.1, so
         *  Solid DatabaseEngine has to be run in another machine.
         */
        args[1] = "tcp 255.255.255.255 2315";
        args[2] = "dba";
        args[3] = "dba";
        args[4] = NULL;
        embed_main(4, args);
        return;
}

#else
#  define SS_CDECL
#endif

int SS_CDECL main(int argc, char* argv[])
{
        if (argc != 4) {
            usage();
            return(-1);
        }

        welcome();
        embedded_to_dynamic_SQL_example(argv[1], argv[2], argv[3]);
        return(0);
}

/*=============================================================
Function Name	: usage
Purpose			: Print usage instructions on standard output.
Parameters		: none
Returns			: none
==============================================================*/
static void usage(void)
{
        printf("\nUsage:  saembed [network name] [username] [password]\n\n");
        welcome();
}

/*=============================================================
Function Name	: welcome
Purpose			: Print welcome texts on standard output.
Parameters		: none
Returns			: none
==============================================================*/
static void welcome(void)
{
        printf("This program automatically executes a set of static SQL \n");
        printf("statements and checks their success through SOLID API(SA) interface.\n\n");
}


/*=============================================================
Function Name	: embedded_to_dynamic_SQL_example
Purpose			: Example C function that constructs SQL statements 
 					within the application. The example comments include
					equivalent embedded SQL calls for illustrative purposes. 
 
 					Please see SOLID Server Programmer's Guide and Reference
 					for detailed information on functions and their parameters.
Parameters		: input
		network name, username, password: character strings 
Returns			: none
==============================================================*/
static void embedded_to_dynamic_SQL_example(char * server, char * uid, char * pwd)
{
        SaRetT		rc = 0;
        SaConnectT	*scon = NULL;
        SaCursorT	*scur = NULL;
        int			id;
        char		*name;
        

        /********************************************************/
        printf("Connecting to server '%s' ...\n", server);

        /* conenct to specified server using SaConnect SA-API */

        scon = SaConnect(server,uid,pwd);
        if(scon == NULL) {
            printf("\n Failed to connect to %s\n",server);
            exit(-1);
        }
        
        /********************************************************/
        printf("Creating table 'NAMEID' ...\n");

		/* DROP TABLE NAMEID if it exists */
        /* CREATE TABLE NAMEID(ID integer, NAME varUCHAR(50)); */
        /* Execute the SQL statement. */

        SaSQLExecDirect(scon,"DROP TABLE NAMEID");
        
        rc = SaTransCommit(scon);
        t_con_check(scon,rc == SA_RC_SUCC);
        
        rc = SaSQLExecDirect(scon, "CREATE TABLE NAMEID (ID INTEGER, NAME VARCHAR(50))");
        t_con_check(scon,rc == SA_RC_SUCC);

        /********************************************************/
        printf("Commit.\n");

		/* commit the transaction using SaTransCommit */
        rc = SaTransCommit(scon);
        t_con_check(scon, rc == SA_RC_SUCC);

        /********************************************************/
        printf("Inserting a row to table 'NAMEID' (500, Babbage) ...\n");

		/* create the cursor scur on the table 'NAMEID' */
		/* bind input data using SaCursorColInt for integer,
			SacursorColStr for string data				*/
		/* open the cursor								*/
		/* Assign parameter values.						*/
		/* now execute SaCursorInsert to insert the data*/
		
        scur = SaCursorCreate(scon,"NAMEID");
        if (scur == NULL) {
            printf("\n Table 'NAMEID' not found \n");
            exit(-1);
        }
        rc = SaCursorColInt(scur, "ID", &id);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        rc = SaCursorColStr(scur, "NAME", &name);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        rc = SaCursorOpen(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        id=500;
        name = (char*)malloc(20);
        strcpy(name, "Babbage");
        
        rc = SaCursorInsert(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        /********************************************************/
        printf("Commit.\n");

        /* Commit the insertion. */

        rc = SaTransCommit(scon);
        t_con_check(scon,rc == SA_RC_SUCC);

        /* Close cursor */
        SaCursorFree(scur);
        free(name);

        /********************************************************/
        /* fetching the data back */
        assert(1 == fetch_data(scon));

        /********************************************************/
        printf("updating rows in table 'NAMEID' ...\n");
        
        scur = SaCursorCreate(scon,"NAMEID");
        if(scur == NULL) {
            printf("\n Table 'NAMEID' not found \n");
            exit(-1);
        }
        rc = SaCursorColStr(scur,"NAME",&name);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        rc = SaCursorOpen(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        rc = SaCursorSearch(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        rc = SaCursorNext(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        name = "updated";
        rc = SaCursorUpdate(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        printf("1 row updated.\n");
        
        /********************************************************/
        printf("Closing cursor ...\n");

        /* Free the cursor */

        SaCursorFree(scur);

        /********************************************************/
        printf("Commit.\n");

        /* EXEC SQL COMMIT WORK;      */

        rc = SaTransCommit(scon);
        t_con_check(scon,rc == SA_RC_SUCC);

        /********************************************************/
        /* fetching the data back */
        assert(1 == fetch_data(scon));

        /********************************************************/
        printf("Deleting all rows from table 'NAMEID' ...\n");
        
        scur = SaCursorCreate(scon,"NAMEID");
        if(scur == NULL) {
            printf("\n Table 'NAMEID' not found \n");
            exit(-1);
        }
        rc = SaCursorOpen(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        rc = SaCursorSearch(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        rc = SaCursorNext(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
        
        rc = SaCursorDelete(scur);
        t_cur_check(scur, rc == SA_RC_SUCC);
		
        printf("1 row deleted.\n");
        
        /********************************************************/
        printf("Closing cursor ...\n");
        
        /* Free the cursor */        
        SaCursorFree(scur);
        
        /********************************************************/
        printf("Commit.\n");
        
        /* EXEC SQL COMMIT WORK;      */

        rc = SaTransCommit(scon);
        t_con_check(scon,rc == SA_RC_SUCC);

        /********************************************************/
        /* fetching the data back */
        assert(0 == fetch_data(scon));

        /********************************************************/
        printf("Dropping table 'NAMEID' ...\n");

        /* EXEC SQL DROP TABLE NAMEID       */
        /* Execute the SQL statement. */

        rc = SaSQLExecDirect(scon,"DROP TABLE NAMEID");
        t_con_check(scon,rc == SA_RC_SUCC);

        /********************************************************/
        printf("Commit.\n");

        /* EXEC SQL COMMIT WORK;      */

        rc = SaTransCommit(scon);
        t_con_check(scon,rc == SA_RC_SUCC);

        /********************************************************/
        printf("Disconnecting from server ...\n");

        /* EXEC SQL DISCONNECT;             */
        /* Disconnect from the data source. */
        /* Free the connection handle.      */
        /* Free the environment handle.     */

        SaDisconnect(scon);               
}

/*=============================================================
Function Name	: t_scon_error()
Purpose			: to get the connection based error message with assertion
Parameters		: 
	scon		: active connection
	rc			: return code
	file		: name of the file
	line		: line number at which the error occured
Returns			: none
==============================================================*/
static void t_scon_error(SaConnectT* scon, int expr, char* file, int line)
{
        char *errstr;
        int errcode;
        
        if (SaErrorInfo(scon, &errstr, &errcode)) {
            printf("\n%d:[%d][%s]", line, errcode, errstr);		
        }	
        if (!expr) {
            print_assert_info(file, line);
        }
}

/*=============================================================
Function Name	: t_scur_error()
Purpose			: to get the cursor based error message with assertion
Parameters		: 
	scur		: active cursor	
	rc			: return code
	file		: name of the file
	line		: line number at which the error occured
Returns			: none
==============================================================*/
static void t_scur_error(SaCursorT* scur, int expr, char* file, int line)
{
        char *errstr;
        int errcode;
        
        if (SaCursorErrorInfo(scur, &errstr, &errcode)) {
            printf("\n%d:[%d][%s]", line, errcode, errstr);			
        }
        if (!expr) {
            print_assert_info(file, line);
        }
}

/*=============================================================
Function Name	: print_assert_info()
Purpose			: to print assert information
Parameters		: 
	file		: name of the file
	line		: line number at which the error occured
Returns			: none
==============================================================*/
static void print_assert_info(char *file, int line)
{
        printf("\n SOLID assertion failed in %d@%s\n", line, file);
        exit(0);
}

/*=============================================================
Function Name	: read_line()
Purpose			: Read one line from standard input into the parameter buffer.
				  Allows continuing to the following by preceding the
                  end-of-line with '\'
Parameters		: 
	string		: string buffer
	maxlen		: max length of string
Returns			: read string
==============================================================*/
static char* read_line(char* string, unsigned int maxlen)
{
        static char retbuf[MAX_NAME_LEN + 1];
        static char buf[MAX_NAME_LEN + 1];
        char* sp = NULL; 
        char* tp = NULL;
        bool loopend = FALSE;
        
        *string = '\0';
        retbuf[0] = '\0';
        retbuf[MAX_NAME_LEN] = '\0';
        buf[0] = '\0';
        buf[MAX_NAME_LEN] = '\0';

        while (!loopend) {
            
            sp = buf;
            
            /* read a line */
            fgets((char*)sp, MAX_NAME_LEN - 1, stdin);
            
            /* trim leading non-alpha UCHARacters */
            while (*sp && *sp != '\n' && isspace(*sp)) {
                sp++;
            }
            
            if (*sp == '\n') {
                *sp = '\0';
            }
            
            if (!*sp) {
                /* empty string */
                break;
            }
            
            /* trim trailing non-alpha UCHARacters */
            tp = sp + strlen(sp) - 1;
            while (tp > sp && *tp != '\\' && isspace(*tp) ) {
                tp--;
            }
            
            /* check if input is continued on the following line */
            if (*tp == '\\') {
                *tp = '\0';
            } else {
                *++tp = '\0';
                loopend = TRUE;
            }
            strcat(retbuf, sp);
        }

        /* check and enforce maxlen */
        if (strlen(retbuf) > maxlen) {
            retbuf[maxlen] = '\0';
        }

        strcpy(string, retbuf);
        return(string);
}

/*=============================================================
Function Name	: fetch_data()
Purpose			: fetches the data from the table 'NAMEID'
Parameters		: 
	scon		: connection parameter
Returns			: total rows
==============================================================*/
static int fetch_data(SaConnectT *scon)
{
        SaCursorT*  scur = NULL;
        SaRetT      rc = 0,rowcount = 0;
        int         id;
        char*       name;
        
		printf("Retrieving data from table 'NAMEID'...\n");

		/* create the cursor scur on the table 'NAMEID' */
		/* bind output data using SaCursorColInt for integer,
			SacursorColStr for string data				*/
		/* open the cursor								*/
		/* start the search operation					*/
		/* now call SaCursorNext to fetch the first row	*/
		/* now execute SaCursorInsert to insert the data*/

        scur = SaCursorCreate(scon,"NAMEID");
		if (scur == NULL) {
			printf("\n Table 'NAMEID' not found \n");
			exit(-1);
		}
		rc = SaCursorColInt(scur,"ID",&id);
		t_cur_check(scur, rc == SA_RC_SUCC);

		rc = SaCursorColStr(scur,"NAME",&name);
		t_cur_check(scur, rc == SA_RC_SUCC);

		rc = SaCursorOpen(scur);
		t_cur_check(scur, rc == SA_RC_SUCC);

		SaCursorSearch(scur);
		
		while (SaCursorNext(scur) == SA_RC_SUCC) {
			rowcount++;
			printf("A row found (%d, %s).\n", id, name);
		}
        rc = SaCursorNext(scur);
		t_cur_check(scur, rc == SA_RC_END);

        SaCursorFree(scur);
        
		rc = SaTransCommit(scon);
        t_con_check(scon, rc == SA_RC_SUCC);
		
		printf("total rows:%d\n", rowcount);

		return(rowcount);
}

/* EOF */
