/*************************************************************************
**  file         * tf1.c
**               *
**  description  * Sample program for TF1
**               * 
**               * (C) Copyright Solid Information Technology Ltd 2005
**************************************************************************/

/*
 * Demonstrates use of Transparent Connectivity (TC) at CONNECTION level.
 *
 * This is the same as the older feature Transparent Failover level 1.
 * (And indeed the sample is exactly the same).
 *
 * In TC we have a virtual connection that may signal that a rollback is
 * needed.  This happens when the real connection below the virtual connection
 * fails or there is a switch-over in the Hotstandby server pair (since we
 * cannot do writes on Secondary server).
 *
 * We are forced to do the rollback as everything (except closing it) done
 * with the connection will result in an error if we don't.
 *
 * So when we receive the native E_EXPECT_ROLLBACK error (25216) we must
 * consider the following:
 *
 *   - uncommitted writes have to be redone after forced rollback
 *   - reads may break in the middle
 *   - commit might not succeed
 *   - prepared statements must be prepared again but statement handles stay
 *     allocated.
 *   - session settings must be redone so set commands must be redone
 *
 * SEE IT IN ACTION:
 *
 * You must externally trigger events to see the failover in action.  This
 * can be done, for example, with:
 *
 *   - use solsql/solcon to connect to the primary server and issue
 *     "ADMIN COMMAND 'hsb switch secondary'".
 *   - terminate the primary server. (If not using a watchdog, you might have
 *     to issue "ADMIN COMMAND 'hsb switch primary'" to the working server).
 *
 */

#ifdef SS_UNIX
#  include <unistd.h>
#  define SLEEP_MULT 1         /* posix sleep is s */
#else
#  ifdef SS_NT
#    include <windows.h>
#    define SLEEP_MULT 1000    /* w32 sleep is ms */
#    define sleep Sleep
#  endif /* SS_NT */
#endif /* !SS_UNIX */

/* Common system headers */
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#include <solidodbc3.h>

/* Solid samples portability header */
#include "c.h"

/* Set the connection string.  With Transparent Connectivity (TC) the
 * connection parameters are more verbose. Here we use connection
 * level transparency.  This is the same as connection string
 * "tf -l1 tcp 2315, 2325" in 4.5.
 */
static char* connstr_default = "TF_LEVEL=CONNECTION SERVERS=tcp 2315, tcp 2325";
static char* username        = "dba";
static char* password        = "dba";

#define E_EXPECT_ROLLBACK    25216
#define E_TABLE_NOT_EXIST    13011

#define MAX_STRLEN           1024
 
/*** prototypes ***************/

static void alloc_env(HENV* henvp);
static void alloc_dbc(HENV henv, HDBC* hdbcp);
static void alloc_stmt(HDBC hdbc, HSTMT* hstmtp);

static void free_env(HENV henv);
static void free_dbc(HDBC hdbc);
static void free_stmt(HSTMT hstmt);

static void sql_connect(HDBC hdbc, char* connstr, char* user, char* pwd);
static void sql_disconnect(HDBC hdbc);

static long get_error_native(SQLHANDLE handle, SQLSMALLINT htype);
static bool check_error_native(SQLHANDLE handle, SQLSMALLINT htype,
                               long expected_err);

/*** assert macros ************/

#define SU(rc) do {                                         \
                   if (rc != SQL_SUCCESS) {                 \
                       printf("\nASSERT: rc = %d\n", (rc)); \
                   }                                        \
                   assert((rc) == SQL_SUCCESS);             \
               } while(0)

#define SUH(rc, h, type)                                                    \
                   do {                                                     \
                       if (rc != SQL_SUCCESS) {                             \
                           printf("\nASSERT: rc = %d, naterr = %ld\n",      \
                                  (rc),                                     \
                                  get_error_native((h), (type)));           \
                       }                                                    \
                       assert((rc) == SQL_SUCCESS);                         \
                   } while(0)

static void alloc_env(HENV* henvp)
{
        RETCODE rc;
        
        rc = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, henvp);
        SU(rc);
}

static void alloc_dbc(HENV henv, HDBC* hdbcp)
{
        RETCODE rc;
        
        rc = SQLAllocHandle(SQL_HANDLE_DBC, henv, hdbcp);
        SU(rc);
}

static void alloc_stmt(HDBC hdbc, HSTMT* hstmtp)
{
        RETCODE rc;

        rc = SQLAllocHandle(SQL_HANDLE_STMT, hdbc, hstmtp);
        SU(rc);
}

static void free_env(HENV henv)
{
        RETCODE rc;
        
        rc = SQLFreeHandle(SQL_HANDLE_ENV, henv);
        SU(rc);
}

static void free_dbc(HDBC hdbc)
{
        RETCODE rc;
        
        rc = SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
        SU(rc);
}

static void free_stmt(HSTMT hstmt)
{
        RETCODE rc;

        rc = SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
        SU(rc);
}

static void sql_connect(HDBC hdbc, char* connstr, char* user, char* pwd)
{
        RETCODE rc;
        
        rc = SQLConnect(hdbc,
                        connstr, SQL_NTS,
                        user,    SQL_NTS,
                        pwd,     SQL_NTS);
        SUH(rc, hdbc, SQL_HANDLE_DBC);
}

static void sql_disconnect(HDBC hdbc)
{
        RETCODE rc;

        rc = SQLDisconnect(hdbc);
        SU(rc);
}

static long get_error_native(SQLHANDLE handle, SQLSMALLINT htype)
{
        SQLSMALLINT recno;
        SQLCHAR     sqlstate[6];
        SQLINTEGER  naterr;
        SQLCHAR     msgtext[MAX_STRLEN];
        SQLSMALLINT len = 0;
        RETCODE     rc;

        for (recno = 1; TRUE; recno++) {   /* get error records */

            /* yes, we could use SQLGetDiagField to get just the naterr
             * but more info may be useful when debugging. */
            rc = SQLGetDiagRec(htype, handle, recno, sqlstate,
                               &naterr, msgtext, sizeof(msgtext), &len);

            if (rc != SQL_SUCCESS) {
                if (rc == SQL_NO_DATA) {   /* no more data */
                    if (recno == 1) {      /* no error records */
                        return -1L;
                    }
                    break;
                }
                SU(rc);
            }  
            printf("get_error_native: %ld (%s: %s)\n",
                   (long) naterr, sqlstate, msgtext);
        }        

        return (long) naterr;
}

static bool check_error_native(SQLHANDLE handle, SQLSMALLINT htype,
                               long expected_err)
{
        SQLSMALLINT recno;
        SQLINTEGER  naterr;
        RETCODE     rc;

        for (recno = 1; TRUE; recno++) { /* get error records */

            rc = SQLGetDiagField(htype, handle, recno, SQL_DIAG_NATIVE,
                                 &naterr, 0, NULL);

            if (rc != SQL_SUCCESS) {
                if (rc == SQL_NO_DATA) { /* no more data */
                    break;
                }
                SUH(rc, handle, htype);
            }  
            if ((long) naterr == expected_err) {
                return TRUE;
            }
        }        

        return FALSE;
}

/*** more assert macros: located here for faster referral need due to goto **/

/* SU() but handles TF exception and restarts process */
#define TFSU(rc, h) do {                                                 \
                        if ((rc) == SQL_ERROR                            \
                             && check_error_native((h), SQL_HANDLE_STMT, \
                                                   E_EXPECT_ROLLBACK)) { \
                            goto tfexception;                            \
                        }                                                \
                        SUH((rc), (h), SQL_HANDLE_STMT);                 \
                    } while(0)

/* SUH() but handles TF exception and restarts process */
#define TFSUH(rc, h, type)                                               \
                    do {                                                 \
                        if ((rc) == SQL_ERROR                            \
                             && check_error_native((h), (type),          \
                                                   E_EXPECT_ROLLBACK)) { \
                            goto tfexception;                            \
                        }                                                \
                        SUH((rc), (h), (type));                          \
                    } while(0)

int SS_CDECL main(int argc, char* argv[])
{
        HENV      henv;
        HDBC      hdbc;
        HSTMT     hstmt;
        RETCODE   rc;

        char* connstr;
        
        int i = 1, j;
        int nloops = 4;
        int nrows;
        
        HSTMT inserth;
        HSTMT selecth;
        
        /* insert vars */
        SQLINTEGER i_index = 1;
        SQLLEN     i_index_ind = 0;
        SQLINTEGER i_block;
        SQLLEN     i_block_ind = 0;
        SQLINTEGER i_data;
        SQLLEN     i_data_ind = 0;

        /* select vars */
        SQLINTEGER s_index;
        SQLLEN     s_index_ind = 0;
        SQLINTEGER s_block;
        SQLLEN     s_block_ind = 0;
        SQLINTEGER s_data;
        SQLLEN     s_data_ind = 0;

        /* handle cmd line params */
        if (argc == 2) {
            connstr = argv[1];
        } else if (argc > 2) {
            printf("Usage:  tf1 [\"<connect-string>\"]\n\n");
            exit(2);
        } else {
            connstr = connstr_default;
        }
        
        alloc_env(&henv);
        alloc_dbc(henv, &hdbc);

        /* connect */
        printf("main: TF connect: %s\n", connstr);
        sql_connect(hdbc, connstr, username, password);

        alloc_stmt(hdbc, &hstmt);

        /* initialize table */  
        printf("    : drop table\n");
        rc = SQLExecDirect(hstmt, "DROP TABLE tf1sample", SQL_NTS);
        /*   ignore table not existing */
        if (rc == SQL_ERROR
            && !check_error_native(hstmt, SQL_HANDLE_STMT,
                                   E_TABLE_NOT_EXIST)) {
            
            SUH(rc, hstmt, SQL_HANDLE_STMT);
        }
        printf("    : create table\n");
        rc = SQLExecDirect(hstmt, "CREATE TABLE tf1sample (idx INTEGER PRIMARY KEY, block INTEGER, data INTEGER)",
                           SQL_NTS);
        SU(rc);
        
        /* we don't need to realloc statements after a TF exception */
        alloc_stmt(hdbc, &inserth);
        alloc_stmt(hdbc, &selecth);

        goto workblock;
        
 tfexception:
        printf("\n");
        printf("  *** TF EXCEPTION ***\n");
        printf("\n");
        printf("workblock: rollback after TF exception\n");
        rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_ROLLBACK);
        SUH(rc, hdbc, SQL_HANDLE_DBC);

        printf("main: redo preparations\n");
        
 workblock:
        /* set autocommit off; if autocommit were on, we'd still need to
         * reprepare statements, but, of course, we wouldn't have to
         * reinsert anything--just check that we don't insert the same
         * things again.
         */
        printf("    : set autocommit off\n");
        SQLSetConnectAttr(hdbc, SQL_ATTR_AUTOCOMMIT, SQL_AUTOCOMMIT_OFF, 0);

        /* we need to (re)set all session SET commands after TF exception */
        printf("    : set isolation level for session\n");
        rc = SQLExecDirect(hstmt, "SET ISOLATION LEVEL READ COMMITTED",
                           SQL_NTS);
        TFSU(rc, hstmt);
        rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT);
        TFSUH(rc, hdbc, SQL_HANDLE_DBC);      
        
        /* prepare statement for insert */
        printf("    : prepare insert\n");
        rc = SQLPrepare(inserth, "INSERT INTO tf1sample VALUES (?,?,?)",
                        SQL_NTS);
        TFSU(rc, inserth);
        rc = SQLBindParameter(inserth, 1, SQL_PARAM_INPUT,
                              SQL_C_SLONG, SQL_INTEGER, 0, 0,
                              &i_index, 0, &i_index_ind);
        rc = SQLBindParameter(inserth, 2, SQL_PARAM_INPUT,
                              SQL_C_SLONG, SQL_INTEGER, 0, 0,
                              &i_block, 0, &i_block_ind);
        rc = SQLBindParameter(inserth, 3, SQL_PARAM_INPUT,
                              SQL_C_SLONG, SQL_INTEGER, 0, 0,
                              &i_data, 0, &i_data_ind);
        TFSU(rc, inserth);

        /* prepare statement for select */
        printf("    : prepare select\n");
        rc = SQLPrepare(selecth, "SELECT idx, block, data FROM tf1sample",
                        SQL_NTS);
        TFSU(rc, selecth);
        rc = SQLBindCol(selecth, 1, SQL_C_SLONG, &s_index,
                        0, &s_index_ind);
        rc = SQLBindCol(selecth, 2, SQL_C_SLONG, &s_block,
                        0, &s_block_ind);
        rc = SQLBindCol(selecth, 3, SQL_C_SLONG, &s_data,
                        0, &s_data_ind);
        TFSU(rc, selecth);

        /* do stuff */
        for (; i <= nloops; i++) {
            printf("main: do block %d\n", i);
            i_block = i;
            i_data  = i * 1000;
            
            /* read old data */
            printf("workblock: read old values\n");
            printf("         :  IDX  BLOCK  DATA\n");
            printf("         : ------------------\n");
            
            rc = SQLExecute(selecth);
            TFSU(rc, selecth);

            for (nrows = 0; TRUE; nrows++) { /* fetch rows */
                rc = SQLFetch(selecth);
                if (SQL_NO_DATA == rc) {
                    if (nrows == 0) {
                        printf("         :  -- none found --\n");
                    }
                    break;
                }
                TFSU(rc, selecth);
                printf("         : %3d  %4d  %6d\n",
                       (int) s_index, (int) s_block, (int) s_data);
            }
        
            /* insert some rows */
            printf("workblock: insert values\n");
            for (j = 0; j < 3; j++) {
                printf("         : ins %d, %d, %d\n", (int) i_index,
                       (int) i_block, (int) i_data);
                rc = SQLExecute(inserth);
                TFSU(rc, inserth);
                
                /* make time to disrupt the processing */
                sleep(1 * SLEEP_MULT);
                
                i_index++;
                i_data++;
            }
        
            /* commit */
            printf("workblock: commit\n");
            rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT);
            TFSUH(rc, hdbc, SQL_HANDLE_DBC);
        }
        
        /* start clean up */
        free_stmt(hstmt);
        free_stmt(inserth);
        free_stmt(selecth);
        
        /* disconnect */
        sql_disconnect(hdbc);

        free_dbc(hdbc);
        free_env(henv);

        printf("\nTF1 sample finished.\n");
        
        return 0;
}

/* EOF */
